package com.maryanovsky.gwtutils.client;

import java.util.Collections;
import java.util.List;
import java.util.Map;
import com.google.gwt.core.client.EntryPoint;
import com.google.gwt.core.client.GWT;
import com.google.gwt.http.client.RequestBuilder;
import com.google.gwt.http.client.RequestCallback;
import com.google.gwt.http.client.RequestException;
import com.google.gwt.user.client.Window;
import com.maryanovsky.gwtutils.client.net.HttpTextRequestMaker;
import com.maryanovsky.gwtutils.client.net.RequestMaker;
import com.maryanovsky.gwtutils.client.net.ResultCallback;
import com.maryanovsky.gwtutils.client.text.ParseException;

/**
 * A class which handles much of the heavy lifting of starting up an
 * application. The class handles the following:
 * <ul>
 *   <li>Parses the URL parameters of the page's URL and puts them in
 *       {@link #URL_ARGS}</li>
 *   <li>Sets the {@link #SETTINGS_URL} field to the URL of the application
 *       settings. By default this is simply "settings.properties", but it can
 *       also be specified by a URL parameter named "settingsUrl".</li>
 *   <li>Enables or disables window scrolling based on the value returned by
 *       {@link #isWindowScrollingEnabled()}, which can be overridden.</li>
 *   <li>Sets an uncaught exception handler to the value returned by
 *       {@link #getUncaughtExceptionHandler()}, which can be overridden.</li>
 *   <li>Makes the requests required to start the application. These are
 *       specified by {@link #getApplicationStartRequests()} which can be
 *       overridden.</li>
 *   <li>Once all the requests required to start the application have finished,
 *       invokes {@link #startApplication()}, which should be overridden</li>.
 * </ul>
 * 
 * @author Maryanovsky Alexander
 */
public abstract class Application implements EntryPoint {

    /**
	 * The arguments passed to us via the URL.
	 */
    protected static final Map<String, String> URL_ARGS;

    static {
        URL_ARGS = Utils.getUrlParams();
    }

    /**
	 * The URL of the settings file.
	 */
    protected final String SETTINGS_URL = getSettingsUrl();

    /**
	 * The {@link ResultCallback} for the settings request.
	 */
    protected final ResultCallback<String> settingsResponseCallback = new ResultCallback<String>() {

        public void requestFinished() {
        }

        public void errorOcurred(String message) {
            onErrorRetrievingSettings(SETTINGS_URL, message);
        }

        public void resultReceived(String response) {
            try {
                onSettingsReceived(response);
            } catch (ParseException c) {
                onInvalidSettings(c.getMessage());
            }
        }
    };

    /***/
    private int a = 0;

    /**
	 * The entry point method.
	 */
    public void onModuleLoad() {
        Window.enableScrolling(isWindowScrollingEnabled());
        GWT.setUncaughtExceptionHandler(getUncaughtExceptionHandler());
        List<RequestMakerResponseHandlerPair> c = getApplicationStartRequests();
        a = c.size();
        for (RequestMakerResponseHandlerPair d : c) {
            RequestMaker<String> e = d.getRequestMaker();
            ResultCallback<String> f = d.getResultCallback();
            try {
                e.sendRequest(new ProxyCallback(f));
            } catch (RequestException g) {
                onErrorMakingRequest(e.getUrl(), g.getMessage());
                b();
            }
        }
    }

    /**
	 * Returns the URL of the settings file.
	 */
    protected String getSettingsUrl() {
        String c = URL_ARGS.get("settingsUrl");
        if (c != null) return c; else return "settings.properties";
    }

    /**
	 * Returns the list of requests that need to be made for the application to
	 * start and their handlers. The default implementation returns a list with
	 * a single item - the pair
	 * <code>({@link #createSettingsRequestMaker()},
	 * {@link #settingsResponseCallback})</code>.
	 * If your application needs to retrieve more information in order to start,
	 * you should override this method and return additional request builders
	 * (and corresponding callbacks).
	 */
    protected List<RequestMakerResponseHandlerPair> getApplicationStartRequests() {
        return Collections.singletonList(new RequestMakerResponseHandlerPair(createSettingsRequestMaker(), settingsResponseCallback));
    }

    /**
	 * Creates the request maker for the settings file.
	 */
    protected RequestMaker<String> createSettingsRequestMaker() {
        return new HttpTextRequestMaker(new RequestBuilder(RequestBuilder.GET, SETTINGS_URL));
    }

    /***/
    private void b() {
        a--;
        if (a == 0) startApplication();
    }

    /**
	 * Returns whether window scrolling should be enabled for the application.
	 * Note that this method is called from {@link #onModuleLoad()}. The default
	 * implementation returns false.
	 */
    protected boolean isWindowScrollingEnabled() {
        return false;
    }

    /**
	 * Returns the uncaught exception handler for the application. Note that
	 * this method is called from {@link #onModuleLoad()}. The default
	 * implementation returns an exception handler which shows an (unlocalized)
	 * alert describing the exception that occurred.
	 */
    protected GWT.UncaughtExceptionHandler getUncaughtExceptionHandler() {
        return new GWT.UncaughtExceptionHandler() {

            public void onUncaughtException(Throwable e) {
                if (GWT.isScript()) {
                    while (e != null) {
                        String c = "";
                        for (StackTraceElement d : e.getStackTrace()) c += d.toString() + "\n";
                        showError("Uncaught exception occurred:\n" + e.getMessage() + "\n" + c);
                        e = e.getCause();
                    }
                } else {
                    GWT.log("Uncaught exception occurred", e);
                }
            }
        };
    }

    /**
	 * Invoked when making one of the requests required to start the application
	 * fails (that is,
	 * {@link RequestBuilder#sendRequest(String, RequestCallback)} threw a
	 * {@link RequestException}). The default implementation displays an
	 * (unlocalized) alert describing the error that occurred.
	 */
    protected void onErrorMakingRequest(String url, String errorMessage) {
        showError("Unable to access URL " + url + "\n" + errorMessage);
    }

    /**
	 * This method is called when an error occurs while retrieving the
	 * application settings. The default implementation shows an (unlocalized)
	 * alert describing the error.
	 */
    protected void onErrorRetrievingSettings(String settingsUrl, String errorMessage) {
        showError("Unable to retrieve settings from " + settingsUrl + "\n" + errorMessage);
    }

    /**
	 * Invoked when the settings are received. The method is meant to parse the
	 * settings, and store them in a field until {@link #startApplication()} is
	 * invoked (where they will be used).
	 */
    protected abstract void onSettingsReceived(String settingsText) throws ParseException;

    /**
	 * Invoked when the settings are invalid. The default implementation shows
	 * an (unlocalized) alert describing the error.
	 */
    protected void onInvalidSettings(String errorMessage) {
        showError("Invalid settings or URL params:\n" + errorMessage);
    }

    /**
	 * Displays the specified error message to the user. This is not an error
	 * that occurs normally. The default implementation shows it in an alert.
	 */
    protected void showError(String errorMessage) {
        Window.alert(errorMessage);
    }

    /**
	 * Starts the application. This method is invoked when all the requests
	 * required to start the application have finished. Note that this happens
	 * even if some of them failed - it is up to you, when overriding this
	 * method, to check whether you have all the information you need.
	 */
    protected abstract void startApplication();

    /***/
    private class ProxyCallback implements ResultCallback<String> {

        /***/
        private final ResultCallback<String> c;

        /**
		 * Creates a new {@link ProxyCallback} with the specified delegate
		 * callback.
		 */
        public ProxyCallback(ResultCallback<String> callback) {
            if (callback == null) throw new IllegalArgumentException("callback may not be null");
            this.c = callback;
        }

        /**
		 * Invokes the delegate callback.
		 */
        public void requestFinished() {
            c.requestFinished();
        }

        /**
		 * Invokes the delegate callback. and remembers that this request has
		 * been answered.
		 */
        public void errorOcurred(String errorMessage) {
            c.errorOcurred(errorMessage);
            b();
        }

        /**
		 * Invokes the delegate callback and remembers that this request has
		 * been answered.
		 */
        public void resultReceived(String result) {
            c.resultReceived(result);
            b();
        }
    }

    public static class RequestMakerResponseHandlerPair {

        private final RequestMaker<String> c;

        private final ResultCallback<String> d;

        public RequestMakerResponseHandlerPair(RequestMaker<String> requestMaker, ResultCallback<String> resultCallback) {
            if (requestMaker == null) throw new IllegalArgumentException("requestMaker may not be null");
            if (resultCallback == null) throw new IllegalArgumentException("resultCallback may not be null");
            this.c = requestMaker;
            this.d = resultCallback;
        }

        public RequestMaker<String> getRequestMaker() {
            return c;
        }

        public ResultCallback<String> getResultCallback() {
            return d;
        }
    }
}
