package com.maryanovsky.gwtutils.client;

/**
 * A boolean condition.
 * 
 * @author Maryanovsky Alexander
 */
public abstract class Condition<T> {

    /***/
    private static final Condition<Object> a = new Condition<Object>() {

        public boolean eval(Object argument) {
            return false;
        }
    };

    /***/
    private static final Condition<Object> b = new Condition<Object>() {

        public boolean eval(Object argument) {
            return true;
        }
    };

    /**
	 * The always-true condition.
	 */
    public static <T> Condition<T> truth() {
        return cast(a);
    }

    /**
	 * The always-false condition.
	 */
    public static <T> Condition<T> falsehood() {
        return cast(b);
    }

    /**
	 * Returns the logical-AND of this condition and the specified one. Note
	 * that the evaluation of the conditions is short-circuited in the standard
	 * way and that this condition is evaluated first.
	 */
    public Condition<T> and(final Condition<? super T> cond) {
        return new Condition<T>() {

            public boolean eval(T argument) {
                return Condition.this.eval(argument) && cond.eval(argument);
            }
        };
    }

    /**
	 * Returns the logical-AND of the specified conditions. Note that the
	 * evaluation of the conditions is short-circuited in the standard way.
	 */
    public static <T> Condition<T> and(final Condition<? super T> cond1, final Condition<? super T> cond2) {
        return new Condition<T>() {

            public boolean eval(T argument) {
                return cond1.eval(argument) && cond2.eval(argument);
            }
        };
    }

    /**
	 * Returns the logical-AND of the specified conditions. Note that the
	 * evaluation of the conditions is short-circuited in the standard way.
	 */
    public static <T> Condition<T> and(final Condition<? super T> cond1, final Condition<? super T> cond2, final Condition<? super T> cond3) {
        return new Condition<T>() {

            public boolean eval(T argument) {
                return cond1.eval(argument) && cond2.eval(argument) && cond3.eval(argument);
            }
        };
    }

    /**
	 * Returns the logical-AND of the specified conditions. Note that the
	 * evaluation of the conditions is short-circuited in the standard way.
	 */
    public static <T> Condition<T> and(final Condition<? super T> cond1, final Condition<? super T> cond2, final Condition<? super T> cond3, final Condition<? super T> cond4) {
        return new Condition<T>() {

            public boolean eval(T argument) {
                return cond1.eval(argument) && cond2.eval(argument) && cond3.eval(argument) && cond4.eval(argument);
            }
        };
    }

    /**
	 * Returns the logical-OR of this condition and the specified one. Note that
	 * the evaluation of the conditions is short-circuited in the standard way
	 * and that this condition is evaluated first.
	 */
    public Condition<T> or(final Condition<? super T> cond) {
        return new Condition<T>() {

            public boolean eval(T argument) {
                return Condition.this.eval(argument) || cond.eval(argument);
            }
        };
    }

    /**
	 * Returns the logical-OR of the specified conditions. Note that the
	 * evaluation of the conditions is short-circuited in the standard way. 
	 */
    public static <T> Condition<T> or(final Condition<? super T> cond1, final Condition<? super T> cond2) {
        return new Condition<T>() {

            public boolean eval(T argument) {
                return cond1.eval(argument) || cond2.eval(argument);
            }
        };
    }

    /**
	 * Returns the logical-OR of the specified conditions. Note that the
	 * evaluation of the conditions is short-circuited in the standard way. 
	 */
    public static <T> Condition<T> or(final Condition<? super T> cond1, final Condition<? super T> cond2, final Condition<? super T> cond3) {
        return new Condition<T>() {

            public boolean eval(T argument) {
                return cond1.eval(argument) || cond2.eval(argument) || cond3.eval(argument);
            }
        };
    }

    /**
	 * Returns the logical-OR of the specified conditions. Note that the
	 * evaluation of the conditions is short-circuited in the standard way. 
	 */
    public static <T> Condition<T> or(final Condition<? super T> cond1, final Condition<? super T> cond2, final Condition<? super T> cond3, final Condition<? super T> cond4) {
        return new Condition<T>() {

            public boolean eval(T argument) {
                return cond1.eval(argument) || cond2.eval(argument) || cond3.eval(argument) || cond4.eval(argument);
            }
        };
    }

    /**
	 * Returns the logical inverse of this condition.
	 */
    public Condition<T> not() {
        return not(this);
    }

    /**
	 * Returns the logical inverse of the specified condition.
	 */
    public static <T> Condition<T> not(final Condition<? super T> condition) {
        return new Condition<T>() {

            public boolean eval(T argument) {
                return !condition.eval(argument);
            }
        };
    }

    /**
	 * Casts a condition parameterized with a supertype of T to a condition
	 * parameterized with T. This is legal, as a condition that evaluates an
	 * argument of a certain type can always evaluate an argument of its
	 * subtype. 
	 */
    @SuppressWarnings("unchecked")
    public static <T> Condition<T> cast(Condition<? super T> condition) {
        return (Condition<T>) condition;
    }

    /**
	 * Casts this condition to a condition parameterized with a subtype of T.
	 * This is legal as a condition can always be evaluated with a subtype of T.
	 */
    public <S extends T> Condition<S> cast() {
        return cast(this);
    }

    /**
	 * Returns the value of the condition based on the specified argument.
	 */
    public abstract boolean eval(T argument);
}
