package com.maryanovsky.gwtutils.client;

import java.util.Map;

/**
 * Parses and provides access to a set of properties mapped to string keys. This
 * is quite similar to the {@link java.util.Properties} class, but provides a
 * little more functionality for accessing and parsing the properties.
 * 
 * 
 * @author Maryanovsky Alexander
 */
public class Properties {

    /***/
    private final Map<String, String> a;

    /***/
    private final Properties b;

    /**
	 * Creates a new {@link Properties} object with the specified map of
	 * actual, unparsed properties, and a backup {@link Properties} object which
	 * is consulted if a certain property is not found locally.
	 */
    public Properties(Map<String, String> props, Properties delegate) {
        if (props == null) throw new IllegalArgumentException("props may not be null");
        this.a = props;
        this.b = delegate;
    }

    /**
	 * Creates a new {@link Properties} object with the specified map of
	 * actual, unparsed properties.
	 */
    public Properties(Map<String, String> props) {
        this(props, null);
    }

    /**
	 * Returns the exception to throw when encountering a missing property. The
	 * default implementation returns an {@link IllegalArgumentException} with a
	 * short message and the name of the property.
	 */
    protected RuntimeException missingPropertyException(String propertyName) {
        return new IllegalArgumentException("Missing property: " + propertyName);
    }

    /**
	 * Returns the exception to throw when encountering an invalid property. The
	 * default implementation returns an {@link IllegalArgumentException} with a
	 * short message and the name and value of the property.
	 */
    protected RuntimeException invalidPropertyException(String propertyName, String propertyValue) {
        return new IllegalArgumentException("Invalid setting: " + propertyName + "=" + propertyValue);
    }

    /**
	 * Returns the value of the specified property, or the specified default
	 * value if there is no property with the specified name.
	 */
    public String get(String propertyName, String defaultValue) {
        String c = a.get(propertyName);
        if (c != null) return c;
        if (b != null) return b.get(propertyName, defaultValue); else return defaultValue;
    }

    /**
	 * Returns the value of specified property. Throws
	 * {@link #missingPropertyException(String)} if there is no property with the
	 * specified name.
	 */
    public String get(String propertyName) {
        String c = get(propertyName, null);
        if (c == null) throw missingPropertyException(propertyName);
        return c;
    }

    /**
	 * Returns the value of the specified property, parsed as an integer. Throws
	 * {@link #missingPropertyException(String)} if there is no property with the
	 * specified name, or {@link #invalidPropertyException(String, String)} if
	 * the property value can't be parsed as an integer.
	 * 
	 */
    public int getInt(String propertyName) {
        String c = get(propertyName);
        try {
            return Integer.parseInt(c);
        } catch (NumberFormatException d) {
            throw invalidPropertyException(propertyName, c);
        }
    }

    /**
	 * Returns the value of the specified property, parsed as an integer, or the
	 * specified default value if there is no property with the specified name.
	 * Throws {@link #invalidPropertyException(String, String)} if the property
	 * value can't be parsed as an integer.
	 */
    public int getInt(String propertyName, int defaultValue) {
        String c = get(propertyName, null);
        if (c == null) return defaultValue;
        try {
            return Integer.parseInt(c);
        } catch (NumberFormatException d) {
            throw invalidPropertyException(propertyName, c);
        }
    }

    /**
	 * Returns the value of the specified property, parsed as a floating point
	 * value. Throws {@link #missingPropertyException(String)} if there is no
	 * property with the specified name, or 
	 * {@link #invalidPropertyException(String, String)} if the property value
	 * can't be parsed as a floating point value.
	 */
    public double getDouble(String propertyName) {
        String c = get(propertyName);
        try {
            return Double.parseDouble(c);
        } catch (NumberFormatException d) {
            throw invalidPropertyException(propertyName, c);
        }
    }

    /**
	 * Returns the value of the specified property, parsed as a floating point
	 * value, or the specified default value if there is no property with the
	 * specified name. Throws {@link #invalidPropertyException(String, String)}
	 * if the property value can't be parsed as a floating point value.
	 */
    public double getDouble(String propertyName, double defaultValue) {
        String c = get(propertyName, null);
        if (c == null) return defaultValue;
        try {
            return Double.parseDouble(c);
        } catch (NumberFormatException d) {
            throw invalidPropertyException(propertyName, c);
        }
    }

    /**
	 * Returns the value of the specified property, parsed as a boolean value.
	 * All strings equal, ignoring case, to <code>"true"</code> are considered
	 * as <code>true</code> values, and all strings equal, ignoring case, to
	 * <code>"false"</code> are considered as <code>false</code>.
	 * Throws {@link #missingPropertyException(String)} if there is no property
	 * with the specified name, or 
	 * {@link #invalidPropertyException(String, String)} if the property value
	 * can't be parsed as a boolean.
	 */
    public boolean getBoolean(String propertyName) {
        String c = get(propertyName);
        if ("true".equalsIgnoreCase(c)) return true; else if ("false".equalsIgnoreCase(c)) return false; else throw invalidPropertyException(propertyName, c);
    }

    /**
	 * Returns the value of the specified property, parsed as a boolean value,
	 * or the specified default value if there is no such property.
	 * All strings equal, ignoring case, to <code>"true"</code> are considered
	 * as <code>true</code> values, and all strings equal, ignoring case, to
	 * <code>"false"</code> are considered as <code>false</code>.
	 * Throws {@link #invalidPropertyException(String, String)} if the property
	 * value can't be parsed as a boolean.
	 */
    public boolean getBoolean(String propertyName, boolean defaultValue) {
        String c = get(propertyName, null);
        if (c == null) return defaultValue; else if ("true".equalsIgnoreCase(c)) return true; else if ("false".equalsIgnoreCase(c)) return false; else throw invalidPropertyException(propertyName, c);
    }

    /**
	 * Returns the value of the specified property, parsed as a character.
	 * Throws {@link #missingPropertyException(String)} if there is no property
	 * with the specified name, or 
	 * {@link #invalidPropertyException(String, String)} if the property value
	 * can't be parsed as a character.
	 */
    public char getChar(String propertyName) {
        String c = get(propertyName);
        if (c.length() != 1) throw invalidPropertyException(propertyName, c); else return c.charAt(0);
    }

    /**
	 * Returns the value of the specified property, parsed as a character, or
	 * the specfied default value if there is no such property.
	 * Throws {@link #invalidPropertyException(String, String)} if the property
	 * value can't be parsed as a character.
	 */
    public char getChar(String propertyName, char defaultValue) {
        String c = get(propertyName);
        if (c == null) return defaultValue; else if (c.length() != 1) throw invalidPropertyException(propertyName, c); else return c.charAt(0);
    }

    /**
	 * Returns the value of the specified property, parsed a list of
	 * comma-separated strings. Throws {@link #missingPropertyException(String)}
	 * if there is no property with the specified name.
	 */
    public String[] getStringList(String propertyName) {
        String c = get(propertyName);
        return c.split(",");
    }

    /**
	 * Returns the value of the specified property, parsed as a list of
	 * comma-separated strings, or the specified default value if there is no
	 * property with the specified name.
	 */
    public String[] getStringList(String propertyName, String[] defaultValue) {
        String c = get(propertyName);
        if (c == null) return defaultValue; else return c.split(",");
    }
}
