package com.maryanovsky.gwtutils.client.event;

import java.util.HashSet;
import java.util.Set;

/**
 * Implements notifying a {@linkplain ReadyListener} when a collection of items
 * change between ready and unready states, as a group. A group of items are
 * ready iff all of them are ready. Note that the class implements
 * <code>ReadyListener</code> itself, so it can be simply used as a ready
 * listener for some set of items.
 * <code>IT</code> is the type of the items.
 * <code>ST</code> is the type of the object report as the source of ready
 * events.
 * 
 * @author Maryanovsky Alexander
 */
public class GroupReadyListener<IT, ST> implements ReadyListener<IT> {

    /***/
    private final ReadyListener<ST> a;

    /***/
    private final ST b;

    /***/
    private final Set<IT> c = new HashSet<IT>();

    /**
	 * Creates a new <code>GroupReadyListener</code> with the specified
	 * ready listener to report to, and the specified object to report as the
	 * source of ready events. All items are considered initially ready.
	 * 
	 * @param readyListener The listener to report ready events to.
	 * @param source The object to report as the source of ready events.
	 */
    public GroupReadyListener(ReadyListener<ST> readyListener, ST source) {
        if (readyListener == null) throw new IllegalArgumentException("readyListener may not be null");
        if (source == null) throw new IllegalArgumentException("source may not be null");
        this.b = source;
        this.a = readyListener;
    }

    /**
	 * Invoked when an item becomes ready.
	 */
    public void becameReady(IT source) {
        if (c.remove(source) && c.isEmpty()) a.becameReady(this.b);
    }

    /**
	 * Invoked when the item becomes unready.
	 */
    public void becameUnready(IT source) {
        boolean d = c.isEmpty();
        c.add(source);
        if (d) a.becameUnready(this.b);
    }

    /**
	 * Marks the specified item as ready, without notifiying the listener.
	 * This method is useful when a potentially unready item, is removed from
	 * the set of interest.
	 */
    public void quietlyMarkReady(IT item) {
        c.remove(item);
    }

    /**
	 * Marks all items as ready, without notifying the listener. This method is
	 * useful when the set of interest changes drastically.
	 */
    public void reset() {
        c.clear();
    }

    /**
	 * Marks the specified item as unready, without notifying the listener.
	 * This method is useful if you wish all your items to be initially unready.  
	 */
    public void quietlyMarkUnready(IT item) {
        c.add(item);
    }
}
