package com.maryanovsky.gwtutils.client.geom;

import com.maryanovsky.gwtutils.client.MathUtils;

/**
 * A two dimensional point, with integer coordinates.
 * 
 * @author Maryanovsky Alexander
 */
public final class Point implements PointView {

    /***/
    private int a;

    /***/
    private int b;

    /**
	 * Creates a new <code>Point</code> with the specified coordinates.
	 */
    public Point(int x, int y) {
        set(x, y);
    }

    /**
	 * Creates a new Point at 0,0.
	 */
    public Point() {
        this(0, 0);
    }

    /**
	 * Creates a new <code>Point</code> with the X coordinate at the width of
	 * the specified size and the Y coordinate at the height of the specified
	 * size.
	 */
    public Point(SizeView size) {
        this(size.getWidth(), size.getHeight());
    }

    /**
	 * Creates a new <code>Point</code> at the same coordinates as the specified
	 * one.
	 */
    public Point(PointView point) {
        this(point.getX(), point.getY());
    }

    /**
	 * Returns a <code>Point</code> with the specified coordinates. If the
	 * given <code>Point</code> is not <code>null</code>, it is reused and
	 * returned.
	 */
    public static Point make(int x, int y, Point point) {
        if (point == null) return new Point(x, y); else {
            point.set(x, y);
            return point;
        }
    }

    /**
	 * Sets both coordinates.
	 * 
	 * @return <code>this</code>.
	 */
    public Point set(int x, int y) {
        setX(x);
        setY(y);
        return this;
    }

    /**
	 * Sets both coordinates.
	 * 
	 * @return <code>this</code>.
	 */
    public Point set(PointView point) {
        return set(point.getX(), point.getY());
    }

    /**
	 * Moves this point by the specified offsets.
	 * 
	 * @return <code>this</code>.
	 */
    public Point moveBy(int dx, int dy) {
        return set(getX() + dx, getY() + dy);
    }

    /**
	 * Moves this point by the offsets specified by the given point.
	 * 
	 * @return <code>this</code>.
	 */
    public Point moveBy(PointView point) {
        return moveBy(point.getX(), point.getY());
    }

    /**
	 * Moves this point by the offsets specified by the given size.
	 * 
	 * @return <code>this</code>.
	 */
    public Point moveBy(SizeView size) {
        return moveBy(size.getWidth(), size.getHeight());
    }

    /**
	 * Multiples the X and Y coordinates of this point by the specified
	 * coefficient.
	 * 
	 * @return <code>this</code>.
	 */
    public Point multiplyBy(double coefficient) {
        return set((int) Math.round(getX() * coefficient), (int) Math.round(getY() * coefficient));
    }

    /**
	 * Returns the distance from this point to 0,0.
	 */
    public double getDistanceFromOrigin() {
        int c = getX();
        int d = getY();
        return Math.sqrt(c * c + d * d);
    }

    /**
	 * Returns the distance from this point to the specified one.
	 */
    public double getDistanceFrom(PointView point) {
        int c = getX() - point.getX();
        int d = getY() - point.getY();
        return Math.sqrt(c * c + d * d);
    }

    /**
	 * Returns the distance from this point to the specified one, as a
	 * <code>PointView</code>.
	 */
    public PointView getDistancePoint(PointView point) {
        return getDistancePoint(point, null);
    }

    /**
	 * Returns the distance from this point to the specified one, as a
	 * <code>Point</code>. If the specified 2nd point is not <code>null</code>,
	 * it is reused and returned.
	 */
    public Point getDistancePoint(PointView point, Point reusable) {
        return Point.make(point.getX() - this.getX(), point.getY() - this.getY(), reusable);
    }

    /**
	 * Returns the distance from this point to the specified one, as a
	 * <code>SizeView</code>. Note that this may differ from
	 * {@linkplain #getDistancePoint(PointView)} because the dimensions of a 
	 * <code>SizeView</code> are always non-negative.
	 */
    public SizeView getDistanceSize(PointView point) {
        return getDistanceSize(point, null);
    }

    /**
	 * Returns the distance from this point to the specified one in the same
	 * manner as {@linkplain #getDistanceSize(PointView)}. If the specified 2nd
	 * point is not <code>null</code>, it is reused and returned.
	 */
    public Size getDistanceSize(PointView point, Size reusable) {
        int c = Math.abs(point.getX() - this.getX());
        int d = Math.abs(point.getY() - this.getY());
        return Size.make(c, d, reusable);
    }

    /**
	 * Returns the point opposite to this one.
	 */
    public PointView getOpposite() {
        return getOpposite(null);
    }

    /**
	 * Returns the point opposite to this one. If the specified reusable point
	 * is not <code>null</code>, it is reused and returned.
	 */
    public Point getOpposite(Point reusable) {
        if (reusable == null) reusable = new Point();
        return reusable.set(-getX(), -getY());
    }

    /**
	 * Moves this point to the nearest location with smaller X and Y coordinates
	 * which is on a grid where each cell is of the specified size.
	 * 
	 * @return <code>this</code>.
	 */
    public Point roundDownToGrid(int cellWidth, int cellHeight) {
        setX(MathUtils.roundDownToMultiple(getX(), cellWidth));
        setY(MathUtils.roundDownToMultiple(getY(), cellHeight));
        return this;
    }

    /**
	 * Moves this point to the nearest location with smaller X and Y coordinates
	 * which is on a grid of the specified size.
	 * 
	 * @return <code>this</code>.
	 */
    public Point roundDownToGrid(SizeView cellSize) {
        return roundDownToGrid(cellSize.getWidth(), cellSize.getHeight());
    }

    /**
	 * Moves the point to the nearest location with greater X and Y coordinates
	 * which is on a grid where each cell is of the specified size.
	 * 
	 * @return <code>this</code>.
	 */
    public Point roundUpToGrid(int cellWidth, int cellHeight) {
        setX(MathUtils.roundUpToMultiple(getX(), cellWidth));
        setY(MathUtils.roundUpToMultiple(getY(), cellHeight));
        return this;
    }

    /**
	 * Moves the point to the nearest location with greater X and Y coordinates
	 * which is on a grid where each cell is of the specified size.
	 * 
	 * @return <code>this</code>.
	 */
    public Point roundUpToGrid(Size cellSize) {
        return roundUpToGrid(cellSize.getWidth(), cellSize.getHeight());
    }

    /**
	 * Reflects this point relative to the specified one.
	 * 
	 * @return <code>this</code>.
	 */
    public Point reflect(Point point) {
        int c = point.getX() - this.getX();
        int d = point.getY() - this.getY();
        return this.moveBy(2 * c, 2 * d);
    }

    /**
	 * Returns the mid-point between the two specified points. If the midpoint
	 * is not an integer, one of the near integer points is returned. If the
	 * specified reusable point is not <code>null</code>, it is reused and
	 * returned.
	 */
    public static Point getMidPoint(PointView point1, PointView point2, Point reusable) {
        if (reusable == null) reusable = new Point();
        return reusable.set((point1.getX() + point2.getX()) / 2, (point1.getY() + point2.getY()) / 2);
    }

    /**
	 * Returns the X coordinate.
	 */
    public int getX() {
        return a;
    }

    /**
	 * Sets the X coordinate.
	 * 
	 * @return <code>this</code>.
	 */
    public Point setX(int x) {
        this.a = x;
        return this;
    }

    /**
	 * Returns the Y coordinate.
	 */
    public int getY() {
        return b;
    }

    /**
	 * Sets the Y coordinate.
	 * 
	 * @return <code>this</code>.
	 */
    public Point setY(int y) {
        this.b = y;
        return this;
    }

    /**
	 * Returns whether the specified coordinates are the same as this point's.
	 */
    public boolean equals(int x, int y) {
        return (this.a == x) && (this.b == y);
    }

    /**
	 * Returns whether this <code>Point</code> is the same as the specified one.	 */
    public boolean equals(PointView point) {
        return equals(point.getX(), point.getY());
    }

    /**
	 * Returns whether this <code>Point</code> is the same as the specified one.
	 */
    public boolean equals(Object o) {
        if (!(o instanceof Point)) return false;
        return equals((Point) o);
    }

    /**
	 * Returns the hashcode of this object.
	 */
    public int hashCode() {
        return 37 * a + b;
    }

    /**
	 * Returns a textual representation of this point.
	 */
    public String toString() {
        return "Point[x=" + getX() + ", y=" + getY() + "]";
    }
}
