package com.maryanovsky.gwtutils.client.geom;

/**
 * A rectangle, in integer coordinates.
 * 
 * @author Maryanovsky Alexander
 */
public final class Rectangle implements RectangleView {

    /***/
    private int a;

    /***/
    private int b;

    /***/
    private int c;

    /***/
    private int d;

    /**
	 * Creates a new <code>Rectangle</code> with the specified arguments.
	 * Note that if <code>width</code> and/or <code>height</code> are negative,
	 * the rectangle is normalized (the top-left point is moved accordingly, and
	 * and the width and/or height are set to the absolute values of the
	 * specified ones).
	 * 
	 * @param x The X coordinate of the top-left point.
	 * @param y The Y coordinate of the top-left point.
	 * @param width The width of the rectangle.
	 * @param height The height of the rectangle.
	 */
    public Rectangle(int x, int y, int width, int height) {
        set(x, y, width, height);
    }

    /**
	 * Creates a new <code>Rectangle</code> of <code>0x0</code> size at 
	 * <code>0,0</code> coordinates.
	 */
    public Rectangle() {
        this(0, 0, 0, 0);
    }

    /**
	 * Creates a new rectangle with the same values as the specified one.
	 */
    public Rectangle(RectangleView rectangle) {
        this(rectangle.getX(), rectangle.getY(), rectangle.getWidth(), rectangle.getHeight());
    }

    /**
	 * Creates a new rectangle at the specified coordinates and with the
	 * specified size.
	 */
    public Rectangle(PointView point, SizeView size) {
        this(point.getX(), point.getY(), size.getWidth(), size.getHeight());
    }

    /**
	 * Returns a <code>Rectangle</code> with the specified values. If the given
	 * <code>Rectangle</code> is not <code>null</code>, it is reused and
	 * returned.
	 */
    public static Rectangle make(int x, int y, int width, int height, Rectangle reusable) {
        if (reusable == null) return new Rectangle(x, y, width, height); else {
            reusable.set(x, y, width, height);
            return reusable;
        }
    }

    /**
	 * Returns the smallest rectangle which contains both specified points. 
	 */
    public static Rectangle makeCover(PointView point1, PointView point2) {
        return new Rectangle().setCover(point1, point2);
    }

    /**
	 * Sets the values of this rectangle.
	 * 
	 * @return <code>this</code>.
	 */
    public Rectangle set(int x, int y, int width, int height) {
        setX(x);
        setY(y);
        setWidth(width);
        setHeight(height);
        return this;
    }

    /**
	 * Sets the values of this rectangle.
	 */
    public Rectangle set(PointView point, SizeView size) {
        return set(point.getX(), point.getY(), size.getWidth(), size.getHeight());
    }

    /**
	 * Sets the values of this rectangle to those of the specified one.
	 */
    public Rectangle set(RectangleView rect) {
        return set(rect.getX(), rect.getY(), rect.getWidth(), rect.getHeight());
    }

    /**
	 * Returns the size of the rectangle.
	 */
    public SizeView getSize() {
        return getSize(null);
    }

    /**
	 * Returns the size of the rectangle. If the specified <code>Size</code>
	 * object is not <code>null</code>, it is reused and returned.
	 */
    public Size getSize(Size reusable) {
        return Size.make(getWidth(), getHeight(), reusable);
    }

    /**
	 * Sets the size of the rectangle. If the specified width and/or height are
	 * negative, the rectangle is normalized.
	 * 
	 * @return <code>this</code>.
	 */
    public Rectangle setSize(int width, int height) {
        setWidth(width);
        setHeight(height);
        return this;
    }

    /**
	 * Sets the size of the rectangle.
	 * 
	 * @return <code>this</code>.
	 */
    public Rectangle setSize(SizeView size) {
        return setSize(size.getWidth(), size.getHeight());
    }

    /**
	 * Returns the location of the top-left point.
	 */
    public PointView getTopLeft() {
        return getTopLeft(null);
    }

    /**
	 * Returns the location of the top-left point of the rectangle. If the
	 * specified <code>Point</code> object is not <code>null</code>, it is
	 * reused and returned.
	 */
    public Point getTopLeft(Point reusable) {
        return Point.make(getX(), getY(), reusable);
    }

    /**
	 * Returns the location of the top-right point.
	 */
    public PointView getTopRight() {
        return getTopRight(null);
    }

    /**
	 * Returns the location of the top-right point of the rectangle. If the
	 * specified <code>Point</code> object is not <code>null</code>, it is
	 * reused and returned.
	 */
    public Point getTopRight(Point reusable) {
        return Point.make(getX() + getWidth(), getY(), reusable);
    }

    /**
	 * Returns the location of the bottom-left point.
	 */
    public PointView getBottomLeft() {
        return getBottomLeft(null);
    }

    /**
	 * Returns the location of the bottom-left point of the rectangle. If the
	 * specified <code>Point</code> object is not <code>null</code>, it is
	 * reused and returned.
	 */
    public Point getBottomLeft(Point reusable) {
        return Point.make(getX(), getY() + getHeight(), reusable);
    }

    /**
	 * Returns the location of the bottom-right point.
	 */
    public PointView getBottomRight() {
        return getBottomRight(null);
    }

    /**
	 * Returns the location of the bottom-right point of the rectangle. If the
	 * specified <code>Point</code> object is not <code>null</code>, it is
	 * reused and returned.
	 */
    public Point getBottomRight(Point reusable) {
        return Point.make(getX() + getWidth(), getY() + getHeight(), reusable);
    }

    /**
	 * Returns the location of the center of the rectangle. If the rectangle's
	 * width and/or height are even (so that the center point's coordinates are
	 * not integers), one of the closest integer points is returned, but which
	 * is unspecified.
	 */
    public PointView getCenter() {
        return getCenter(null);
    }

    /**
	 * Returns the location of the center of the rectangle, in the same manner
	 * as the {@linkplain #getCenter()} method. If the specified
	 * <code>Point</code> object is not <code>null</code>, it is reused and
	 * returned.
	 */
    public Point getCenter(Point reusable) {
        return Point.make(getX() + getWidth() / 2, getY() + getHeight() / 2, reusable);
    }

    /**
	 * Sets the location of the top-left point.
	 * 
	 * @return <code>this</code>.
	 */
    public Rectangle setTopLeft(int x, int y) {
        setX(x);
        setY(y);
        return this;
    }

    /**
	 * Sets the location of the top-left point.
	 * 
	 * @return <code>this</code>.
	 */
    public Rectangle setTopLeft(PointView point) {
        return setTopLeft(point.getX(), point.getY());
    }

    /**
	 * Moves the rectangle by the specified offsets.
	 * 
	 * @return <code>this</code>.
	 */
    public Rectangle moveBy(int dx, int dy) {
        return setTopLeft(getX() + dx, getY() + dy);
    }

    /**
	 * Moves the rectangle by the offsets specified by the given size.
	 * 
	 * @return <code>this</code>.
	 */
    public Rectangle moveBy(SizeView size) {
        return moveBy(size.getWidth(), size.getHeight());
    }

    /**
	 * Moves the rectangle by the offsets specified by the given point.
	 * 
	 * @return <code>this</code>.
	 */
    public Rectangle moveBy(PointView point) {
        return moveBy(point.getX(), point.getY());
    }

    /**
	 * Grows the dimensions of the rectangle by the specified values.
	 * 
	 * @return <code>this</code>.
	 */
    public Rectangle growBy(int dx, int dy) {
        setSize(getWidth() + dx, getHeight() + dy);
        return this;
    }

    /**
	 * Grows the dimensions of the rectangle by the specified size.
	 * 
	 * @return <code>this</code>.
	 */
    public Rectangle growBy(SizeView size) {
        return growBy(size.getWidth(), size.getHeight());
    }

    /**
	 * Grows the dimensions of the rectangle by the values of the specified
	 * point.
	 */
    public Rectangle growBy(PointView point) {
        return growBy(point.getX(), point.getY());
    }

    /**
	 * Multiplies the coordinates and dimensions of this rectangle by the
	 * specified coefficient.
	 * 
	 * @return <code>this</code>. 
	 */
    public Rectangle multiplyBy(double coefficient) {
        return set((int) Math.round(getX() * coefficient), (int) Math.round(getY() * coefficient), (int) Math.round(getWidth() * coefficient), (int) Math.round(getHeight() * coefficient));
    }

    /**
	 * Sets this rectangle to the smallest rectangle which contains both
	 * specified points.
	 * 
	 * @return <code>this</code>.
	 */
    public Rectangle setCover(PointView point1, PointView point2) {
        return set(point1.getX(), point1.getY(), point2.getX() - point1.getX(), point2.getY() - point1.getY());
    }

    /**
	 * Returns whether this rectangle contains the specified point.
	 */
    public boolean contains(PointView point) {
        return (point.getX() >= getX()) && (point.getY() >= getY()) && (point.getX() <= getX() + getWidth()) && (point.getY() <= getY() + getHeight());
    }

    /**
	 * Returns whether this rectangle contains the specified one.
	 */
    public boolean contains(RectangleView rectangle) {
        return (this.getX() <= rectangle.getX()) && (this.getY() <= rectangle.getY()) && (this.getX() + this.getWidth() >= rectangle.getX() + rectangle.getWidth()) && (this.getY() + this.getHeight() >= rectangle.getY() + rectangle.getHeight());
    }

    /**
	 * Returns whether this rectangle intersects with the specified one.
	 */
    public boolean intersects(RectangleView rectangle) {
        return (this.getX() < rectangle.getX() + rectangle.getWidth()) && (this.getY() < rectangle.getY() + rectangle.getHeight()) && (this.getX() + this.getWidth() > rectangle.getX()) && (this.getY() + this.getHeight() > rectangle.getY());
    }

    /**
	 * Returns the intersection between this rectangle and the specified one.
	 * Returns <code>null</code> if the two rectangles do not intersect and do
	 * not even "touch". Returns an empty rectangle if the two rectangles
	 * "touch".
	 */
    public RectangleView getIntersection(RectangleView rectangle) {
        return getIntersection(rectangle, null);
    }

    /**
	 * Returns the intersection between this rectangle and the specified one.
	 * Returns <code>null</code> if the two rectangles do not intersect and do
	 * not even "touch". Returns an empty rectangle if the two rectangles
	 * "touch".
	 * If the specified <code>Rectangle</code> object is not <code>null</code>,
	 * it is reused and returned (unless, as above, <code>null</code> is
	 * returned).
	 */
    public Rectangle getIntersection(RectangleView rectangle, Rectangle reusable) {
        int e = Math.max(this.getLeft(), rectangle.getLeft());
        int f = Math.max(this.getTop(), rectangle.getTop());
        int g = Math.min(this.getRight(), rectangle.getRight());
        int h = Math.min(this.getBottom(), rectangle.getBottom());
        int i = g - e;
        int j = h - f;
        if ((i < 0) || (j < 0)) return null;
        if (reusable == null) reusable = new Rectangle();
        reusable.set(e, f, i, j);
        return reusable;
    }

    /**
	 * Expands this rectangle so that it lies on the grid with cells of the
	 * specified size.
	 * 
	 * @return <code>this</code>.
	 */
    public Rectangle expandToGrid(int cellWidth, int cellHeight) {
        Point e = new Point(getTopLeft()).roundDownToGrid(cellWidth, cellHeight);
        Point f = new Point(getBottomRight()).roundUpToGrid(cellWidth, cellHeight);
        return setCover(e, f);
    }

    /**
	 * Expands this rectangle so that it lies on the grid with cells of the
	 * specified size.
	 * 
	 * @return <code>this</code>.
	 */
    public Rectangle expandToGrid(SizeView cellSize) {
        return expandToGrid(cellSize.getWidth(), cellSize.getHeight());
    }

    /**
	 * Returns the X coordinate of the top-left point.
	 */
    public int getX() {
        return a;
    }

    /**
	 * Sets the X coordinate of the top-left point.
	 * 
	 * @return <code>this</code>.
	 */
    public Rectangle setX(int x) {
        this.a = x;
        return this;
    }

    /**
	 * Returns the Y coordinate of the top-left point.
	 */
    public int getY() {
        return b;
    }

    /**
	 * Sets the Y coordinate of the top-left point.
	 * 
	 * @return <code>this</code>.
	 */
    public Rectangle setY(int y) {
        this.b = y;
        return this;
    }

    /**
	 * Returns the width of the rectangle.
	 */
    public int getWidth() {
        return c;
    }

    /**
	 * Sets the width of the rectangle. If the specified value is negative, the
	 * rectangle is normalized (the X coordinate is moved accordingly and the
	 * width is set to the absolute value of the specified one).
	 * 
	 * @return <code>this</code>.
	 */
    public Rectangle setWidth(int width) {
        this.a = Math.min(this.a, this.a + width);
        this.c = Math.abs(width);
        return this;
    }

    /**
	 * Returns the height of the rectangle.
	 */
    public int getHeight() {
        return d;
    }

    /**
	 * Sets the height of the rectangle. If the specified value is negative, the
	 * rectangle is normalized (the Y coordinate is moved accordingly and the
	 * height is set to the absolute value of the specified one).
	 * 
	 * @return <code>this</code>.
	 */
    public Rectangle setHeight(int height) {
        this.b = Math.min(this.b, this.b + height);
        this.d = Math.abs(height);
        return this;
    }

    /**
	 * Same as {@link #getX()}.
	 */
    public int getLeft() {
        return a;
    }

    /**
	 * Same as {@link #getY()}.
	 */
    public int getTop() {
        return b;
    }

    /**
	 * Same as <code>getX() + getWidth()</code>.
	 */
    public int getRight() {
        return a + c;
    }

    /**
	 * Same as <code>getY() + getHeight()</code>.
	 */
    public int getBottom() {
        return b + d;
    }

    /**
	 * Returns whether the coordinates and size of this rectangle are the
	 * specified ones.
	 */
    public boolean equals(int x, int y, int width, int height) {
        return (this.a == x) && (this.b == y) && (this.c == width) && (this.d == height);
    }

    /**
	 * Returns whether the specified rectangle is the same as this one.
	 */
    public boolean equals(RectangleView rect) {
        return equals(rect.getX(), rect.getY(), rect.getWidth(), rect.getHeight());
    }

    /**
	 * Returns whether this <code>Rectangle</code> object has the same
	 * coordinates and size as this one.
	 */
    public boolean equals(Object o) {
        if (!(o instanceof Rectangle)) return false;
        return equals((Rectangle) o);
    }

    /**
	 * Returns the hashcode of this object.
	 */
    public int hashCode() {
        int e = a;
        e = 37 * e + b;
        e = 37 * e + c;
        e = 37 * e + d;
        return e;
    }

    /**
	 * Returns a textual representation of this rectangle.
	 */
    public String toString() {
        return "Rectangle[x=" + getX() + ", y=" + getY() + ", width=" + getWidth() + ", height=" + getHeight() + "]";
    }
}
