package com.maryanovsky.gwtutils.client.geom;

/**
 * Provides a view of a {@linkplain Rectangle} without allowing modifications.
 * 
 * @author Maryanovsky Alexander
 */
public interface RectangleView {

    /**
	 * Returns the X coordinate of the top-left point.
	 */
    int getX();

    /**
	 * Returns the Y coordinate of the top-left point.
	 */
    int getY();

    /**
	 * Returns the width of the rectangle.
	 */
    int getWidth();

    /**
	 * Returns the height of the rectangle.
	 */
    int getHeight();

    /**
	 * Same as {@link #getX()}.
	 */
    int getLeft();

    /**
	 * Same as {@link #getY()}.
	 */
    int getTop();

    /**
	 * Same as <code>getX() + getWidth()</code>.
	 */
    int getRight();

    /**
	 * Same as <code>getY() + getHeight()</code>.
	 */
    int getBottom();

    /**
	 * Returns the size of the rectangle.
	 */
    SizeView getSize();

    /**
	 * Returns the size of the rectangle. If the specified <code>Size</code>
	 * object is not <code>null</code>, it is reused and returned.
	 */
    Size getSize(Size reusable);

    /**
	 * Returns the location of the top-left point of the rectangle.
	 */
    PointView getTopLeft();

    /**
	 * Returns the location of the top-left point of the rectangle. If the
	 * specified <code>Point</code> object is not <code>null</code>, it is
	 * reused and returned.
	 */
    Point getTopLeft(Point reusable);

    /**
	 * Returns the location of the top-right point of the rectangle.
	 */
    PointView getTopRight();

    /**
	 * Returns the location of the top-right point of the rectangle. If the
	 * specified <code>Point</code> object is not <code>null</code>, it is
	 * reused and returned.
	 */
    Point getTopRight(Point reusable);

    /**
	 * Returns the location of the bottom-left point of the rectangle.
	 */
    PointView getBottomLeft();

    /**
	 * Returns the location of the bottom-left point of the rectangle. If the
	 * specified <code>Point</code> object is not <code>null</code>, it is
	 * reused and returned.
	 */
    Point getBottomLeft(Point reusable);

    /**
	 * Returns the location of the bottom-right point of the rectangle.
	 */
    PointView getBottomRight();

    /**
	 * Returns the location of the bottom-right point of the rectangle. If the
	 * specified <code>Point</code> object is not <code>null</code>, it is
	 * reused and returned.
	 */
    Point getBottomRight(Point reusable);

    /**
	 * Returns the location of the center of the rectangle. If the rectangle's
	 * width and/or height are even (so that the center point's coordinates are
	 * not integers), one of the closest integer points is returned, but which
	 * is unspecified.
	 */
    PointView getCenter();

    /**
	 * Returns the location of the center of the rectangle, in the same manner
	 * as the {@linkplain #getCenter()} method. If the specified
	 * <code>Point</code> object is not <code>null</code>, it is reused and
	 * returned.
	 */
    Point getCenter(Point reusable);

    /**
	 * Returns whether this rectangle contains the specified point.
	 */
    boolean contains(PointView point);

    /**
	 * Returns whether this rectangle contains the specified one.
	 */
    boolean contains(RectangleView rectangle);

    /**
	 * Returns whether this rectangle intersects with the specified one.
	 */
    boolean intersects(RectangleView rectangle);

    /**
	 * Returns the intersection between this rectangle and the specified one.
	 * Returns <code>null</code> if the two rectangles do not intersect and do
	 * not even "touch". Returns an empty rectangle if the two rectangles
	 * "touch".
	 */
    RectangleView getIntersection(RectangleView rectangle);

    /**
	 * Returns the intersection between this rectangle and the specified one.
	 * Returns <code>null</code> if the two rectangles do not intersect and do
	 * not even "touch". Returns an empty rectangle if the two rectangles
	 * "touch".
	 * If the specified <code>Rectangle</code> object is not <code>null</code>,
	 * it is reused and returned.
	 */
    Rectangle getIntersection(RectangleView rectangle, Rectangle reusable);
}
