package com.maryanovsky.gwtutils.client.geom;

/**
 * Defines two dimensional bounds, with integer coordinates.
 * 
 * @author Maryanovsky Alexander
 */
public final class Size implements SizeView {

    /***/
    private int a;

    /***/
    private int b;

    /**
	 * Creates a new <code>Size</code> with the specified width and height.
	 */
    public Size(int width, int height) {
        set(width, height);
    }

    /**
	 * Creates a new <code>0x0</code> <code>Size</code>.
	 */
    public Size() {
        this(0, 0);
    }

    /**
	 * Creates a new <code>Size</code> of the same size as the specified one.
	 */
    public Size(SizeView size) {
        this(size.getWidth(), size.getHeight());
    }

    /**
	 * Returns a <code>Size</code> with the specified size. If the given
	 * <code>Size</code> object is not <code>null</code>, it is reused and
	 * returned.
	 */
    public static Size make(int width, int height, Size reusable) {
        if (reusable == null) return new Size(width, height); else {
            reusable.set(width, height);
            return reusable;
        }
    }

    /**
	 * Sets the size of this <code>Size</code> object.
	 * 
	 * @return <code>this</code>.
	 */
    public Size set(int width, int height) {
        setWidth(width);
        setHeight(height);
        return this;
    }

    /**
	 * Sets the size of this <code>Size</code> object from the values of the
	 * specified one.
	 * 
	 * @return <code>this</code>.
	 */
    public Size set(SizeView size) {
        return set(size.getWidth(), getHeight());
    }

    /**
	 * Multiplies the dimensions of the <code>Size</code> by the specified
	 * coefficient.
	 * 
	 * @return <code>this</code>. 
	 */
    public Size multiplyBy(double coefficient) {
        return set((int) Math.round(getWidth() * coefficient), (int) Math.round(getHeight() * coefficient));
    }

    /**
	 * Returns the width.
	 */
    public int getWidth() {
        return a;
    }

    /**
	 * Sets the width.
	 * 
	 * @return <code>this</code>.
	 */
    public Size setWidth(int width) {
        if (width < 0) throw new IllegalArgumentException("width may not be negative: " + width);
        this.a = width;
        return this;
    }

    /**
	 * Returns the height.
	 * 
	 */
    public int getHeight() {
        return b;
    }

    /**
	 * Sets the height.
	 * 
	 * @return <code>this</code>.
	 */
    public Size setHeight(int height) {
        if (height < 0) throw new IllegalArgumentException("height may not be negative: " + height);
        this.b = height;
        return this;
    }

    /**
	 * Returns <code>true</code> if either the width or height are
	 * <code>0</code>.
	 */
    public boolean isEmpty() {
        return (b == 0) || (a == 0);
    }

    /**
	 * Returns whether this <code>Size</code> object is of the specified width
	 * and height.
	 */
    public boolean equals(int width, int height) {
        return (this.a == width) && (this.b == height);
    }

    /**
	 * Returns whether this <code>Size</code> object has the same width and
	 * height as the specified one.
	 */
    public boolean equals(SizeView size) {
        return equals(size.getWidth(), size.getHeight());
    }

    /**
	 * Returns whether this <code>Size</code> object has the same width and
	 * height as the specified one.
	 */
    public boolean equals(Object o) {
        if (!(o instanceof Size)) return false;
        return equals((Size) o);
    }

    /**
	 * Returns the hashcode of this object.
	 */
    public int hashCode() {
        return 37 * a + b;
    }

    /**
	 * Returns a textual representation of this size.
	 */
    public String toString() {
        return "Size[width=" + getWidth() + ", height=" + getHeight() + "]";
    }
}
