package com.maryanovsky.gwtutils.client.net;

import com.google.gwt.http.client.Request;
import com.google.gwt.http.client.RequestCallback;
import com.google.gwt.http.client.Response;
import com.google.gwt.user.client.Window;

/**
 * A base implementation of a <code>RequestCallback</code>.
 */
public abstract class AbstractResponseHandler implements RequestCallback {

    /**
	 * The URL of the request.
	 */
    protected final String requestUrl;

    /**
	 * Creates a new {@link AbstractResponseHandler} with the specified URL of
	 * the request.
	 */
    public AbstractResponseHandler(String requestUrl) {
        if (requestUrl == null) throw new IllegalArgumentException("requestUrl may not be null");
        this.requestUrl = requestUrl;
    }

    /**
	 * Invokes {@link #onAnyResponse()} and then
	 * {@link #onRequestError(Request, Throwable)}.
	 */
    public final void onError(Request request, Throwable exception) {
        onAnyResponse();
        onRequestError(request, exception);
    }

    /**
	 * If the response code is 200, parses the response and passes the result
	 * to {@link #onSuccessfulResponse(Response)}. If the response code is not
	 * 200, invokes {@link #onUnexpectedStatusCode(Request, Response)}.
	 */
    public void onResponseReceived(Request request, Response response) {
        onAnyResponse();
        int a = response.getStatusCode();
        if ((a != 200) && (a != 0)) {
            onUnexpectedStatusCode(request, response);
            return;
        }
        onSuccessfulResponse(response);
    }

    /**
	 * Invoked when a successful response has arrived. The default
	 * implementation does nothing.
	 */
    protected abstract void onSuccessfulResponse(Response response);

    /**
	 * Invoked when an error occurs during the request. The default
	 * implementation displays an error message in an alert window.
	 * 
	 * @param request The request during which the error occurred.
	 * @param exception The error that occurred.
	 */
    protected void onRequestError(Request request, Throwable exception) {
        Window.alert("Error retrieving URL:\n" + requestUrl + "\n" + exception.getMessage());
    }

    /**
	 * Invoked when the HTTP response code is not 200. The default
	 * implementation displays an error message in an alert window.
	 * 
	 * @param request The request that was sent to the server.
	 * @param response The response received from the server.
	 */
    protected void onUnexpectedStatusCode(Request request, Response response) {
        Window.alert("Unexpected response code (" + response.getStatusCode() + " " + response.getStatusText() + ") " + "retrieving URL:\n" + requestUrl);
    }

    /**
	 * Invoked whenever any kind of response arrives, before any of the other
	 * methods. The default implementation does nothing.
	 */
    protected void onAnyResponse() {
    }
}
