package com.maryanovsky.gwtutils.client.net;

import java.util.Map;
import com.google.gwt.http.client.URL;

/**
 * Constructs a URL from a base URL and parameters.
 * 
 * @author Maryanovsky Alexander
 */
public final class UrlBuilder {

    /***/
    private final StringBuffer url;

    /***/
    private boolean hasParameters;

    /**
	 * Creates a new {@link UrlBuilder} with the specified base URL.
	 */
    public UrlBuilder(String baseUrl) {
        if (baseUrl == null) throw new IllegalArgumentException("baseUrl may not be null");
        this.url = new StringBuffer(baseUrl);
        this.hasParameters = url.indexOf("?") != -1;
    }

    /**
	 * Creates a copy of the specified {@link UrlBuilder}. This is useful if you
	 * will be creating many URLs which share some of the parameters.
	 */
    public UrlBuilder(UrlBuilder urlBuilder) {
        this.url = new StringBuffer(urlBuilder.url.toString());
        this.hasParameters = urlBuilder.hasParameters;
    }

    /**
	 * Adds the specified parameter to the URL.
	 * 
	 * @param name The name of the URL parameter. May not be <code>null</code>
	 * or an empty string.
	 * @param value The value of the URL parameter. Converted to a
	 * <code>String</code> by invoking <code>toString</code> on it. May be
	 * <code>null</code>, in which case only the parameter name is appended
	 * (no "=" is appended).
	 * 
	 * @return this
	 */
    public UrlBuilder addParam(String name, Object value) {
        if (name == null) throw new IllegalArgumentException("name may not be null");
        if ("".equals(name)) throw new IllegalArgumentException("name may not be empty");
        url.append(hasParameters ? "&" : "?");
        if (value == null) url.append(URL.encodeQueryString(name)); else {
            url.append(URL.encodeQueryString(name));
            url.append("=");
            url.append(URL.encodeQueryString(value.toString()));
        }
        hasParameters = true;
        return this;
    }

    /**
	 * Adds the specified integer parameter to the URL.
	 * 
	 * @param name The name of the URL parameter. May not be <code>null</code>
	 * or an empty string.
	 * @param value The value of the URL parameter. Converted to a
	 * <code>String</code> via {@link String#valueOf(int)}.
	 * 
	 * @return this
	 */
    public UrlBuilder addParam(String name, int value) {
        return addParam(name, String.valueOf(value));
    }

    /**
	 * Adds the specified double parameter to the URL.
	 * 
	 * @param name The name of the URL parameter. May not be <code>null</code>
	 * or an empty string.
	 * @param value The value of the URL parameter. Converted to a
	 * <code>String</code> via {@link String#valueOf(double)}.
	 * 
	 * @return this
	 */
    public UrlBuilder addParam(String name, double value) {
        return addParam(name, String.valueOf(value));
    }

    /**
	 * Adds all the specified parameters to the URL.
	 * 
	 * @param parameters A map from parameter names to parameter values.
	 */
    public UrlBuilder addParams(Map<String, ? extends Object> parameters) {
        for (Map.Entry<String, ? extends Object> a : parameters.entrySet()) addParam(a.getKey(), a.getValue());
        return this;
    }

    /**
	 * Returns the built URL.
	 */
    public String getUrl() {
        return url.toString();
    }

    /**
	 * Returns the URL.
	 */
    public String toString() {
        return getUrl();
    }
}
