package com.maryanovsky.gwtutils.client.net;

import com.google.gwt.http.client.Response;
import com.google.gwt.user.client.Window;
import com.google.gwt.xml.client.*;
import com.maryanovsky.gwtutils.client.text.ParseException;

/**
 * A base implementation of a <code>RequestCallback</code> which handles XML
 * responses. An {@link XmlResponseHandler} can be either parameterized by a
 * parser and a callback to parse and handle the response XML document or by
 * subclassing and overriding the appropriate method. See the constructor(s) for
 * more information.
 *
 * @author Maryanovsky Alexander
 */
public class XmlResponseHandler<T> extends AbstractResponseHandler {

    /***/
    private final String a;

    /***/
    private final String b;

    /***/
    private final XmlParser<T> c;

    /***/
    private final ResultCallback<T> d;

    /**
	 * Creates a new {@link XmlResponseHandler} with the specified arguments.
	 * If <code>parser</code> is not <code>null</code>, the default
	 * implementation of {@link #parseResponse(Document)} will use it to parse
	 * the received XML document; otherwise, it will return the document itself 
	 * as the result. This means that <code>T</code> <strong>must</strong> be
	 * {@link Document} if <code>parser</code> is <code>null</code> and
	 * {@link #parseResponse(Document)} is not overridden.
	 * If <code>callback</code> is not <code>null</code>,
	 * {@link #onAnyResponse()} and {@link #onResult(Object)} will invoke the
	 * appropriate method on it. 
	 * 
	 * @param requestUrl The URL of the request.
	 * @param successfulResponseTagName The name of the document's main tag in
	 * case of a successful response.
	 * @param errorResponseTagName The name of the document's main tag in case
	 * of an error response.
	 * @param parser The parser we use to parse the response XML document; may
	 * be <code>null</code>.
	 * @param callback The callback to notify when a response/result arrives;
	 * may be <code>null</code>.
	 */
    public XmlResponseHandler(String requestUrl, String successfulResponseTagName, String errorResponseTagName, XmlParser<T> parser, ResultCallback<T> callback) {
        super(requestUrl);
        if (successfulResponseTagName == null) throw new IllegalArgumentException("successfulResponseTagName may not be null");
        if (errorResponseTagName == null) throw new IllegalArgumentException("errorResponseTagName may not be null");
        this.a = successfulResponseTagName;
        this.b = errorResponseTagName;
        this.c = parser;
        this.d = callback;
    }

    /**
	 * Returns the request URL.
	 */
    protected String getRequestUrl() {
        return requestUrl;
    }

    /**
	 * Invoked when a successful response arrives. Parses the response as an XML
	 * document and passes it on to either
	 * {@link #onSuccessfulXmlResponse(Document)},
	 * {@link #onErrorXmlResponse(Document)} or
	 * {@link #onUnrecognizedResponse(Document)} depending on the root tag name
	 * of the response document. If the response could not be parsed as an XML
	 * document, invoked {@link #onBadlyFormedXml(Response, DOMException)}.
	 */
    protected void onSuccessfulResponse(Response response) {
        Document e;
        try {
            e = XMLParser.parse(response.getText());
        } catch (DOMException f) {
            onBadlyFormedXml(response, f);
            return;
        }
        Element g = e.getDocumentElement();
        String h = g.getTagName();
        if (b.equals(h)) onErrorXmlResponse(e); else if (a.equals(h)) onSuccessfulXmlResponse(e); else onUnrecognizedResponse(e);
    }

    /**
	 * Invoked when the result could not be parsed as an XML document.
	 * The default implementation shows an error message.
	 */
    protected void onBadlyFormedXml(Response response, DOMException e) {
        showErrorMessage("Badly formed XML.\nError:\n" + e.getMessage() + "\n" + response.getText());
    }

    /**
	 * Parses the specified response and returns the resulting object. If the
	 * response is invalid, throws a {@link ParseException}. The default
	 * implementation returns uses the parser passed to the constructor to
	 * parse the response. If the parser is <code>null</code>, returns the
	 * document itself.
	 */
    @SuppressWarnings("unchecked")
    protected T parseResponse(Document document) throws ParseException {
        return c == null ? (T) document : c.parse(document);
    }

    /**
	 * Invoked when a successful XML response arrives. The default
	 * implementation attempts to parse it and invoke {@link #onResult(Object)}.
	 */
    protected void onSuccessfulXmlResponse(Document document) {
        try {
            onResult(parseResponse(document));
        } catch (ParseException e) {
            onBadResponse(document, e);
        }
    }

    /**
	 * Invoked when an error response arrives. The default implementation
	 * displays the text contents of the document's main tag as an error. 
	 */
    protected void onErrorXmlResponse(Document document) {
        Node e = document.getDocumentElement().getFirstChild();
        String f;
        if (e instanceof Text) f = ((Text) e).getData(); else f = "Received invalid error response";
        showErrorMessage(f);
    }

    /**
	 * Invoked when an unknown response arrives (a document whose main tag name
	 * is not recognized). The default implementation shows an error message.
	 */
    protected void onUnrecognizedResponse(Document document) {
        showErrorMessage("Unrecognized response (" + document.getDocumentElement().getTagName() + ") received");
    }

    /**
	 * Invoked when the received response cannot be parsed by
	 * {@link #parseResponse(Document)}. The default implementation shows an
	 * error message.
	 * 
	 * @param document The response document.
	 * @param e The exception thrown when parsing the response.
	 */
    protected void onBadResponse(Document document, ParseException e) {
        showErrorMessage("Bad response received:\n" + e.getMessage());
    }

    /**
	 * Invoked whenever any kind of response arrives, before any of the other
	 * methods. The default implementation invokes
	 * {@link ResultCallback#requestFinished()} on the callback, if any.
	 */
    protected void onAnyResponse() {
        if (d != null) d.requestFinished();
    }

    /**
	 * Invoked when a response successfully parsed by
	 * {@link #parseResponse(Document)} arrives. The default implementation
	 * invokes {@link ResultCallback#resultReceived(Object)} on the callback, if
	 * any.
	 * 
	 * @param result The object returned by {@link #parseResponse(Document)}.
	 */
    protected void onResult(T result) {
        if (d != null) d.resultReceived(result);
    }

    /**
	 * Displays the specified error message. The default implementation shows it
	 * in an alert window.
	 */
    protected void showErrorMessage(String errorMessage) {
        Window.alert("From " + getRequestUrl() + "\n" + errorMessage);
    }
}
