package com.maryanovsky.gwtutils.client.text;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import com.google.gwt.xml.client.Element;
import com.google.gwt.xml.client.Node;
import com.google.gwt.xml.client.NodeList;
import com.google.gwt.xml.client.Text;

/**
 * A utility class providing useful methods for parsing.
 *
 * @author Maryanovsky Alexander
 */
public class ParseUtils {

    /**
	 * Parses a simplified properties file.
	 */
    public static Map<String, String> parseProperties(String propertiesText) throws ParseException {
        return parseProperties(propertiesText, false);
    }

    /**
	 * Parses a simplified properties file, optionally trimming the property
	 * names and values. The parser supports comments (lines whose first
	 * non-whitespace character is <code>'#'</code>) and allows empty lines.
	 * Otherwise, every line must has a <code>'='</code> character, the first of
	 * which separates the property name and the value. A logical line may be
	 * continued on a new physical line by ending the physical line with a
	 * backslash character.
	 */
    public static Map<String, String> parseProperties(String propertiesText, boolean trim) throws ParseException {
        Map<String, String> a = new HashMap<String, String>();
        String[] b = propertiesText.split("\n");
        for (int c = 0; c < b.length; c++) {
            String d = b[c];
            if (d.trim().length() == 0) continue; else if (d.trim().charAt(0) == '#') continue;
            while (d.endsWith("\\")) {
                d = d.substring(0, d.length() - 1) + b[c + 1];
                c++;
            }
            int e = d.indexOf('=');
            if (e == -1) throw new ParseException("Missing '=' on line " + c + ": " + d);
            String f = d.substring(0, e);
            String g = d.substring(e + 1);
            if (trim) {
                f = f.trim();
                g = g.trim();
            }
            a.put(f, g);
        }
        return a;
    }

    /**
	 * If the specified element's tag name equals to the specified value,
	 * returns quietly. Otherwise throws an appropriate ParseException.
	 */
    public static void checkTagName(Element element, String tagName) throws ParseException {
        if (!tagName.equals(element.getTagName())) throw new ParseException("Element tag name is \"" + element.getTagName() + "\";" + " must be \"" + tagName + "\"");
    }

    /**
	 * Returns the list of (immediate) child elements with the specified tag
	 * name.
	 */
    public static List<Element> getNamedChildElements(Element parentElement, String tagName) {
        NodeList a = parentElement.getChildNodes();
        int b = a.getLength();
        List<Element> c = new ArrayList<Element>(b);
        for (int d = 0; d < b; d++) {
            Node e = a.item(d);
            if (e instanceof Element) {
                Element f = (Element) e;
                if (tagName.equals(f.getTagName())) c.add(f);
            }
        }
        return c;
    }

    /**
	 * Returns the child element with the specified tag name. If there is more
	 * than one such child element, throws a {@link ParseException}. If the
	 * element is missing, then depending on the value of
	 * <code>isMandatory</code> either a {@link ParseException} is thrown or
	 * <code>null</code> is returned.
	 */
    public static Element getNamedChildElement(Element parentElement, String tagName, boolean isMandatory) throws ParseException {
        Element a = null;
        for (Node b = parentElement.getFirstChild(); b != null; b = b.getNextSibling()) {
            if (b instanceof Element) {
                Element c = (Element) b;
                if (tagName.equals(c.getTagName())) {
                    if (a != null) throw new ParseException("There may be at most one " + tagName + " element within element " + c.getTagName()); else a = c;
                }
            }
        }
        if (a == null) {
            if (isMandatory) throw new ParseException(tagName + " element is missing within element " + parentElement.getTagName()); else return null;
        }
        return a;
    }

    /**
	 * Equivalent to <code>getNamedChildElement(element, tagName, true)</code>.
	 */
    public static Element getMandatoryNamedElement(Element parentElement, String tagName) throws ParseException {
        return getNamedChildElement(parentElement, tagName, true);
    }

    /**
	 * Equivalent to <code>getNamedChildElement(element, tagName, false)</code>.
	 */
    public static Element getOptionalNamedElement(Element parentElement, String tagName) throws ParseException {
        return getNamedChildElement(parentElement, tagName, false);
    }

    /**
	 * Returns the text content of the specified element by concatenating the
	 * data of all its {@link Text} child nodes.
	 */
    public static String getTextContent(Element element) throws ParseException {
        String a = "";
        for (Node b = element.getFirstChild(); b != null; b = b.getNextSibling()) if (b instanceof Text) a += ((Text) b).getData();
        return a;
    }

    /**
	 * Returns the value of the specified element's attribute. If the attribute
	 * is missing, then depending on the value of <code>isMandatory</code>
	 * either a {@link ParseException} is thrown or <code>null</code> is
	 * returned.
	 */
    public static String getAttribute(Element element, String attributeName, boolean isMandatory) throws ParseException {
        String a = element.getAttribute(attributeName);
        if (isMandatory && (a == null)) throw new ParseException("Missing \"" + attributeName + "\" attribute in element \"" + element.getTagName() + "\"");
        return a;
    }

    /**
	 * Equivalent to <code>getAttribute(element, attributeName, true)</code>.
	 */
    public static String getMandatoryAttribute(Element element, String attributeName) throws ParseException {
        return getAttribute(element, attributeName, true);
    }

    /**
	 * Returns the value of the specified element's attribute. If the attribute
	 * is missing, returns the specified default value.
	 */
    public static String getOptionalAttribute(Element element, String attributeName, String defaultValue) {
        try {
            String a = getAttribute(element, attributeName, false);
            return (a == null) ? defaultValue : a;
        } catch (ParseException a) {
            throw new IllegalStateException("This may never happen");
        }
    }

    /***/
    private static double parseDouble(String a, Element b, String c) throws ParseException {
        try {
            return Double.parseDouble(a);
        } catch (NumberFormatException d) {
            throw new ParseException("Unable to parse \"" + c + "\" attribute in element \"" + b.getTagName() + "\" as a double (value: " + a + ")");
        }
    }

    /**
	 * Returns the value of the specified element's attribute, parsed as a
	 * double.
	 * Throws a {@link ParseException} if the attribute is missing or if its
	 * value can't be parsed as a double.
	 */
    public static double getMandatoryDoubleAttribute(Element element, String attributeName) throws ParseException {
        return parseDouble(getMandatoryAttribute(element, attributeName), element, attributeName);
    }

    /**
	 * Returns the value of the specified element's attribute, parsed as a
	 * double. If the attribute is missing, returns the specified default value.
	 * If the attribute's value can't be parsed as a double, throws a
	 * {@link ParseException}.
	 */
    public static double getOptionalDoubleAttribute(Element element, String attributeName, double defaultValue) throws ParseException {
        String a = getOptionalAttribute(element, attributeName, null);
        return (a == null) ? defaultValue : parseDouble(a, element, attributeName);
    }

    /***/
    private static int parseInt(String a, Element b, String c) throws ParseException {
        try {
            return Integer.parseInt(a);
        } catch (NumberFormatException d) {
            throw new ParseException("Unable to parse \"" + c + "\" attribute in element \"" + b.getTagName() + "\" as an int (value: " + a + ")");
        }
    }

    /**
	 * Returns the value of the specified element's attribute, parsed as an int.
	 * Throws a {@link ParseException} if the attribute is missing or if its
	 * value can't be parsed as an int.
	 */
    public static int getMandatoryIntAttribute(Element element, String attributeName) throws ParseException {
        return parseInt(getMandatoryAttribute(element, attributeName), element, attributeName);
    }

    /**
	 * Returns the value of the specified element's attribute, parsed as an int.
	 * If the attribute is missing, returns the specified default value. If the
	 * attribute's value can't be parsed as an int, throws a
	 * {@link ParseException}.
	 */
    public static int getOptionalIntAttribute(Element element, String attributeName, int defaultValue) throws ParseException {
        String a = getOptionalAttribute(element, attributeName, null);
        return (a == null) ? defaultValue : parseInt(a, element, attributeName);
    }

    /**
	 * Returns the content of the child element of the specified element with
	 * the specified name. Throws a {@link ParseException} if the element is
	 * missing.
	 */
    public static String getMandatoryElementTextContent(Element element, String childElementName) throws ParseException {
        return getTextContent(getMandatoryNamedElement(element, childElementName));
    }

    /**
	 * Returns the content of the child element of the specified element with
	 * the specified name. Returns the specified default value if the element is
	 * missing.
	 */
    public static String getOptionalElementTextContent(Element element, String childElementName, String defaultValue) throws ParseException {
        Element a = getOptionalNamedElement(element, childElementName);
        return (a == null) ? defaultValue : getTextContent(a);
    }

    /**
	 * Returns a list of the text contents of the child elements of the
	 * specified element with the specified name.
	 */
    public static String[] getElementListTextContent(Element element, String childElementsName) throws ParseException {
        List<Element> a = getNamedChildElements(element, childElementsName);
        String[] b = new String[a.size()];
        for (int c = 0; c < b.length; c++) b[c] = getTextContent(a.get(c));
        return b;
    }
}
