package com.maryanovsky.gwtutils.client.text;

/**
 * An interface for classes which can both decode and encode objects from/to
 * text.
 * 
 * @author Maryanovsky Alexander
 */
public interface Transcoder<T> extends Encoder<T>, Decoder<T> {

    /**
	 * A string transcoder.
	 */
    public static final Transcoder<String> STRING_TRANSCODER = new Transcoder<String>() {

        public String encode(String value) {
            return value;
        }

        public String decode(String s) {
            return s;
        }
    };

    /**
	 * A boolean transcoder. The decode method accepts only the values "true"
	 * and "false" (ignoring case).
	 */
    public static final Transcoder<Boolean> BOOLEAN_TRANSCODER = new Transcoder<Boolean>() {

        public Boolean decode(String s) throws ParseException {
            if ("true".equalsIgnoreCase(s)) return true; else if ("false".equalsIgnoreCase(s)) return false; else throw new ParseException("Value cannot be parsed as boolean: " + s);
        }

        public String encode(Boolean value) {
            return value.toString();
        }
    };

    /**
	 * An int transcoder (to/from decimal representation).
	 */
    public static final Transcoder<Integer> INTEGER_TRANSCODER = new Transcoder<Integer>() {

        public Integer decode(String s) throws ParseException {
            try {
                return Integer.valueOf(s);
            } catch (NumberFormatException a) {
                throw new ParseException("Value cannot be parsed as an integer: " + s);
            }
        }

        public String encode(Integer i) {
            return i.toString();
        }
    };

    /**
	 * A double transcoder.
	 */
    public static final Transcoder<Double> DOUBLE_TRANSCODER = new Transcoder<Double>() {

        public Double decode(String s) throws ParseException {
            try {
                return Double.valueOf(s);
            } catch (NumberFormatException a) {
                throw new ParseException("Value cannot be parsed as a double: " + s);
            }
        }

        public String encode(Double d) {
            return d.toString();
        }
    };

    /**
	 * A char transcoder.
	 */
    public static final Transcoder<Character> CHARACTER_TRANSCODER = new Transcoder<Character>() {

        public Character decode(String s) throws ParseException {
            if (s.length() != 1) throw new ParseException("Value is not a single a character: " + s); else return s.charAt(0);
        }

        public String encode(Character c) {
            return c.toString();
        }
    };
}
