package com.maryanovsky.gwtutils.client.ui;

import java.util.LinkedList;
import java.util.List;
import com.google.gwt.dom.client.Style.Unit;
import com.google.gwt.event.dom.client.ClickEvent;
import com.google.gwt.event.dom.client.ClickHandler;
import com.google.gwt.event.shared.HandlerRegistration;
import com.google.gwt.user.client.ui.*;

/**
 * An "Accordion" panel. It arranges its children vertically with each child
 * having a "header" that can be clicked to show/hide the widget. This is
 * similar to {@link StackLayoutPanel}, only more than one widget can be visible
 * at a time.
 * 
 * @author Maryanovsky Alexander
 */
public class Accordion extends ResizeComposite implements RequiresResize, ProvidesResize {

    /***/
    private static class Header extends Composite {

        /**
		 * Creates a new <code>Header</code> with the specified widget.
		 */
        public Header(Widget widget) {
            if (widget == null) throw new IllegalArgumentException("widget may not be null");
            initWidget(widget);
        }

        /**
		 * Returns the widget we're wrapping.
		 */
        public Widget getWidget() {
            return super.getWidget();
        }

        /**
		 * Adds a click handler to the header.
		 */
        public HandlerRegistration addClickHandler(ClickHandler handler) {
            return this.addDomHandler(handler, ClickEvent.getType());
        }
    }

    /***/
    private static class LayoutData {

        /***/
        private final Header h;

        /***/
        private final double i;

        /***/
        private final Widget j;

        /***/
        private final double k;

        /***/
        private boolean l = false;

        /**
		 * Creates a new <code>LayoutData</code> object with the specified
		 * header, widget, and their heights.
		 */
        public LayoutData(Header header, double headerHeight, Widget widget, double widgetHeight) {
            if (header == null) throw new IllegalArgumentException("header may not be null");
            if (widget == null) throw new IllegalArgumentException("widget may not be null");
            this.h = header;
            this.i = headerHeight;
            this.j = widget;
            this.k = widgetHeight;
        }

        /**
		 * Returns the header.
		 */
        public Header getHeader() {
            return h;
        }

        /**
		 * Returns the height of the header.
		 */
        public double getHeaderHeight() {
            return i;
        }

        /**
		 * Returns the widget.
		 */
        public Widget getWidget() {
            return j;
        }

        /**
		 * Returns the height of the widget.
		 */
        public double getWidgetHeight() {
            return k;
        }

        /**
		 * Returns whether the widget is visible.
		 */
        public boolean isWidgetVisible() {
            return l;
        }

        /**
		 * Sets whether the widget is visible.
		 */
        public void setWidgetVisible(boolean isWidgetVisible) {
            this.l = isWidgetVisible;
        }
    }

    /***/
    private final int a = 250;

    /***/
    private final Unit b;

    /***/
    private final LayoutPanel c = new LayoutPanel();

    /***/
    private final List<LayoutData> d = new LinkedList<LayoutData>();

    /***/
    private double e = 0;

    /**
	 * Creates a new <code>Accordion</code> which will have children whose sizes
	 * will be specified in the given unit.
	 */
    public Accordion(Unit unit) {
        if (unit == null) throw new IllegalArgumentException("unit may not be null");
        this.b = unit;
        initWidget(c);
    }

    /**
	 * Returns the number of widgets in the Accordion.
	 */
    public int getWidgetCount() {
        return d.size();
    }

    /**
	 * Returns the index of the specified widget; -1 if the widget is not a
	 * child.
	 */
    public int getWidgetIndex(Widget widget) {
        for (int h = 0; h < d.size(); h++) if (d.get(h).getWidget() == widget) return h;
        return -1;
    }

    /**
	 * Returns the header for the widget at the specified index.
	 */
    public Widget getHeader(int index) {
        return d.get(index).getHeader().getWidget();
    }

    /**
	 * Returns the header for the specified widget.
	 */
    public Widget getHeader(Widget widget) {
        return getHeader(getWidgetIndex(widget));
    }

    /**
	 * Returns whether the widget at the specified index is visible.
	 */
    public boolean isWidgetVisible(int index) {
        return d.get(index).isWidgetVisible();
    }

    /**
	 * Returns whether the specified widget is visible.
	 */
    public boolean isWidgetVisible(Widget widget) {
        return isWidgetVisible(getWidgetIndex(widget));
    }

    /**
	 * Sets the visibility of the widget at the specified index.
	 */
    public void setWidgetVisible(int index, boolean isVisible) {
        d.get(index).setWidgetVisible(isVisible);
        f(a);
    }

    /**
	 * Sets the visibility of the specified widget.
	 */
    public void setWidgetVisible(Widget widget, boolean isVisible) {
        setWidgetVisible(getWidgetIndex(widget), isVisible);
    }

    /**
	 * Adds a widget with the specified header at the bottom. The heights of the
	 * widget and header must be specified. 
	 */
    public void add(Widget widget, double widgetHeight, Widget header, double headerHeight) {
        insert(widget, widgetHeight, header, headerHeight, getWidgetCount());
    }

    /**
	 * Adds a widget with the specified header at the specified index. Also
	 * specified are their heights and the location where to insert them.  
	 */
    public void insert(Widget widget, double widgetHeight, Widget header, double headerHeight, int index) {
        g(widget, widgetHeight, new Header(header), headerHeight, index);
    }

    /**
	 * Returns the spacing between the widgets.
	 */
    public double getSpacing() {
        return e;
    }

    /**
	 * Sets the spacing between the widgets.
	 */
    public void setSpacing(double spacing) {
        if (spacing < 0) throw new IllegalArgumentException("spacing may not be negative");
        this.e = spacing;
    }

    /**
	 * Lays the children out when the widget is attached.
	 */
    @Override
    protected void onLoad() {
        f(0);
    }

    /***/
    private void f(int h) {
        double i = 0;
        for (LayoutData j : d) {
            c.setWidgetTopHeight(j.getHeader(), i, b, j.getHeaderHeight(), b);
            i += j.getHeaderHeight();
            if (j.isWidgetVisible()) {
                c.setWidgetTopHeight(j.getWidget(), i, b, j.getWidgetHeight(), b);
                i += j.getWidgetHeight();
            } else {
                c.setWidgetTopHeight(j.getWidget(), i, b, 0, b);
            }
            i += getSpacing();
        }
        c.animate(h);
    }

    /***/
    private void g(final Widget h, double i, Header j, double k, int l) {
        if ((l < 0) || (l > getWidgetCount())) throw new IllegalArgumentException("index (" + l + ") out of bounds");
        c.add(j);
        c.add(h);
        c.setWidgetLeftRight(j, 0, Unit.PX, 0, Unit.PX);
        c.setWidgetLeftRight(h, 0, Unit.PX, 0, Unit.PX);
        LayoutData m = new LayoutData(j, k, h, i);
        d.add(l, m);
        j.addClickHandler(new ClickHandler() {

            public void onClick(ClickEvent event) {
                int n = getWidgetIndex(h);
                setWidgetVisible(n, !isWidgetVisible(n));
            }
        });
        if (isAttached()) f(0);
    }
}
