package com.maryanovsky.gwtutils.client.ui;

import com.google.gwt.core.client.Scheduler;
import com.google.gwt.core.client.Scheduler.ScheduledCommand;
import com.google.gwt.user.client.Timer;

/**
 * A convenient method to perform animations.
 * 
 * @author Maryanovsky Alexander
 */
public abstract class Animation {

    /**
	 * The default frame rate.
	 */
    public static final int DEFAULT_FPS = 25;

    /***/
    private final int a;

    /***/
    private final double b;

    /***/
    private long c = -1;

    /***/
    private boolean d = false;

    /***/
    private final Timer e = new Timer() {

        public void run() {
            if (!isRunning()) return;
            long f = System.currentTimeMillis() - c;
            if (f >= a) end(); else nextFrame();
        }
    };

    /**
	 * Creates a new <code>Animation</code> with the specified duration (in
	 * milliseconds) and frames per second to display.
	 */
    public Animation(int duration, int fps) {
        if (duration < 0) throw new IllegalArgumentException("duration may not be negative");
        if (fps <= 0) throw new IllegalArgumentException("fps must be positive");
        this.a = duration;
        this.b = fps;
    }

    /**
	 * Creates a new <code>Animation</code> with the specified duration and
	 * default frame rate.
	 */
    public Animation(int duration) {
        this(duration, DEFAULT_FPS);
    }

    /**
	 * Returns the duration of the animation, in milliseconds.
	 */
    public final int getDuration() {
        return a;
    }

    /**
	 * Returns the number of frames per second this animation displays.
	 */
    public final double getFps() {
        return b;
    }

    /**
	 * Starts the animation (and returns immediately).
	 */
    public void start() {
        if (d) throw new IllegalStateException("Animation already running");
        d = true;
        Scheduler.get().scheduleDeferred(new ScheduledCommand() {

            public void execute() {
                if (!d) return;
                c = System.currentTimeMillis();
                animationStart();
                nextFrame();
                e.scheduleRepeating((int) (1000 / b));
            }
        });
    }

    /**
	 * Returns whether the animation is currently running. This returns
	 * <code>true</code> between the return from {@linkplain #start()} and until
	 * the return from {@linkplain #animationEnd()}.
	 */
    public final boolean isRunning() {
        return d;
    }

    /**
	 * Ends the animation immediately.
	 */
    public void end() {
        if (!d) throw new IllegalArgumentException("Animation already not running");
        if (c == -1) animationStart();
        animationEnd();
        e.cancel();
        d = false;
        c = -1;
    }

    /**
	 * Returns the linear progress of the animation, i.e. a value that changes
	 * from 0 to 1 linearly with time.
	 */
    public final double getLinearProgress() {
        return (System.currentTimeMillis() - c) / (double) a;
    }

    /**
	 * Returns the sinusoidal progression of the animation, i.e. a value that
	 * changes from 0 to 1 as the sine changes between -pi/2 and +pi/2.
	 */
    public final double getSinusoidalProgress() {
        return (Math.sin((getLinearProgress() - 0.5) * Math.PI) + 1) / 2;
    }

    /**
	 * Returns the half-sinusoidal progression of the animation, i.e. a value
	 * that changes from 0 to 1 as the sine changes between 0 and pi/2.
	 */
    public final double getHalfSinusoidalProgress() {
        return Math.sin(getLinearProgress() * Math.PI / 2);
    }

    /**
	 * Invoked before the first frame of the animation is displayed. This method
	 * will always be called, even if the animation ends prematurely.
	 */
    protected abstract void animationStart();

    /**
	 * Displays the next frame of the animation. Use one of the
	 * <code>get[Type]Progress</code> methods to determine how far along the
	 * animation is. Note that this method is not guaranteed to get called
	 * because the animation may be ended prematurely.
	 */
    protected abstract void nextFrame();

    /**
	 * Invoked after the animation is done. This method will always be
	 * called, even if the animation ends prematurely.
	 */
    protected abstract void animationEnd();
}
