package com.maryanovsky.gwtutils.client.ui;

import com.google.gwt.user.client.Event;
import com.google.gwt.user.client.ui.SimplePanel;
import com.google.gwt.user.client.ui.Widget;

/**
 * A panel which allows cancelling the bubbling of and preventing the default
 * action from occurring on a specified set of events it receives.
 * 
 * @author Maryanovsky Alexander
 */
public class EventCancellingPanel extends SimplePanel {

    /***/
    private int a = 0;

    /***/
    private int b = 0;

    /**
	 * Creates a new {@link EventCancellingPanel} with the specified child
	 * widget and which will cancel the bubbling of events matching the first
	 * specified mask and prevent the default action for events matching the
	 * second specified mask.
	 */
    public EventCancellingPanel(Widget child, int cancelBubbleEventsMask, int preventDefaultEventsMask) {
        setCancelBubbleEventsMask(cancelBubbleEventsMask);
        setPreventDefaultEventsMask(preventDefaultEventsMask);
        if (child != null) setWidget(child);
    }

    /**
	 * Creates a new {@link EventCancellingPanel}.
	 */
    public EventCancellingPanel() {
        this(null, 0, 0);
    }

    /**
	 * Sets the bitmask of events whose bubbling to cancel.
	 */
    public void setCancelBubbleEventsMask(int cancelBubbleEventsMask) {
        this.a = cancelBubbleEventsMask;
        unsinkEvents(0xffffffff);
        sinkEvents(cancelBubbleEventsMask | b);
    }

    /**
	 * Returns the bitmask of events whose bubbling we are currently cancelling.
	 */
    public int getCancelBubbleEventsMask() {
        return a;
    }

    /**
	 * Sets the bitmask of events for which we prevent the default action.
	 */
    public void setPreventDefaultEventsMask(int preventDefaultEventsMask) {
        this.b = preventDefaultEventsMask;
        unsinkEvents(0xffffffff);
        sinkEvents(a | preventDefaultEventsMask);
    }

    /**
	 * Returns the bitmask of events for which we prevent the default action.
	 */
    public int getPreventDefaultEventsMask() {
        return b;
    }

    /**
	 * Returns whether we should cancel the bubbling of the the specified event.
	 * By default, this method uses the bitmask to determine whether to cancel
	 * bubbling, but you may override it for a more fine-grained control.
	 */
    protected boolean isCancelBubbling(Event event) {
        return (a & event.getTypeInt()) != 0;
    }

    /**
	 * Returns whether we should prevent the default action for the specified
	 * event. By default, this method uses the bitmask to determine whether to
	 * prevent the default action, but you may override it for a more
	 * fine-grained control.
	 */
    protected boolean isPreventDefault(Event event) {
        return (b & event.getTypeInt()) != 0;
    }

    /**
	 * Cancels the bubbling of and/or prevents the default action of the
	 * specified event.
	 */
    @Override
    public void onBrowserEvent(Event event) {
        if (isCancelBubbling(event)) event.stopPropagation();
        if (isPreventDefault(event)) event.preventDefault();
        super.onBrowserEvent(event);
    }
}
