package com.maryanovsky.gwtutils.client.ui;

import java.util.EnumMap;
import java.util.Map;
import com.google.gwt.user.client.DOM;
import com.google.gwt.user.client.Element;
import com.google.gwt.user.client.ui.*;

/**
 * Defines various usefil UI related values, such as preferred spacing between
 * elements, difference in element sizes which allows aligning them etc.
 * Currently, the implementation is tailored to Internet Explorer. To better
 * support other browsers, it should detect the browser we're running in and
 * use an appropriate delegate. Also, the implementation is not complete.
 *
 * @author Maryanovsky Alexander
 */
public class GwtUiDefaults {

    /**
	 * The enumeration of types of widgets for which we offer special treatment.
	 */
    public static enum WidgetElementType {

        TEXT_BOX, TEXT_AREA, LIST_BOX, BUTTON
    }

    ;

    /***/
    private static final Map<WidgetElementType, Integer> a = new EnumMap<WidgetElementType, Integer>(WidgetElementType.class);

    /**
	 * Initializes <code>HORIZONTAL_VISUAL_MARGIN</code>.
	 */
    static {
        a.put(WidgetElementType.TEXT_BOX, 3);
        a.put(WidgetElementType.TEXT_AREA, 3);
        a.put(WidgetElementType.LIST_BOX, 0);
        a.put(WidgetElementType.BUTTON, 1);
    }

    /**
	 * Makes a "dialog" panel from the specified widget. 
	 */
    public static void makeDialogPanel(Widget widget) {
        Element b = widget.getElement();
        DOM.setStyleAttribute(b, "backgroundColor", "#e0e0e0");
        DOM.setStyleAttribute(b, "borderColor", "darkGray");
        DOM.setStyleAttribute(b, "borderStyle", "outset");
        DOM.setStyleAttribute(b, "borderWidth", "2px");
        DOM.setStyleAttribute(b, "padding", "12px");
    }

    /**
	 * Makes a "pallette" panel from the specified widget.
	 */
    public static void makePallettePanel(Widget widget) {
        Element b = widget.getElement();
        DOM.setStyleAttribute(b, "backgroundColor", "#e0e0e0");
        DOM.setStyleAttribute(b, "borderColor", "darkGray");
        DOM.setStyleAttribute(b, "borderStyle", "outset");
        DOM.setStyleAttribute(b, "borderWidth", "2px");
        DOM.setStyleAttribute(b, "padding", "0px 5px");
        DOM.setStyleAttribute(b, "cursor", "default");
    }

    /**
	 * UI elements have a visual edge at which they should be aligned. This
	 * returns the horizontal margin between their visual edge and their actual
	 * boundary.
	 */
    public static int getHorizontalVisualMargin(WidgetElementType elementType) {
        return a.get(elementType);
    }

    /**
	 * Similar to {@link #getHorizontalVisualMargin(WidgetElementType)}, but the
	 * element type is guessed from the widget using
	 * {@link #getElementType(Widget)}. Due to this, the method, only works for
	 * the standard GWT widgets, while the version that takes an element type
	 * can work for custom widgets as well.
	 */
    public static int getHorizontalVisualMargin(Widget widget) {
        return getHorizontalVisualMargin(getElementType(widget));
    }

    /**
	 * Returns the element type of the specified standard GWT widget,
	 * if supported. Throws an exception if the widget is not a standard GWT
	 * widget or is an unsupported widget.
	 */
    public static WidgetElementType getElementType(Widget widget) {
        if (widget instanceof TextBox) return WidgetElementType.TEXT_BOX; else if (widget instanceof TextArea) return WidgetElementType.TEXT_AREA; else if (widget instanceof ListBox) return WidgetElementType.LIST_BOX; else if (widget instanceof Button) return WidgetElementType.BUTTON; else throw new IllegalArgumentException("Unsupported widget: " + widget);
    }

    /**
	 * Returns the preferred horizontal distance between related elements.
	 */
    public static int getRelatedHorizontalDistance() {
        return 6;
    }

    /**
	 * Returns the preferred horizontal distance between unrelated elements.
	 */
    public static int getUnrelatedHorizontalDistance() {
        return 10;
    }

    /**
	 * Returns the preferred vertical distance between related elements.
	 */
    public static int getRelatedVerticalDistance() {
        return 4;
    }

    /**
	 * Returns the preferred vertical distance between unrelated elements.
	 */
    public static int getUnrelatedVerticalDistance() {
        return 10;
    }

    /**
	 * Returns the gap between an icon image and its label.
	 */
    public static int getIconLabelGap() {
        return 2;
    }

    /**
	 * Returns the indent size. This is used when related UI is placed on the
	 * next line. This is the number of pixels by which it is indented.  
	 */
    public static int getIndentSize() {
        return 10;
    }

    /**
	 * Returns the vertical distance between a form's title and its fields.
	 */
    public static int getFormTitleFieldsGap() {
        return 16;
    }

    /**
	 * Returns the vertical distance between a form's fields and its buttons.
	 */
    public static int getFormFieldsButtonsGap() {
        return 16;
    }
}
