package com.maryanovsky.gwtutils.client.ui;

import static com.google.gwt.user.client.ui.HasHorizontalAlignment.ALIGN_LEFT;
import static com.google.gwt.user.client.ui.HasHorizontalAlignment.ALIGN_RIGHT;
import static com.google.gwt.user.client.ui.HasVerticalAlignment.ALIGN_BOTTOM;
import static com.google.gwt.user.client.ui.HasVerticalAlignment.ALIGN_TOP;
import java.util.*;
import com.google.gwt.core.client.GWT;
import com.google.gwt.dom.client.*;
import com.google.gwt.dom.client.Style.Position;
import com.google.gwt.dom.client.Style.Unit;
import com.google.gwt.event.dom.client.*;
import com.google.gwt.event.logical.shared.HasResizeHandlers;
import com.google.gwt.event.logical.shared.ResizeEvent;
import com.google.gwt.event.logical.shared.ResizeHandler;
import com.google.gwt.event.shared.EventHandler;
import com.google.gwt.event.shared.GwtEvent;
import com.google.gwt.event.shared.HandlerRegistration;
import com.google.gwt.http.client.URL;
import com.google.gwt.user.client.DOM;
import com.google.gwt.user.client.Element;
import com.google.gwt.user.client.Event;
import com.google.gwt.user.client.Window;
import com.google.gwt.user.client.ui.*;
import com.google.gwt.user.client.ui.HTMLTable.CellFormatter;
import com.google.gwt.user.client.ui.HasHorizontalAlignment.HorizontalAlignmentConstant;
import com.google.gwt.user.client.ui.HasVerticalAlignment.VerticalAlignmentConstant;
import com.maryanovsky.gwtutils.client.BrowserUtils;
import com.maryanovsky.gwtutils.client.event.Callback;
import com.maryanovsky.gwtutils.client.event.MultipleHandlerRegistration;
import com.maryanovsky.gwtutils.client.geom.*;
import com.maryanovsky.gwtutils.client.ui.mouse.gestures.drag.*;
import com.maryanovsky.gwtutils.client.ui.mouse.gestures.drag.DragEndEvent;
import com.maryanovsky.gwtutils.client.ui.mouse.gestures.drag.DragStartEvent;

/**
 * A utility class for UI related functionality.
 *
 * @author Maryanovsky Alexander
 */
public class UiUtils {

    /**
	 * The delegate for left-to-right languages.
	 */
    public static final Delegate LTR = new LtrDelegate();

    /**
	 * The delegate for right-to-left languages.
	 */
    public static final Delegate RTL = new RtlDelegate();

    /***/
    private static Delegate a = LTR;

    /**
	 * Sets the current delegate. Possible values are {@link #LTR} and
	 * {@link #RTL}. 
	 */
    public static void setDelegate(Delegate delegate) {
        if (delegate == null) throw new IllegalArgumentException("delegate may not be null");
        UiUtils.a = delegate;
    }

    /**
	 * Creates a label to be a label for some other element.
	 */
    public static Label makeLabelFor(String text) {
        Label e = new Label(text, false);
        e.setHorizontalAlignment(getTrailingAlignment());
        return e;
    }

    /**
	 * Sets the margin leading the element, in the reading direction (left when
	 * in LTR, right when in RTL mode).
	 */
    public static void setLeadingMargin(Widget widget, int size) {
        a.setLeadingMargin(widget, size);
    }

    /**
	 * Sets the margin trailing the element, in the reading direction (right
	 * when in LTR, left when in RTL mode).
	 */
    public static void setTrailingMargin(Widget widget, int size) {
        a.setTrailingMargin(widget, size);
    }

    /**
	 * Puts the specified widgets into the specified row in the grid, in
	 * language order.
	 */
    public static void setGridRow(Grid grid, int row, List<? extends Widget> widgets) {
        setGridRow(grid, row, widgets, Collections.nCopies(widgets.size(), (VerticalAlignmentConstant) null), Collections.nCopies(widgets.size(), (HorizontalAlignmentConstant) null));
    }

    /**
	 * Puts the specified widgets into the specified row in the grid, in
	 * language order. Additionally, the corresponding cells are set to the
	 * specified alignments (if the alignment values are
	 * non-<code>null</code>).
	 */
    public static void setGridRow(Grid grid, int row, List<? extends Widget> widgets, List<VerticalAlignmentConstant> vAlignments, List<HorizontalAlignmentConstant> hAlignments) {
        if (grid.getColumnCount() < widgets.size()) grid.resizeColumns(widgets.size());
        a.setGridRow(grid, row, widgets, vAlignments, hAlignments);
    }

    /**
	 * Adds the specified widgets to the specified horizontal panel in reading
	 * order.
	 */
    public static void addWidgets(HorizontalPanel panel, List<? extends Widget> widgets) {
        a.addWidgets(panel, widgets);
    }

    /**
	 * Adds the specified dialog buttons to the specified panel. The order of
	 * the arguments should be consistent with MS Windows standards ("OK" first,
	 * "Cancel" second) when using an LTR language. The actual order in which
	 * the buttons will be placed depends on the current language and platform.
	 */
    public static void addDialogButtons(HorizontalPanel panel, List<? extends Widget> buttons) {
        if (buttons.isEmpty()) return;
        a.addDialogButtons(panel, buttons);
    }

    /**
	 * Returns the alignment constant corresponding to the beginning of the line
	 * in reading order.
	 */
    public static HorizontalAlignmentConstant getLeadingAlignment() {
        return a.getLeadingAlignment();
    }

    /**
	 * Returns the alignment constant corresponding to the end of the line in
	 * reading order.
	 */
    public static HorizontalAlignmentConstant getTrailingAlignment() {
        return a.getTrailingAlignment();
    }

    /**
	 * Sets the direction of the specified widget to the direction of the
	 * application's language.
	 */
    public static void setTextDirection(Widget widget) {
        a.setTextDirection(widget);
    }

    /**
	 * Sets the direction of the specified widget to LTR.
	 */
    public static void setDataDirection(Widget widget) {
        a.setDataDirection(widget);
    }

    /***/
    private static interface Delegate {

        /**
		 * Sets the leading margin of the widget to the specified value.
		 */
        void setLeadingMargin(Widget widget, int size);

        /**
		 * Sets the trailing margin of the widget to the specified value.
		 */
        void setTrailingMargin(Widget widget, int size);

        /**
		 * Returns the leading alignment constant.
		 */
        HorizontalAlignmentConstant getLeadingAlignment();

        /**
		 * Returns the trailing alignment constant.
		 */
        HorizontalAlignmentConstant getTrailingAlignment();

        /**
		 * Adds the specified widgets, in language order, to the specified
		 * grid in the specified row panel. Additionally, the corresponding
		 * cells are set to the specified alignments (when the alignment
		 * values are non-<code>null</code>).
		 */
        void setGridRow(Grid grid, int row, List<? extends Widget> widgets, List<VerticalAlignmentConstant> vAlignments, List<HorizontalAlignmentConstant> hAlignments);

        /**
		 * Adds the specified widgets, in reading order, to the specified
		 * horizontal panel.
		 */
        void addWidgets(HorizontalPanel panel, List<? extends Widget> widgets);

        /**
		 * Adds the specified dialog buttons to the horizontal panel, in the
		 * correct order. The order in which the buttons are passed is the
		 * Microsoft Windows order ("OK" then "Cancel").
		 */
        void addDialogButtons(HorizontalPanel panel, List<? extends Widget> buttons);

        /**
		 * Sets the direction of the specified widget to the direction of the
		 * application's language.
		 */
        void setTextDirection(Widget widget);

        /**
		 * Sets the direction of the specified widget to LTR.
		 */
        void setDataDirection(Widget widget);
    }

    /***/
    private static class LtrDelegate implements Delegate {

        /**
		 * {@inheritDoc}
		 */
        public void setLeadingMargin(Widget widget, int size) {
            DOM.setStyleAttribute(widget.getElement(), "marginLeft", size + "px");
        }

        /**
		 * {@inheritDoc}
		 */
        public void setTrailingMargin(Widget widget, int size) {
            DOM.setStyleAttribute(widget.getElement(), "marginRight", size + "px");
        }

        /**
		 * {@inheritDoc}
		 */
        public HorizontalAlignmentConstant getLeadingAlignment() {
            return ALIGN_LEFT;
        }

        /**
		 * {@inheritDoc}
		 */
        public HorizontalAlignmentConstant getTrailingAlignment() {
            return ALIGN_RIGHT;
        }

        /**
		 * {@inheritDoc}
		 */
        public void setGridRow(Grid grid, int row, List<? extends Widget> widgets, List<VerticalAlignmentConstant> vAlignments, List<HorizontalAlignmentConstant> hAlignments) {
            CellFormatter e = grid.getCellFormatter();
            int f = widgets.size();
            for (int g = 0; g < f; g++) {
                int h = g;
                grid.setWidget(row, h, widgets.get(g));
                VerticalAlignmentConstant i = vAlignments.get(g);
                HorizontalAlignmentConstant j = hAlignments.get(g);
                if (i != null) e.setVerticalAlignment(row, h, i);
                if (j != null) e.setHorizontalAlignment(row, h, j);
            }
        }

        /**
		 * {@inheritDoc}
		 */
        public void addWidgets(HorizontalPanel panel, List<? extends Widget> widgets) {
            for (Widget e : widgets) panel.add(e);
        }

        /**
		 * {@inheritDoc}
		 */
        public void addDialogButtons(HorizontalPanel panel, List<? extends Widget> buttons) {
            Iterator<? extends Widget> e = buttons.iterator();
            panel.add(e.next());
            while (e.hasNext()) {
                panel.add(getRelatedHorizontalStrut());
                panel.add(e.next());
            }
        }

        /**
		 * {@inheritDoc}
		 */
        public void setTextDirection(Widget widget) {
            DOM.setElementProperty(widget.getElement(), "dir", "ltr");
        }

        /**
		 * {@inheritDoc}
		 */
        public void setDataDirection(Widget widget) {
            DOM.setElementProperty(widget.getElement(), "dir", "ltr");
        }
    }

    /***/
    private static class RtlDelegate implements Delegate {

        /**
		 * {@inheritDoc}
		 */
        public void setLeadingMargin(Widget widget, int size) {
            DOM.setStyleAttribute(widget.getElement(), "marginRight", size + "px");
        }

        /**
		 * {@inheritDoc}
		 */
        public void setTrailingMargin(Widget widget, int size) {
            DOM.setStyleAttribute(widget.getElement(), "marginLeft", size + "px");
        }

        /**
		 * {@inheritDoc}
		 */
        public HorizontalAlignmentConstant getLeadingAlignment() {
            return ALIGN_RIGHT;
        }

        /**
		 * {@inheritDoc}
		 */
        public HorizontalAlignmentConstant getTrailingAlignment() {
            return ALIGN_LEFT;
        }

        /**
		 * {@inheritDoc}
		 */
        public void setGridRow(Grid grid, int row, List<? extends Widget> widgets, List<VerticalAlignmentConstant> vAlignments, List<HorizontalAlignmentConstant> hAlignments) {
            CellFormatter e = grid.getCellFormatter();
            int f = widgets.size();
            for (int g = 0; g < f; g++) {
                int h = f - g - 1;
                grid.setWidget(row, h, widgets.get(g));
                VerticalAlignmentConstant i = vAlignments.get(g);
                HorizontalAlignmentConstant j = hAlignments.get(g);
                if (i != null) e.setVerticalAlignment(row, h, i);
                if (j != null) e.setHorizontalAlignment(row, h, j);
            }
        }

        /**
		 * {@inheritDoc}
		 */
        public void addWidgets(HorizontalPanel panel, List<? extends Widget> widgets) {
            for (ListIterator<? extends Widget> e = widgets.listIterator(widgets.size()); e.hasPrevious(); ) panel.add(e.previous());
        }

        /**
		 * {@inheritDoc}
		 */
        public void addDialogButtons(HorizontalPanel panel, List<? extends Widget> buttons) {
            ListIterator<? extends Widget> e = buttons.listIterator(buttons.size());
            panel.add(e.previous());
            while (e.hasPrevious()) {
                panel.add(getRelatedHorizontalStrut());
                panel.add(e.previous());
            }
        }

        /**
		 * {@inheritDoc}
		 */
        public void setTextDirection(Widget widget) {
            DOM.setElementProperty(widget.getElement(), "dir", "rtl");
        }

        /**
		 * {@inheritDoc}
		 */
        public void setDataDirection(Widget widget) {
            DOM.setElementProperty(widget.getElement(), "dir", "ltr");
        }
    }

    /**
	 * Sets the width of the specified widget so that its visual width is as
	 * specified.
	 */
    public static void setVisualWidth(Widget widget, GwtUiDefaults.WidgetElementType elementType, int width) {
        int e = width - 2 * GwtUiDefaults.getHorizontalVisualMargin(elementType);
        widget.setWidth(e + "px");
    }

    /**
	 * Sets the width of the specified widget so that its visual width is as
	 * specified. This is different from
	 * {@link #setVisualWidth(Widget, GwtUiDefaults.WidgetElementType, int)}
	 * in that it guesses the element type of the widget itself, but only works
	 * for standard GWT widgets. 
	 */
    public static void setVisualWidth(Widget widget, int width) {
        setVisualWidth(widget, GwtUiDefaults.getElementType(widget), width);
    }

    /**
	 * Creates a horizontal spacer of the specified width.
	 */
    public static Widget getHorizontalStrut(int width) {
        return getRigidArea(width, 0);
    }

    /**
	 * Creates a vertical spacer of the specified height.
	 */
    public static Widget getVerticalStrut(int height) {
        return getRigidArea(0, height);
    }

    /**
	 * Creates a spacer of the specified dimensions.
	 */
    public static Widget getRigidArea(int width, int height) {
        Widget e = new SimplePanel();
        e.setPixelSize(width, height);
        return e;
    }

    /**
	 * Returns a horizontal spacer of with the size of the preferred distance
	 * between related elements.
	 */
    public static Widget getRelatedHorizontalStrut() {
        return getHorizontalStrut(GwtUiDefaults.getRelatedHorizontalDistance());
    }

    /**
	 * Returns a horizontal spacer of with the size of the preferred distance
	 * between unrelated elements.
	 */
    public static Widget getUnrelatedHorizontalStrut() {
        return getHorizontalStrut(GwtUiDefaults.getUnrelatedHorizontalDistance());
    }

    /**
	 * Returns a vertical spacer of with the size of the preferred distance
	 * between related elements.
	 */
    public static Widget getRelatedVerticalStrut() {
        return getVerticalStrut(GwtUiDefaults.getRelatedVerticalDistance());
    }

    /**
	 * Returns a vertical spacer of with the size of the preferred distance
	 * between unrelated elements.
	 */
    public static Widget getUnrelatedVerticalStrut() {
        return getVerticalStrut(GwtUiDefaults.getUnrelatedVerticalDistance());
    }

    /**
	 * Sets the specified widget's position attribute to "absolute", its 
	 * top and left attributes to "0px" and its width and height attributes to
	 * "100%".
	 */
    public static void setAbsoluteFullSize(Widget widget) {
        DOM.setStyleAttribute(widget.getElement(), "position", "absolute");
        setFullSize(widget);
    }

    /**
	 * Sets the specified widget's top and left attributes to "0px" and its
	 * width and height attributes to "100%".
	 */
    public static void setFullSize(Widget widget) {
        Element e = widget.getElement();
        DOM.setStyleAttribute(e, "left", "0px");
        DOM.setStyleAttribute(e, "top", "0px");
        DOM.setStyleAttribute(e, "width", "100%");
        DOM.setStyleAttribute(e, "height", "100%");
    }

    /**
	 * Adds the specified widget to the specified absolute panel and makes it
	 * occupy the panel fully.
	 */
    public static void addFullSize(Panel panel, Widget widget) {
        setAbsoluteFullSize(widget);
        panel.add(widget);
    }

    /**
	 * Makes the specified widget positioned absolutely and sets its bounds.
	 */
    public static void setAbsoluteBounds(Widget widget, int left, int top, int width, int height) {
        widget.getElement().getStyle().setPosition(Position.ABSOLUTE);
        setBounds(widget, left, top, width, height);
    }

    /**
	 * Makes the specified widget positioned absolutely and sets its bounds.
	 */
    public static void setAbsoluteBounds(Widget widget, RectangleView bounds) {
        setAbsoluteBounds(widget, bounds.getX(), bounds.getY(), bounds.getWidth(), bounds.getHeight());
    }

    /**
	 * Sets the "left", "top", "width" and "height" style attributes of the
	 * widget to the specified values, in pixels.
	 */
    public static void setBounds(Widget widget, int left, int top, int width, int height) {
        setBounds(widget.getElement(), left, top, width, height);
    }

    /**
	 * Sets the bounds ("left", "top", "width" and "height" style attributes)
	 * of the specified widget.
	 */
    public static void setBounds(Widget widget, RectangleView bounds) {
        setBounds(widget, bounds.getX(), bounds.getY(), bounds.getWidth(), bounds.getHeight());
    }

    /**
	 * Sets the bounds ("left", "top", "width" and "height" style attributes)
	 * of the specified element.
	 */
    public static void setBounds(Element element, RectangleView bounds) {
        setBounds(element, bounds.getX(), bounds.getY(), bounds.getWidth(), bounds.getHeight());
    }

    /**
	 * Sets the "left", "top", "width" and "height" style attributes of the
	 * widget to the specified values, in pixels.
	 */
    public static void setBounds(Element element, int left, int top, int width, int height) {
        setPosition(element, left, top);
        setSize(element, width, height);
    }

    /**
	 * Sets the "width" and "height" style attributes of the widget to the
	 * specified values, in pixels.
	 */
    public static void setSize(Widget widget, int width, int height) {
        setSize(widget.getElement(), width, height);
    }

    /**
	 * Sets the "width" and "height" attributes of the element to the specified
	 * values, in pixels.
	 */
    public static void setSize(Element element, int width, int height) {
        Style e = element.getStyle();
        e.setWidth(width, Unit.PX);
        e.setHeight(height, Unit.PX);
    }

    /**
	 * Sets the size of the widget to the specified one, in pixels. 
	 */
    public static void setSize(Widget widget, SizeView size) {
        setSize(widget, size.getWidth(), size.getHeight());
    }

    /**
	 * Sets the "left" and "top" style attributes of the widget to the specified
	 * values, in pixels.
	 */
    public static void setPosition(Widget widget, int left, int top) {
        setPosition(widget.getElement(), left, top);
    }

    /**
	 * Sets the "left" and "top" style attributes of the widget to the specified
	 * values, in pixels.
	 */
    public static void setPosition(Element element, int left, int top) {
        Style e = element.getStyle();
        e.setLeft(left, Unit.PX);
        e.setTop(top, Unit.PX);
    }

    /**
	 * Sets the position of the widget to the specified one, in pixels.
	 */
    public static void setPosition(Widget widget, PointView position) {
        setPosition(widget, position.getX(), position.getY());
    }

    /**
	 * Sets the bounds of the specified panel so that all of its widgets are
	 * within them (obviously, except those placed at negative coordinates).
	 */
    public static void setContainingSize(AbsolutePanel panel) {
        int e = 0;
        int f = 0;
        for (Widget g : panel) {
            e = Math.max(e, panel.getWidgetLeft(g) + g.getOffsetWidth());
            f = Math.max(f, panel.getWidgetTop(g) + g.getOffsetHeight());
        }
        panel.setPixelSize(e, f);
    }

    /**
	 * Adds the specified widget to the specified corner of the specified panel
	 * with the specified offset from the borders.
	 * 
	 * @param panel The panel to add to.
	 * @param widget The widget to place in a corner.
	 * @param corner The corner to place the widget in. This must be one of the
	 * alignment constants which define a corner (not anything in the middle).
	 * @param offset The offset, in pixels, from the borders.
	 */
    public static void addInCorner(Panel panel, Widget widget, Alignment corner, SizeView offset) {
        Element e = widget.getElement();
        DOM.setStyleAttribute(e, "position", "absolute");
        if (corner.getHorizontalAlignment() == ALIGN_LEFT) DOM.setStyleAttribute(e, "left", offset.getWidth() + "px"); else if (corner.getHorizontalAlignment() == ALIGN_RIGHT) DOM.setStyleAttribute(e, "right", offset.getWidth() + "px"); else throw new IllegalArgumentException("Illegal horizontal alignment value: " + corner.getHorizontalAlignment().getTextAlignString());
        if (corner.getVerticalAlignment() == ALIGN_TOP) DOM.setStyleAttribute(e, "top", offset.getHeight() + "px"); else if (corner.getVerticalAlignment() == ALIGN_BOTTOM) DOM.setStyleAttribute(e, "bottom", offset.getHeight() + "px"); else throw new IllegalArgumentException("Illegal vertical alignment value: " + corner.getVerticalAlignment().getVerticalAlignString());
        panel.add(widget);
    }

    /**
	 * Returns the top-left location of an absolutely positioned widget from
	 * its "top" and "left" style attributes. Note that this only works for
	 * absolutely positioned widgets.
	 */
    public static Point getTopLeft(Widget widget) {
        Element e = widget.getElement();
        return new Point(DOM.getIntStyleAttribute(e, "left"), DOM.getIntStyleAttribute(e, "top"));
    }

    /**
	 * Sets the widget's position via the "top" and "left" style attributes.
	 */
    public static void setTopLeft(Widget widget, PointView point) {
        setTopLeft(widget, point.getX(), point.getY());
    }

    /**
	 * Sets the widget's position via the "top" and "left" style attributes.
	 */
    public static void setTopLeft(Widget widget, int x, int y) {
        Element e = widget.getElement();
        DOM.setStyleAttribute(e, "left", x + "px");
        DOM.setStyleAttribute(e, "top", y + "px");
    }

    /**
	 * Returns the bounds of the specified widget, in the client coordinate
	 * system.
	 */
    public static Rectangle getBoundsClient(Widget widget) {
        int e = widget.getAbsoluteLeft();
        int f = widget.getAbsoluteTop();
        int g = widget.getOffsetWidth();
        int h = widget.getOffsetHeight();
        return new Rectangle(e, f, g, h);
    }

    /**
	 * Returns the size of the specified widget (offsetWidth, offsetHeight).
	 */
    public static Size getWidgetSize(Widget widget) {
        return new Size(widget.getOffsetWidth(), widget.getOffsetHeight());
    }

    /**
	 * Returns the location of the specified touch in the coordinate system
	 * of the target element.
	 */
    public static PointView getTargetCoordinates(Touch touch) {
        return getRelativeCoordinates(touch, Element.as(touch.getTarget()));
    }

    /**
	 * Returns the location of the specified touch in the coordinate system of
	 * the specified event's current target element.
	 */
    public static PointView getCurrentTargetCoordinates(Touch touch, TouchEvent<?> evt) {
        return getRelativeCoordinates(touch, Element.as(evt.getNativeEvent().getCurrentEventTarget()));
    }

    /**
	 * Returns the location of the specified touch in the coordinate system of
	 * the specified element.
	 */
    public static PointView getRelativeCoordinates(Touch touch, com.google.gwt.dom.client.Element target) {
        return new Point(touch.getRelativeX(target), touch.getRelativeY(target));
    }

    /**
	 * Disables the browser context menu for the specified element (and enables
	 * right-click detection). Note that this is not reliable across browsers
	 * (Opera, for example, doesn't care and still displays the context menu).
	 */
    public static native void disableContextMenu(Element element)/*-{
		element.oncontextmenu = function(a,b){return false};
	}-*/;

    /**
	 * Sets the id attribute of the specified widget's DOM element to the
	 * specified value.
	 */
    public static void setElementId(Widget widget, String id) {
        DOM.setElementAttribute(widget.getElement(), "id", id);
    }

    /**
	 * Returns the tag name of the specified element.
	 */
    public static String getTagName(Element element) {
        return DOM.getElementProperty(element, "tagName");
    }

    /**
	 * Finds the children of the specified element whose tag name is the
	 * specified one, and adds them to the specified list. Tag name comparison
	 * is done in a case-insensitive manner.  
	 */
    public static void findElementsByTagName(Element parent, String tagName, List<Element> list) {
        int e = DOM.getChildCount(parent);
        for (int f = 0; f < e; f++) {
            Element g = DOM.getChild(parent, f);
            findElementsByTagName(g, tagName, list);
            if (getTagName(g).equalsIgnoreCase(tagName)) list.add(g);
        }
    }

    /**
	 * Returns the "primary" style name based on the specified full style name.
	 */
    public static String getPrimaryStyleName(String fullStyleName) {
        return fullStyleName.split(" ")[0];
    }

    /**
	 * Returns the "primary" style name of the specified widget.
	 */
    public static String getPrimaryStyleName(Widget widget) {
        return getPrimaryStyleName(widget.getStyleName());
    }

    /**
	 * Adds a dependent style name to the specified widget. See
	 * {@linkplain UIObject#addStyleName(String)} for a description of dependent
	 * styles.
	 * 
	 * @param styleSuffix The suffix of the dependent style (not including the
	 * '-' separator).
	 */
    public static void addDependentStyleName(Widget widget, String styleSuffix) {
        widget.addStyleName(getPrimaryStyleName(widget) + "-" + styleSuffix);
    }

    /**
	 * Sets the visibility of the specified widget. This is different from
	 * {@link Widget#setVisible(boolean)} in that it uses the "visibility" CSS
	 * property, and not "display".
	 */
    public static void setVisibility(Widget widget, boolean isVisible) {
        DOM.setStyleAttribute(widget.getElement(), "visibility", isVisible ? "" : "hidden");
    }

    /**
	 * Returns whether the specified widget's "visibility" CSS property is not
	 * set to "hidden" or "collapsed" or "inherit" and the parent's visibility
	 * is false.
	 */
    public static boolean getVisibility(Widget widget) {
        String e = DOM.getStyleAttribute(widget.getElement(), "visibility");
        return !("hidden".equals(e) || "collapsed".equals(e) || ("inherit".equals(e) && !getVisibility(widget.getParent())));
    }

    /**
	 * Sets the z-indices of the specified widgets to their indices in the
	 * specified array, making widgets with greater indices appear on top of
	 * widgets with smaller indices.
	 */
    public static void setZOrder(Widget... widgets) {
        for (int e = 0; e < widgets.length; e++) DOM.setStyleAttribute(widgets[e].getElement(), "zIndex", String.valueOf(e));
    }

    /**
	 * Sets the opacity of the specified widget to the specified value, where 0
	 * stands for completely transparent and 1 for completely opaque.
	 */
    public static void setOpacity(Widget widget, double opacity) {
        Element e = widget.getElement();
        if (BrowserUtils.isIE8()) DOM.setStyleAttribute(e, "MsFilter", "progid:DXImageTransform.Microsoft.Alpha(Opacity=" + (int) Math.round(opacity * 100) + ")");
        if (BrowserUtils.isIE()) DOM.setStyleAttribute(e, "filter", "progid:DXImageTransform.Microsoft.Alpha(opacity = " + (int) Math.round(opacity * 100) + ")"); else DOM.setStyleAttribute(e, "opacity", String.valueOf(opacity));
    }

    /***/
    private static String b = null;

    /***/
    private static String c = null;

    /**
	 * Sets the grab cursor to the specified URL. This will be used by
	 * {@link #setGrabCursor(Widget)}.
	 */
    public static void setGrabCursorUrl(String url) {
        b = url;
    }

    /**
	 * Sets the grabbing cursor to the specified URL. This will be used by
	 * {@link #setGrabbingCursor(Widget)}.
	 */
    public static void setGrabbingCursorUrl(String url) {
        c = url;
    }

    /**
	 * Sets the "grab" cursor on the specified widget, if supported. You can set
	 * the cursor to be used via {@link #setGrabCursorUrl(String)} or it will
	 * currently only work under Mozilla browsers.
	 */
    public static void setGrabCursor(Widget widget) {
        String e = null;
        if (b != null) e = "url(" + b + ")"; else if (BrowserUtils.isMozilla()) e = "-moz-grab";
        if (e != null) DOM.setStyleAttribute(widget.getElement(), "cursor", e);
    }

    /**
	 * Sets the "grabbing" cursor on the specified widget, if supported. You can
	 * set the cursor to be used via {@link #setGrabbingCursorUrl(String)} or it
	 * will currently only work under Mozilla browsers.
	 */
    public static void setGrabbingCursor(Widget widget) {
        String e = null;
        if (b != null) e = "url(" + c + ")"; else if (BrowserUtils.isMozilla()) e = "-moz-grabbing";
        if (e != null) DOM.setStyleAttribute(widget.getElement(), "cursor", e);
    }

    /**
	 * Fixes a strange problem with an image being placed in a table cell.
	 * The problem is that there are extra 4 pixels below the image.
	 * Known anomalies, if you want to try and fix it:
	 * - If a DIV is put instead of an image, it works fine.
	 * - It works fine in HTML Transitional, but not Strict.
	 * - The behaviour is not due to GWT or anything - a plain HTML table also behaves this way.
	 * Note that tdElement must be an actual TD element.
	 */
    public static void extraImageCellHeightWorkaround(Element tdElement) {
        if (!getTagName(tdElement).equalsIgnoreCase("td")) throw new IllegalArgumentException("tdElement must be a TD element");
        DOM.setStyleAttribute(tdElement, "lineHeight", "0px");
        DOM.setStyleAttribute(tdElement, "fontSize", "1px");
    }

    /**
	 * Fixes the problem described by
	 * {@linkplain #extraImageCellHeightWorkaround(Element)} for every TD
	 * element of the specified panel.  
	 */
    public static void extraImageCellHeightWorkaround(Widget panel) {
        List<Element> e = new Vector<Element>();
        findElementsByTagName(panel.getElement(), "td", e);
        for (Element f : e) extraImageCellHeightWorkaround(f);
    }

    /**
	 * Fills the specified row of a table with vertical struts of the specified
	 * size. Note that the row must already exist.
	 */
    public static void fillRow(HTMLTable table, int row, int height) {
        int e = table.getCellCount(row);
        for (int f = 0; f < e; f++) table.setWidget(row, f, UiUtils.getVerticalStrut(height));
        Element f = table.getRowFormatter().getElement(row);
        DOM.setStyleAttribute(f, "fontSize", "0px");
    }

    /**
	 * Prepares the specified table to be a layout table by clearing its cell
	 * padding, spacing and border.
	 */
    public static HTMLTable prepareLayoutTable(HTMLTable table) {
        table.setCellPadding(0);
        table.setCellSpacing(0);
        table.setBorderWidth(0);
        return table;
    }

    /**
	 * Adds the specified mouse event handler to the specified mouse event
	 * source and returns the proper registration. This method is needed because
	 * {@link HandlesAllMouseEvents#handle(HasAllMouseHandlers, MouseDownHandler)}
	 * doesn't bother to return a proper {@link HandlerRegistration}.
	 */
    public static <H extends MouseDownHandler & MouseUpHandler & MouseOutHandler & MouseOverHandler & MouseMoveHandler & MouseWheelHandler> HandlerRegistration handleAllMouseEvents(HasAllMouseHandlers source, H receiver) {
        return new MultipleHandlerRegistration(source.addMouseDownHandler(receiver), source.addMouseUpHandler(receiver), source.addMouseOutHandler(receiver), source.addMouseOverHandler(receiver), source.addMouseMoveHandler(receiver), source.addMouseWheelHandler(receiver));
    }

    /**
	 * Attempts to track changes in the size of the specified widget and notify
	 * the specified change listener. Note that there isn't much we can actually
	 * do to get notified when the size changes - what we do is listen to size
	 * changes of the containing window, and look if the size of the widget
	 * changed as well. It's not good, but it's better than nothing.
	 */
    public static HandlerRegistration addSizeChangeHandler(final Widget widget, final ResizeHandler handler) {
        final HasResizeHandlers e = new HasResizeHandlers() {

            public HandlerRegistration addResizeHandler(ResizeHandler handler) {
                return null;
            }

            public void fireEvent(GwtEvent<?> event) {
                handler.onResize((ResizeEvent) event);
            }
        };
        ResizeHandler f = new ResizeHandler() {

            private Size lastKnownSize = new Size(widget.getOffsetWidth(), widget.getOffsetHeight());

            public void onResize(ResizeEvent event) {
                int g = widget.getOffsetWidth();
                int h = widget.getOffsetHeight();
                if (!lastKnownSize.equals(g, h)) {
                    lastKnownSize.set(g, h);
                    ResizeEvent.fire(e, g, h);
                }
            }
        };
        return Window.addResizeHandler(f);
    }

    /**
	 * Cancels processing of the current event by stopping its propagation and
	 * preventing the default action from occurring.
	 */
    public static void cancelCurrentEvent() {
        Event e = DOM.eventGetCurrentEvent();
        if (e != null) cancelEvent(e);
    }

    /**
	 * Cancels processing of the specified event by stopping its propagation and
	 * preventing the default action from occurring.
	 */
    public static void cancelEvent(NativeEvent evt) {
        evt.stopPropagation();
        evt.preventDefault();
    }

    /**
	 * Cancels the processing of the specified {@link DomEvent} by stopping its
	 * propagation and preventing the default action from occurring.
	 */
    public static <H extends EventHandler> void cancelEvent(DomEvent<H> evt) {
        evt.stopPropagation();
        evt.preventDefault();
    }

    /**
	 * Makes the specified widget draggable via the specified drag handle. The
	 * dragging will also be reported to the specified drag handler,
	 * if not <code>null</code>, with all parameters passed being relative to
	 * the parent of the dragged widget. That is, the drag target widget will be
	 * the parent and the pressed point and offset will be relative to the
	 * parent.
	 * You may make the widget un-draggable again by passing the return value to
	 * {@linkplain #disableDragging(Object)}.
	 */
    public static Object makeDraggable(Widget widget, HasAllMouseHandlers dragHandle, HandlesAllDragEvents handler) {
        return new DragTracker(widget, dragHandle, handler);
    }

    /**
	 * Makes the specified widget un-draggable, after it was made draggable with
	 * {@linkplain #makeDraggable(Widget, HasAllMouseHandlers,
	 * HandlesAllDragEvents)}.
	 * 
	 * @param dragTracker The object returned by <code>makeDraggable</code>.
	 */
    public static void disableDragging(Object dragTracker) {
        if (!(dragTracker instanceof DragTracker)) throw new IllegalArgumentException("dragTracker must be the object returned by makeDraggable");
        ((DragTracker) dragTracker).disable();
    }

    /***/
    private static class DragTracker extends HandlesAllDragEvents {

        /***/
        private final Widget e;

        /***/
        private final HandlesAllDragEvents f;

        /***/
        private final DragRecognizer g;

        /***/
        private Point h = null;

        /***/
        private Point i = null;

        /**
		 * Creates a new <code>DragTracker</code> for moving the specified
		 * widget via the specified drag handle. Dragging will also be reported
		 * to the specified handler, if not <code>null</code>, with all
		 * arguments as if the dragging occurred on the parent of the dragged
		 * widget.
		 */
        public DragTracker(Widget widget, HasAllMouseHandlers dragHandle, HandlesAllDragEvents handler) {
            if (widget == null) throw new IllegalArgumentException("widget may not be null");
            if (dragHandle == null) throw new IllegalArgumentException("dragHandle may not be null");
            this.e = widget;
            this.f = handler;
            this.g = new DragRecognizer(widget, dragHandle, DragRecognizer.STANDARD_DRAG_CONDITION);
            handle(g);
            DOM.setStyleAttribute(widget.getElement(), "position", "absolute");
        }

        /**
		 * Disables this <code>DragTracker</code>. It cannot be enabled again.
		 */
        public void disable() {
            g.setMouseEventsSource((HasAllMouseHandlers) null);
        }

        /**
		 * {@inheritDoc} 
		 */
        public void onDragStart(DragStartEvent evt) {
            Widget j = e.getParent();
            int k = e.getAbsoluteLeft() - j.getAbsoluteLeft();
            int l = e.getAbsoluteTop() - j.getAbsoluteTop();
            setTopLeft(e, new Point(k, l));
            h = new Point(k, l).moveBy(evt.getLocation());
            i = new Point(0, 0);
            if (f != null) f.onDragStart(new DragStartEvent(evt.getSourceEvent(), evt.getButton(), h));
        }

        /**
		 * {@inheritDoc}
		 */
        public void onDragCancel(DragCancelEvent evt) {
            if (f != null) f.onDragCancel(new DragCancelEvent(evt.getSourceEvent()));
            h = null;
            i = null;
        }

        /**
		 * {@inheritDoc}
		 */
        public void onDragEnd(DragEndEvent evt) {
            if (f != null) f.onDragEnd(new DragEndEvent(evt.getSourceEvent(), evt.getButton(), h, i));
            h = null;
            i = null;
        }

        /**
		 * {@inheritDoc}
		 */
        public void onDragMove(DragMoveEvent evt) {
            Point j = getTopLeft(e).moveBy(evt.getOffset());
            setTopLeft(e, j);
            i.moveBy(evt.getOffset());
            if (f != null) f.onDragMove(new DragMoveEvent(evt.getSourceEvent(), evt.getButton(), h, i));
        }
    }

    /***/
    private static final String d = GWT.getModuleBaseURL() + "clear.cache.gif";

    /***/
    private static class PngHackLoadHandler implements LoadHandler, ErrorHandler {

        /***/
        private HandlerRegistration e = null;

        /***/
        private HandlerRegistration f = null;

        /***/
        private Image g = null;

        /**
		 * Registers with the specified image.
		 */
        public void registerWith(Image image) {
            if (this.g != null) throw new IllegalStateException("This PngHackLoadHandler is already registered with another image");
            e = image.addErrorHandler(this);
            f = image.addLoadHandler(this);
            image.getElement().setPropertyBoolean("hasPngHackLoadListener", true);
            this.g = image;
        }

        /***/
        private void h() {
            if (g != null) {
                e.removeHandler();
                f.removeHandler();
                e = null;
                f = null;
                g.getElement().setPropertyBoolean("hasPngHackLoadListener", false);
                g = null;
            }
        }

        /**
		 * Invoked when an error occurs loading the image. 
		 */
        public void onError(ErrorEvent evt) {
            h();
        }

        /**
		 * Invoked when the image is loaded successfully.
		 */
        public void onLoad(LoadEvent evt) {
            Image i = this.g;
            h();
            applyPngHack(i, i.getUrl(), null);
        }

        /**
		 * Returns whether the specified image has a {@link PngHackLoadHandler}
		 * registered with it.
		 */
        public static boolean hasHandler(Image image) {
            return image.getElement().getPropertyBoolean("hasPngHackLoadListener");
        }
    }

    ;

    /**
	 * Applies the IE PNG hack to the specified image. When using the PNG hack,
	 * never call {@link Image#setUrl(String)} directly; always use this method
	 * instead. Also, using {@link Image#getUrl()} on a PNG-hacked image will
	 * not return the expected URL. You don't have to, but are advised to
	 * specify the image's size - the PNG hack is only 100% reliable if you do.
	 * 
	 * @param image The image to apply the PNG hack to.
	 * @param url The URL of the PNG image.
	 * @param size The size of the image; may be <code>null</code>.
	 */
    public static void applyPngHack(Image image, String url, Size size) {
        if ((size != null) && (size.getWidth() == 0) || (size.getHeight() == 0)) throw new IllegalArgumentException("Neither width nor height may be 0");
        if (url.indexOf("://") == -1) url = GWT.getModuleBaseURL() + url;
        String e = DOM.getElementProperty(image.getElement(), "pngHackImage");
        if (url.equals(e)) return;
        if (!url.equals(image.getUrl())) image.setUrl(url);
        int f = size == null ? image.getWidth() : size.getWidth();
        int g = size == null ? image.getHeight() : size.getHeight();
        if ((f == 0) || (g == 0)) {
            if (!PngHackLoadHandler.hasHandler(image)) new PngHackLoadHandler().registerWith(image);
            return;
        }
        image.setWidth(f + "px");
        image.setHeight(g + "px");
        DOM.setStyleAttribute(image.getElement(), "filter", "progid:DXImageTransform.Microsoft.AlphaImageLoader(src='" + url + "', sizingMethod='scale')");
        DOM.setElementProperty(image.getElement(), "pngHackImage", url);
        image.setUrl(d);
    }

    /**
	 * Returns the HTML for an image at the specified URL with the PNG hack
	 * applied to it. You don't have to, but are advised to
	 * specify the image's size - the PNG hack is only 100% reliable if you do.
	 */
    public static String getPngHackedImageHtml(String url, Size size) {
        if (size == null) return "<img src=\"" + URL.encode(url) + "\" " + "onload=\"" + "if (this.pngHackApplied) return;" + "this.pngHackApplied = true;" + "var width = this.width;" + "var height = this.height;" + "this.style.width = width + 'px';" + "this.style.height = height + 'px';" + "this.src = '" + d + "';" + "this.style.filter = " + "'progid:DXImageTransform.Microsoft.AlphaImageLoader(src=\\'" + url + "\\', sizingMethod=\\'scale\\')'" + "\" " + "/>"; else return "<img src=\"" + d + "\" " + "style=\"" + "width: " + size.getWidth() + "px;" + "height: " + size.getWidth() + "px;" + "filter: " + "'progid:DXImageTransform.Microsoft.AlphaImageLoader(src=\\'" + url + "\\', sizingMethod=\\'scale\\')'" + "\" " + "/>";
    }

    /**
	 * If the event system has not yet been initialized, initializes it.
	 */
    public static native void maybeInitializeEventSystem()/*-{
		@com.google.gwt.user.client.DOM::maybeInitializeEventSystem()();
	}-*/;

    /**
	 * Returns whether the specified value is one of
	 * {@link NativeEvent#BUTTON_LEFT}, {@link NativeEvent#BUTTON_MIDDLE} or
	 * {@link NativeEvent#BUTTON_RIGHT}.
	 */
    public static boolean isNativeMouseButton(int button) {
        return (button == NativeEvent.BUTTON_LEFT) || (button == NativeEvent.BUTTON_MIDDLE) || (button == NativeEvent.BUTTON_RIGHT);
    }

    /**
	 * Returns whether <code>ancestor</code> is an ancestor of the specified
	 * widget.
	 */
    public static boolean isAncestorOf(Widget ancestor, Widget widget) {
        if ((widget == null) || (ancestor == null)) return false;
        Widget e = widget.getParent();
        return (ancestor == e) || isAncestorOf(ancestor, e);
    }

    /**
	 * Applies the specified transform CSS property to the specified element.
	 * This is needed because currently different browsers have different names
	 * for the property. Note that IE doesn't currently implement it at all.
	 */
    public static native void setTransform(Element element, String transform)/*-{
		var propertyName;
		if ("transform" in document.body.style)
			propertyName = "transform";
		else if ("WebkitTransform" in document.body.style)
			propertyName = "WebkitTransform";
		else if ("MozTransform" in document.body.style)
			propertyName = "MozTransform";
		else if ("OTransform" in document.body.style)
			propertyName = "OTransform"; 
		else
			return;
		
		element.style[propertyName] = transform;
	 }-*/;

    /**
	 * Applies the specified transform-origin CSS property to the specified
	 * element. This is needed because currently different browsers have
	 * different names for the property. Note that IE doesn't currently
	 * implement it at all.
	 */
    public static native void setTransformOrigin(Element element, String origin)/*-{
		var propertyName;
		if ("transformOrigin" in document.body.style)
			propertyName = "transformOrigin";
		else if ("WebkitTransformOrigin" in document.body.style)
			propertyName = "WebkitTransformOrigin";
		else if ("MozTransformOrigin" in document.body.style)
			propertyName = "MozTransformOrigin";
		else if ("OTransformOrigin" in document.body.style)
			propertyName = "OTransformOrigin";
		else
			return;
		
		element.style[propertyName] = origin;
	}-*/;

    /**
	 * Registers a load listener for the specified element which will invoke
	 * the specified callback with the element. Returns an object which may be
	 * passed to 
	 */
    public static native Object addLoadListener(com.google.gwt.dom.client.Element element, Callback<com.google.gwt.dom.client.Element> callback)/*-{
		var listener = function(){
			callback.@com.maryanovsky.gwtutils.client.event.Callback::callback(Ljava/lang/Object;)(element);
		};
		
		if (element.addEventListener)
			element.addEventListener("load", listener, false);
		else if (element.attachEvent)
			element.attachEvent("onload", listener);
			
		return listener;
	}-*/;

    /**
	 * Unregisters a load listener from the specified element. The listenerId
	 * argument should be the object returned by
	 * {@link #addLoadListener(com.google.gwt.dom.client.Element, Callback)}.
	 */
    public static native void removeListener(com.google.gwt.dom.client.Element element, Object listenerId)/*-{
		if (element.removeEventListener)
			element.removeEventListener("load", listenerId, false);
		else if (element.detachEvent)
			element.detachEvent("onload", listenerId);
	}-*/;
}
