package com.maryanovsky.gwtutils.client.ui.mouse.gestures;

import java.util.ArrayList;
import java.util.List;
import com.google.gwt.event.dom.client.*;
import com.google.gwt.event.shared.GwtEvent;
import com.google.gwt.event.shared.HandlerManager;
import com.google.gwt.event.shared.HandlerRegistration;

/**
 * The base class for all mouse gesture recognizers.
 * 
 * @author Maryanovsky Alexander
 */
public abstract class MouseGestureRecognizer {

    /**
	 * The manager of handlers.
	 */
    protected final HandlerManager handlerManager;

    /***/
    private final HandlesAllMouseEvents a = new HandlesAllMouseEvents() {

        public void onMouseDown(MouseDownEvent event) {
            MouseGestureRecognizer.this.onMouseDown(event);
        }

        public void onMouseUp(MouseUpEvent event) {
            MouseGestureRecognizer.this.onMouseUp(event);
        }

        public void onMouseMove(MouseMoveEvent event) {
            MouseGestureRecognizer.this.onMouseMove(event);
        }

        public void onMouseOut(MouseOutEvent event) {
            MouseGestureRecognizer.this.onMouseOut(event);
        }

        public void onMouseOver(MouseOverEvent event) {
            MouseGestureRecognizer.this.onMouseOver(event);
        }

        public void onMouseWheel(MouseWheelEvent event) {
            MouseGestureRecognizer.this.onMouseWheel(event);
        }
    };

    /***/
    private final List<HandlerRegistration> b = new ArrayList<HandlerRegistration>(4);

    /**
	 * Creates a new {@link MouseGestureRecognizer} which will dispatch events
	 * with the specified source object.
	 */
    public MouseGestureRecognizer(Object firedEventsSource) {
        this(firedEventsSource, (HasAllMouseHandlers) null);
    }

    /**
	 * Creates a new {@link MouseGestureRecognizer} which will dispatch events
	 * with the specified source object. Additionally, you may pass the source
	 * of mouse events.
	 */
    public <H extends HasMouseDownHandlers & HasMouseUpHandlers & HasMouseMoveHandlers & HasMouseOutHandlers & HasMouseOverHandlers & HasMouseWheelHandlers> MouseGestureRecognizer(Object firedEventsSource, H mouseEventsSource) {
        handlerManager = new HandlerManager(firedEventsSource);
        setMouseEventsSource(mouseEventsSource);
    }

    /**
	 * Sets the source of mouse events for this recognizer. Passing
	 * <code>null</code> will cause the recognizer to deregister itself from the
	 * previous source.
	 */
    public <H extends HasMouseDownHandlers & HasMouseUpHandlers & HasMouseMoveHandlers & HasMouseOutHandlers & HasMouseOverHandlers & HasMouseWheelHandlers> void setMouseEventsSource(H mouseEventsSource) {
        for (HandlerRegistration c : b) c.removeHandler();
        b.clear();
        if (mouseEventsSource != null) {
            b.add(mouseEventsSource.addMouseDownHandler(a));
            b.add(mouseEventsSource.addMouseUpHandler(a));
            b.add(mouseEventsSource.addMouseMoveHandler(a));
            b.add(mouseEventsSource.addMouseOutHandler(a));
            b.add(mouseEventsSource.addMouseOverHandler(a));
            b.add(mouseEventsSource.addMouseWheelHandler(a));
        }
    }

    /**
	 * Fires the specified event to all its handlers.
	 */
    protected void fireEvent(GwtEvent<?> event) {
        handlerManager.fireEvent(event);
    }

    /**
	 * Invoked when a mouse button is pressed. 
	 */
    protected void onMouseDown(MouseDownEvent event) {
    }

    /**
	 * Invoked when a mouse button is released.
	 */
    protected void onMouseUp(MouseUpEvent event) {
    }

    /**
	 * Invoked when the mouse is moved.
	 */
    protected void onMouseMove(MouseMoveEvent event) {
    }

    /**
	 * Invoked when the mouse cursor leaves the element.
	 */
    protected void onMouseOut(MouseOutEvent event) {
    }

    /**
	 * Invoked when the mouse cursor enters an element.
	 */
    protected void onMouseOver(MouseOverEvent event) {
    }

    /**
	 * Invoked when the mouse wheel is scrolled.
	 */
    protected void onMouseWheel(MouseWheelEvent event) {
    }
}
