package com.maryanovsky.gwtutils.client.ui.mouse.gestures.drag;

import com.google.gwt.dom.client.NativeEvent;
import com.google.gwt.event.dom.client.*;
import com.google.gwt.event.shared.HandlerRegistration;
import com.google.gwt.user.client.DOM;
import com.google.gwt.user.client.ui.UIObject;
import com.maryanovsky.gwtutils.client.Condition;
import com.maryanovsky.gwtutils.client.geom.Point;
import com.maryanovsky.gwtutils.client.geom.PointView;
import com.maryanovsky.gwtutils.client.ui.mouse.gestures.MouseGestureRecognizer;

/**
 * Recognizes and fires mouse-drag events.
 * 
 * @author Maryanovsky Alexander
 */
public class DragRecognizer extends MouseGestureRecognizer implements HasAllDragHandlers {

    /**
	 * The condition for the standard drag - left mouse button and no keyboard
	 * modifiers.
	 */
    public static Condition<DragStartEvent> STANDARD_DRAG_CONDITION = new Condition<DragStartEvent>() {

        public boolean eval(DragStartEvent evt) {
            MouseDownEvent h = evt.getSourceEvent();
            return (h.getNativeButton() == NativeEvent.BUTTON_LEFT) && !(h.isAltKeyDown() || h.isControlKeyDown() || h.isMetaKeyDown() || h.isShiftKeyDown());
        }
    };

    /***/
    private final Condition<? super DragStartEvent> a;

    /***/
    private boolean b = false;

    /***/
    private int c = 0;

    /***/
    private Point d = null;

    /***/
    private Point e = null;

    /***/
    private boolean f = false;

    /***/
    private boolean g = false;

    /**
	 * Creates a new {@link DragRecognizer} which will fire drag events with the
	 * specified source object.
	 */
    public DragRecognizer(Object dragEventsSource) {
        this(dragEventsSource, (HasAllMouseHandlers) null, null);
    }

    /**
	 * Creates a new {@link DragRecognizer} which will fire drag events with the
	 * specified source object and use the specified source of mouse events.
	 */
    public <H extends HasMouseDownHandlers & HasMouseUpHandlers & HasMouseMoveHandlers & HasMouseOutHandlers & HasMouseOverHandlers & HasMouseWheelHandlers> DragRecognizer(Object dragEventsSource, H mouseEventsSource) {
        this(dragEventsSource, mouseEventsSource, null);
    }

    /**
	 * Creates a new {@link DragRecognizer} which will fire drag events with
	 * the specified source object and consult the specified drag condition
	 * before doing so.
	 */
    public DragRecognizer(Object dragEventsSource, Condition<? super DragStartEvent> dragCondition) {
        this(dragEventsSource, (HasAllMouseHandlers) null, dragCondition);
    }

    /**
	 * Creates a new {@link DragRecognizer} which will fire events with the
	 * specified source object, use the specified source of mouse events and
	 * consult the specified drag condition before firing the drag event(s).
	 * See the documentation of {@link #dragCondition(DragStartEvent)} for more
	 * information on the drag condition.
	 */
    public <H extends HasMouseDownHandlers & HasMouseUpHandlers & HasMouseMoveHandlers & HasMouseOutHandlers & HasMouseOverHandlers & HasMouseWheelHandlers> DragRecognizer(Object dragEventsSource, H mouseEventsSource, Condition<? super DragStartEvent> dragCondition) {
        super(dragEventsSource, mouseEventsSource);
        this.a = dragCondition;
    }

    /**
	 * Adds a drag start handler. 
	 */
    public HandlerRegistration addDragStartHandler(DragStartHandler handler) {
        return handlerManager.addHandler(DragStartEvent.getType(), handler);
    }

    /**
	 * Adds a drag move handler.
	 */
    public HandlerRegistration addDragMoveHandler(DragMoveHandler handler) {
        return handlerManager.addHandler(DragMoveEvent.getType(), handler);
    }

    /**
	 * Adds a drag end handler.
	 */
    public HandlerRegistration addDragEndHandler(DragEndHandler handler) {
        return handlerManager.addHandler(DragEndEvent.getType(), handler);
    }

    /**
	 * Adds a drag cancel handler.
	 */
    public HandlerRegistration addDragCancelHandler(DragCancelHandler handler) {
        return handlerManager.addHandler(DragCancelEvent.getType(), handler);
    }

    /**
	 * Returns whether the specified {@link DragStartEvent} and the drag events
	 * that follow it in the drag gesture should be delivered to the handlers.
	 * This allows filtering drag gestures to only those you really want.
	 * For example, if you only want to recognize left-mouse-button drags, you
	 * would override this method and return
	 * <code>evt.getButton() == NativeEvent.BUTTON_LEFT</code>.
	 * The default implementation delegates the decision to the drag condition
	 * passed in the constructor. 
	 */
    protected boolean dragCondition(DragStartEvent evt) {
        return a == null ? true : a.eval(evt);
    }

    /**
	 * {@inheritDoc}
	 */
    @Override
    protected void onMouseDown(MouseDownEvent event) {
        if (b) return;
        d = Point.make(event.getX(), event.getY(), d);
        b = true;
        c = event.getNativeButton();
        f = false;
        DragStartEvent h = new DragStartEvent(event, c, d);
        g = dragCondition(h);
        if (g) handlerManager.fireEvent(h);
        if (event.getSource() instanceof UIObject) DOM.setCapture(((UIObject) event.getSource()).getElement());
    }

    /**
	 * {@inheritDoc}
	 */
    @Override
    protected void onMouseUp(MouseUpEvent event) {
        if (!b) return;
        if (event.getNativeButton() != c) return;
        if (f) {
            e = Point.make(event.getX(), event.getY(), e);
            PointView h = d.getDistancePoint(e);
            if (g) handlerManager.fireEvent(new DragEndEvent(event, c, e, h));
        } else {
            if (g) handlerManager.fireEvent(new DragCancelEvent(event));
        }
        b = false;
        if (event.getSource() instanceof UIObject) DOM.releaseCapture(((UIObject) event.getSource()).getElement());
    }

    /**
	 * {@inheritDoc}
	 */
    @Override
    protected void onMouseMove(MouseMoveEvent event) {
        if ((e != null) && e.equals(event.getX(), event.getY())) return;
        e = Point.make(event.getX(), event.getY(), e);
        if (b) {
            f = true;
            PointView h = d.getDistancePoint(e);
            if (g) handlerManager.fireEvent(new DragMoveEvent(event, c, e, h));
        }
    }
}
