package com.maryanovsky.gwtutils.client.ui.mouse.gestures.wheelscroll;

import com.google.gwt.event.dom.client.*;
import com.google.gwt.event.shared.HandlerRegistration;
import com.maryanovsky.gwtutils.client.Condition;
import com.maryanovsky.gwtutils.client.geom.Point;
import com.maryanovsky.gwtutils.client.ui.mouse.gestures.MouseGestureRecognizer;

/**
 * Recognizes and fires mouse wheel scroll events.
 * 
 * @author Maryanovsky Alexander
 */
public class WheelScrollRecognizer extends MouseGestureRecognizer implements HasAllWheelScrollHandlers {

    /***/
    private final Condition<WheelScrollUpEvent> a;

    /***/
    private final Condition<WheelScrollDownEvent> b;

    /***/
    private static final int c = 300;

    /***/
    private long d = 0;

    /**
	 * Creates a new {@link WheelScrollRecognizer} which will fire mouse wheel
	 * scroll events with the specified source object.
	 */
    public WheelScrollRecognizer(Object wheelScrollEventsSource) {
        this(wheelScrollEventsSource, (HasAllMouseHandlers) null, null, null);
    }

    /**
	 * Creates a new {@link WheelScrollRecognizer} which will fire mouse wheel
	 * scroll events with the specified source object and use the specified
	 * source of mouse events.
	 */
    public <H extends HasMouseDownHandlers & HasMouseUpHandlers & HasMouseMoveHandlers & HasMouseOutHandlers & HasMouseOverHandlers & HasMouseWheelHandlers> WheelScrollRecognizer(Object wheelScrollEventsSource, H mouseEventsSource) {
        this(wheelScrollEventsSource, mouseEventsSource, null, null);
    }

    /**
	 * Creates a new {@link WheelScrollRecognizer} which will fire mouse wheel
	 * scroll events with the specified source object and consult the specified
	 * conditions before doing so.
	 */
    public WheelScrollRecognizer(Object wheelScrollEventsSource, Condition<WheelScrollUpEvent> wheelScrollUpCondition, Condition<WheelScrollDownEvent> wheelScrollDownCondition) {
        this(wheelScrollEventsSource, (HasAllMouseHandlers) null, wheelScrollUpCondition, wheelScrollDownCondition);
    }

    /**
	 * Creates a new {@link WheelScrollRecognizer} which will fire mouse wheel
	 * scroll events with the specified source object, use the specified source
	 * of mouse events and consult the specified conditions before firing the
	 * event(s).
	 * See the documentation of
	 * {@link #wheelScrollUpCondition(WheelScrollUpEvent)} and
	 * {@link #wheelScrollDownCondition(WheelScrollDownEvent)} for information
	 * on the wheel scroll conditions.
	 */
    public <H extends HasMouseDownHandlers & HasMouseUpHandlers & HasMouseMoveHandlers & HasMouseOutHandlers & HasMouseOverHandlers & HasMouseWheelHandlers> WheelScrollRecognizer(Object wheelScrollEventsSource, H mouseEventsSource, Condition<WheelScrollUpEvent> wheelScrollUpCondition, Condition<WheelScrollDownEvent> wheelScrollDownCondition) {
        super(wheelScrollEventsSource, mouseEventsSource);
        this.a = wheelScrollUpCondition != null ? wheelScrollUpCondition : Condition.<WheelScrollUpEvent>truth();
        this.b = wheelScrollDownCondition != null ? wheelScrollDownCondition : Condition.<WheelScrollDownEvent>truth();
    }

    /**
	 * Adds a mouse wheel scroll-up handler. 
	 */
    public HandlerRegistration addWheelScrollUpHandler(WheelScrollUpHandler handler) {
        return handlerManager.addHandler(WheelScrollUpEvent.getType(), handler);
    }

    /**
	 * Adds a mouse wheel scroll-down handler. 
	 */
    public HandlerRegistration addWheelScrollDownHandler(WheelScrollDownHandler handler) {
        return handlerManager.addHandler(WheelScrollDownEvent.getType(), handler);
    }

    /**
	 * Returns whether the specified {@link WheelScrollUpEvent} should be
	 * delivered to the handlers. This allows filtering mouse wheel scroll up
	 * gestures to only those you really want. For example, if you only want to
	 * recognize scrolls while the control key is held, you would override this
	 * method and return
	 * <code>evt.getSourceEvent().isControlKeyDown()</code>.
	 * The default implementation delegates the decision to the wheel scroll up
	 * condition passed in the constructor.
	 */
    protected boolean wheelScrollUpCondition(WheelScrollUpEvent evt) {
        return a.eval(evt);
    }

    /**
	 * Returns whether the specified {@link WheelScrollDownEvent} should be
	 * delivered to the handlers. This allows filtering mouse wheel scroll down
	 * gestures to only those you really want. For example, if you only want to
	 * recognize scrolls while the control key is held, you would override this
	 * method and return
	 * <code>evt.getSourceEvent().isControlKeyDown()</code>.
	 * The default implementation delegates the decision to the wheel scroll
	 * down condition passed in the constructor.
	 */
    protected boolean wheelScrollDownCondition(WheelScrollDownEvent evt) {
        return b.eval(evt);
    }

    /**
	 * {@inheritDoc}
	 */
    @Override
    protected void onMouseWheel(MouseWheelEvent event) {
        long e = System.currentTimeMillis();
        if (e - d > c) {
            Point f = new Point(event.getX(), event.getY());
            if (event.isNorth()) {
                d = e;
                WheelScrollUpEvent g = new WheelScrollUpEvent(event, f);
                if (wheelScrollUpCondition(g)) handlerManager.fireEvent(g);
            } else if (event.isSouth()) {
                d = e;
                WheelScrollDownEvent g = new WheelScrollDownEvent(event, f);
                if (wheelScrollDownCondition(g)) handlerManager.fireEvent(g);
            }
        }
    }
}
