package com.maryanovsky.gwtutils.client.ui.touch.gestures;

import java.util.ArrayList;
import java.util.List;
import com.google.gwt.event.dom.client.*;
import com.google.gwt.event.shared.GwtEvent;
import com.google.gwt.event.shared.HandlerManager;
import com.google.gwt.event.shared.HandlerRegistration;

/**
 * The base class for all touch gesture recognizers.
 * 
 * @author Maryanovsky Alexander
 */
public abstract class TouchGestureRecognizer {

    /**
	 * The manager of handlers.
	 */
    protected final HandlerManager handlerManager;

    /***/
    private final HandlesAllTouchEvents a = new HandlesAllTouchEvents() {

        public void onTouchStart(TouchStartEvent evt) {
            TouchGestureRecognizer.this.onTouchStart(evt);
        }

        public void onTouchMove(TouchMoveEvent evt) {
            TouchGestureRecognizer.this.onTouchMove(evt);
        }

        public void onTouchEnd(TouchEndEvent evt) {
            TouchGestureRecognizer.this.onTouchEnd(evt);
        }

        public void onTouchCancel(TouchCancelEvent evt) {
            TouchGestureRecognizer.this.onTouchCancel(evt);
        }
    };

    /***/
    private final List<HandlerRegistration> b = new ArrayList<HandlerRegistration>(4);

    /**
	 * Creates a new {@link TouchGestureRecognizer} which will dispatch events
	 * with the specified source object.
	 */
    public TouchGestureRecognizer(Object firedEventsSource) {
        handlerManager = new HandlerManager(firedEventsSource);
    }

    /**
	 * Creates a new {@link TouchGestureRecognizer} which will dispatch events
	 * with the specified source object. Additionally, you may pass the source
	 * of touch events.
	 */
    public <H extends HasTouchStartHandlers & HasTouchMoveHandlers & HasTouchEndHandlers & HasTouchCancelHandlers> TouchGestureRecognizer(Object eventSource, H touchEventsSource) {
        this(eventSource);
        setTouchEventsSource(touchEventsSource);
    }

    /**
	 * Sets the source of touch events for this recognizer. Passing
	 * <code>null</code> will cause the recognizer to deregister itself from the
	 * previous source.
	 */
    public <H extends HasTouchStartHandlers & HasTouchMoveHandlers & HasTouchEndHandlers & HasTouchCancelHandlers> void setTouchEventsSource(H touchEventsSource) {
        for (HandlerRegistration c : b) c.removeHandler();
        b.clear();
        if (touchEventsSource != null) {
            b.add(touchEventsSource.addTouchStartHandler(a));
            b.add(touchEventsSource.addTouchMoveHandler(a));
            b.add(touchEventsSource.addTouchEndHandler(a));
            b.add(touchEventsSource.addTouchCancelHandler(a));
        }
    }

    /**
	 * Fires the specified event to all its handlers.
	 */
    protected void fireEvent(GwtEvent<?> event) {
        handlerManager.fireEvent(event);
    }

    /**
	 * Invoked when a touch starts.
	 */
    protected void onTouchStart(TouchStartEvent evt) {
    }

    /**
	 * Invoked when a finger touching the screen is moved.
	 */
    protected void onTouchMove(TouchMoveEvent evt) {
    }

    /**
	 * Invoked when a finger touching the screen is lifted.
	 */
    protected void onTouchEnd(TouchEndEvent evt) {
    }

    /**
	 * Invoked when a touch event is canceled.
	 */
    protected void onTouchCancel(TouchCancelEvent evt) {
    }
}
