package com.maryanovsky.gwtutils.client.ui.touch.gestures.buttontap;

import com.google.gwt.dom.client.Element;
import com.google.gwt.event.dom.client.*;
import com.google.gwt.event.shared.HandlerRegistration;
import com.maryanovsky.gwtutils.client.Condition;
import com.maryanovsky.gwtutils.client.geom.PointView;
import com.maryanovsky.gwtutils.client.geom.Rectangle;
import com.maryanovsky.gwtutils.client.geom.RectangleView;
import com.maryanovsky.gwtutils.client.ui.UiUtils;
import com.maryanovsky.gwtutils.client.ui.touch.gestures.TouchGestureRecognizer;
import com.maryanovsky.gwtutils.client.ui.touch.gestures.tap.TapRecognizer;

/**
 * Recognizes button-taps and fires tap events. Unlike a regular tap (recognized
 * by {@link TapRecognizer}), a button-tap gesture allows the finger to move
 * between the touch-start and touch-end events, does not limit the amount of
 * time the finger can be pressed. Additionally, there is no such thing as a
 * multi-button-tap and the event object does not include the location of the
 * tap.
 * 
 * @author Maryanovsky Alexander
 */
public class ButtonTapRecognizer extends TouchGestureRecognizer implements HasButtonTapHandlers {

    /***/
    private final Condition<? super ButtonTapEvent> a;

    /**
	 * Creates a new {@link ButtonTapRecognizer} which will fire events with the
	 * specified source object.
	 */
    public ButtonTapRecognizer(Object buttonTapEventsSource) {
        this(buttonTapEventsSource, (HasAllTouchHandlers) null, null);
    }

    /**
	 * Creates a new {@link ButtonTapRecognizer} which will fire events with the
	 * specified source object and use the specified source of touch events.
	 */
    public <H extends HasTouchStartHandlers & HasTouchMoveHandlers & HasTouchEndHandlers & HasTouchCancelHandlers> ButtonTapRecognizer(Object buttonTapEventsSource, H touchEventsSource) {
        this(buttonTapEventsSource, touchEventsSource, null);
    }

    /**
	 * Creates a new {@link ButtonTapRecognizer} which will fire events with the
	 * specified source object and consult the specified button-tap condition
	 * before firing button-tap events. 
	 */
    public ButtonTapRecognizer(Object buttonTapEventsSource, Condition<? super ButtonTapEvent> buttonTapCondition) {
        this(buttonTapEventsSource, (HasAllTouchHandlers) null, buttonTapCondition);
    }

    /**
	 * Creates a new {@link ButtonTapRecognizer} which will fire events with the
	 * specified source object, use the specified source of touch events and
	 * consult the specified button-tap condition before firing button-tap
	 * events.
	 * See the documentation of {@link #buttonTapCondition(ButtonTapEvent)} for
	 * more information on the button-tap condition.
	 */
    public <H extends HasTouchStartHandlers & HasTouchMoveHandlers & HasTouchEndHandlers & HasTouchCancelHandlers> ButtonTapRecognizer(Object buttonTapEventsSource, H touchEventsSource, Condition<? super ButtonTapEvent> buttonTapCondition) {
        super(buttonTapEventsSource, touchEventsSource);
        this.a = buttonTapCondition;
    }

    /**
	 * Adds a button-tap handler.
	 */
    public HandlerRegistration addButtonTapHandler(ButtonTapHandler handler) {
        return handlerManager.addHandler(ButtonTapEvent.getType(), handler);
    }

    /**
	 * Returns whether the specified {@link ButtonTapEvent} should be delivered
	 * to the handlers. This allows filtering tap gestures to only those you
	 * really want.
	 * The default implementation delegates the decision to the button-tap
	 * condition passed in the constructor. 
	 */
    protected boolean buttonTapCondition(ButtonTapEvent evt) {
        return a == null ? true : a.eval(evt);
    }

    /**
	 * {@inheritDoc}
	 */
    @Override
    protected void onTouchEnd(TouchEndEvent evt) {
        if (evt.getTargetTouches().length() != 0) return;
        PointView b = UiUtils.getCurrentTargetCoordinates(evt.getChangedTouches().get(0), evt);
        Element c = Element.as(evt.getNativeEvent().getCurrentEventTarget());
        RectangleView d = new Rectangle(0, 0, c.getOffsetWidth(), c.getOffsetHeight());
        if (!d.contains(b)) return;
        ButtonTapEvent e = new ButtonTapEvent(evt);
        if (buttonTapCondition(e)) fireEvent(e);
    }
}
