package com.maryanovsky.gwtutils.client.ui.touch.gestures.mouse;

import com.google.gwt.dom.client.Document;
import com.google.gwt.dom.client.NativeEvent;
import com.google.gwt.event.dom.client.ClickEvent;
import com.google.gwt.event.dom.client.ClickHandler;
import com.google.gwt.event.dom.client.DomEvent;
import com.google.gwt.event.dom.client.HasClickHandlers;
import com.google.gwt.event.shared.GwtEvent;
import com.google.gwt.event.shared.HandlerManager;
import com.google.gwt.event.shared.HandlerRegistration;
import com.google.gwt.user.client.Element;
import com.google.gwt.user.client.ui.FocusPanel;
import com.google.gwt.user.client.ui.UIObject;
import com.maryanovsky.gwtutils.client.ui.touch.gestures.buttontap.ButtonTapEvent;
import com.maryanovsky.gwtutils.client.ui.touch.gestures.buttontap.ButtonTapHandler;
import com.maryanovsky.gwtutils.client.ui.touch.gestures.buttontap.HasButtonTapHandlers;
import com.maryanovsky.gwtutils.client.ui.touch.gestures.tap.HasTapHandlers;
import com.maryanovsky.gwtutils.client.ui.touch.gestures.tap.TapEvent;
import com.maryanovsky.gwtutils.client.ui.touch.gestures.tap.TapHandler;

/**
 * Generates synthetic click events from corresponding tap events. This is
 * useful for widget implementers that want to add basic iPhone support without 
 * modifying their widget's API. The converter can be used in two ways:
 * <ul>
 *   <li>As a direct source of click events, by registering mouse handlers with
 *       it.
 *   <li>As an indirect source of mouse events, by specifying a non-null target
 *       widget, which will cause the converter to dispatch synthetic click
 *       events on that widget.
 * </ul>
 *  
 * @author Maryanovsky Alexander
 */
public class TapClickEventsConverter implements HasClickHandlers {

    /***/
    private final UIObject a;

    /***/
    private final HandlerManager b;

    /***/
    private final TapHandler c = new TapHandler() {

        public void onTap(TapEvent evt) {
            TapClickEventsConverter.this.onTap(evt);
        }
    };

    /***/
    private final ButtonTapHandler d = new ButtonTapHandler() {

        public void onButtonTap(ButtonTapEvent evt) {
            TapClickEventsConverter.this.onButtonTap(evt);
        }
    };

    /***/
    private HandlerRegistration e;

    /**
	 * Creates a new {@link TapClickEventsConverter} with the specified target
	 * widget. If the target widget is not <code>null</code>, synthetic click
	 * events will be dispatched on it, and its element will be used as the
	 * element relative to which event coordinates are measured.
	 */
    public TapClickEventsConverter(UIObject target) {
        this.a = target;
        this.b = new HandlerManager(target);
    }

    /**
	 * Creates a new {@link TapClickEventsConverter} with the specified target
	 * widget. If the target widget is not <code>null</code>, synthetic click
	 * events will be dispatched on it, and its element will be used as the
	 * element relative to which event coordinates are measured. Additionally,
	 * you may pass the source of tap events.
	 */
    public <H extends HasTapHandlers> TapClickEventsConverter(UIObject target, H tapEventsSource) {
        this(target);
        setTapEventsSource(tapEventsSource);
    }

    /**
	 * Creates a new {@link TapClickEventsConverter} with the specified target
	 * widget. If the target widget is not <code>null</code>, synthetic click
	 * events will be dispatched on it, and its element will be used as the
	 * element relative to which event coordinates are measured. Additionally,
	 * you may pass the source of button-tap events.
	 */
    public <H extends HasButtonTapHandlers> TapClickEventsConverter(UIObject target, H buttonTapEventsSource) {
        this(target);
        setButtonTapEventsSource(buttonTapEventsSource);
    }

    /**
	 * Sets the source of tap events for this recognizer. Passing
	 * <code>null</code> will cause the recognizer to deregister itself from the
	 * previous source. You may not simultaneously set both a regular and a
	 * button-tap source.
	 */
    public <H extends HasTapHandlers> void setTapEventsSource(H source) {
        if (e != null) {
            e.removeHandler();
            e = null;
        }
        if (source != null) e = source.addTapHandler(c);
    }

    /**
	 * Sets the source of button-tap events for this recognizer. Passing
	 * <code>null</code> will cause the recognizer to deregister itself from the
	 * previous source. You may not simultaneously set both a regular and a
	 * button-tap source.
	 */
    public <H extends HasButtonTapHandlers> void setButtonTapEventsSource(H source) {
        if (e != null) {
            e.removeHandler();
            e = null;
        }
        if (source != null) e = source.addButtonTapHandler(d);
    }

    /**
	 * {@inheritDoc}
	 */
    public HandlerRegistration addClickHandler(ClickHandler handler) {
        return b.addHandler(ClickEvent.getType(), handler);
    }

    /**
	 * {@inheritDoc}
	 */
    public void fireEvent(GwtEvent<?> event) {
        new FocusPanel();
        b.fireEvent(event);
    }

    /**
	 * Unregisters this {@link TapClickEventsConverter} from its source, which
	 * will cause it to stop firing any click events. You cannot "restart" a
	 * the converter after it has been stopped.
	 */
    public void stop() {
        if (e != null) {
            e.removeHandler();
            e = null;
        }
    }

    /**
	 * Invoked when a tap occurs.
	 */
    protected void onTap(TapEvent evt) {
        f(evt.getSourceEvent().getNativeEvent(), evt.getTapCount());
    }

    /**
	 * Invoked when a button-tap occurs.
	 */
    protected void onButtonTap(ButtonTapEvent evt) {
        f(evt.getSourceEvent().getNativeEvent(), 1);
    }

    /***/
    private void f(NativeEvent g, int h) {
        int i = g.getScreenX();
        int j = g.getScreenY();
        int k = g.getClientX();
        int l = g.getClientY();
        Document m = Document.get();
        Element n = a == null ? null : a.getElement();
        NativeEvent o = m.createClickEvent(h, i, j, k, l, false, false, false, false);
        DomEvent.fireNativeEvent(o, this, n);
        if (n != null) n.dispatchEvent(o);
    }
}
