package com.maryanovsky.gwtutils.client.ui.touch.gestures.mouse;

import com.google.gwt.dom.client.Document;
import com.google.gwt.dom.client.NativeEvent;
import com.google.gwt.dom.client.Touch;
import com.google.gwt.event.dom.client.*;
import com.google.gwt.event.shared.GwtEvent;
import com.google.gwt.event.shared.HandlerRegistration;
import com.google.gwt.user.client.Element;
import com.google.gwt.user.client.Event;
import com.google.gwt.user.client.ui.UIObject;
import com.maryanovsky.gwtutils.client.geom.Point;
import com.maryanovsky.gwtutils.client.ui.touch.gestures.TouchGestureRecognizer;

/**
 * Generates synthetic mouse events from corresponding touch events. This is
 * useful for widget implementers that want to add basic iPhone support without 
 * modifying their widget's API. The converter can be used in two ways:
 * <ul>
 *   <li>As a direct source of mouse events, by registering mouse handlers with
 *       it.
 *   <li>As an indirect source of mouse events, by specifying a non-null target
 *       widget and the , which will cause the converter to dispatch synthetic mouse
 *       events on that widget.
 * </ul>
 *  
 * @author Maryanovsky Alexander
 */
public class TouchMouseEventsConverter extends TouchGestureRecognizer implements HasAllMouseHandlers {

    /***/
    private final UIObject a;

    /***/
    private final Point b = new Point();

    /***/
    private final Point c = new Point();

    /**
	 * Creates a new {@link TouchMouseEventsConverter} with the specified source
	 * of fired events and the target widget on which synthetic mouse events are
	 * to be dispatched. The target widget may be <code>null</code>, in which
	 * case synthetic mouse events will only be sent to handlers directly
	 * registered with this {@link TouchMouseEventsConverter}.
	 */
    public TouchMouseEventsConverter(Object firedEventsSource, UIObject target) {
        super(firedEventsSource);
        this.a = target;
    }

    /**
	 * Creates a new {@link TouchMouseEventsConverter} with the specified source
	 * of fired events and the target widget on which synthetic mouse events are
	 * to be dispatched. The target widget may be <code>null</code>, in which
	 * case synthetic mouse events will only be sent to handlers directly
	 * registered with this {@link TouchMouseEventsConverter}. Additionally, you
	 * may pass the source of touch events.
	 */
    public <H extends HasTouchStartHandlers & HasTouchMoveHandlers & HasTouchEndHandlers & HasTouchCancelHandlers> TouchMouseEventsConverter(Object firedEventsSource, UIObject target, H touchEventsSource) {
        super(firedEventsSource, touchEventsSource);
        this.a = target;
    }

    /**
	 * {@inheritDoc}
	 */
    public HandlerRegistration addMouseDownHandler(MouseDownHandler handler) {
        return handlerManager.addHandler(MouseDownEvent.getType(), handler);
    }

    /**
	 * {@inheritDoc}
	 */
    public HandlerRegistration addMouseMoveHandler(MouseMoveHandler handler) {
        return handlerManager.addHandler(MouseMoveEvent.getType(), handler);
    }

    /**
	 * {@inheritDoc}
	 */
    public HandlerRegistration addMouseUpHandler(MouseUpHandler handler) {
        return handlerManager.addHandler(MouseUpEvent.getType(), handler);
    }

    /**
	 * {@inheritDoc}
	 */
    public HandlerRegistration addMouseOverHandler(MouseOverHandler handler) {
        return handlerManager.addHandler(MouseOverEvent.getType(), handler);
    }

    /**
	 * {@inheritDoc}
	 */
    public HandlerRegistration addMouseOutHandler(MouseOutHandler handler) {
        return handlerManager.addHandler(MouseOutEvent.getType(), handler);
    }

    /**
	 * {@inheritDoc}
	 */
    public HandlerRegistration addMouseWheelHandler(MouseWheelHandler handler) {
        return handlerManager.addHandler(MouseWheelEvent.getType(), handler);
    }

    /**
	 * {@inheritDoc}
	 */
    public void fireEvent(GwtEvent<?> event) {
        super.fireEvent(event);
    }

    /**
	 * {@inheritDoc}
	 */
    @Override
    protected void onTouchStart(TouchStartEvent evt) {
        if (evt.getTouches().length() != 1) return;
        Touch d = evt.getTouches().get(0);
        int e = d.getScreenX();
        int f = d.getScreenY();
        b.set(e, f);
        int g = d.getClientX();
        int h = d.getClientY();
        c.set(g, h);
        Document i = Document.get();
        NativeEvent j = i.createMouseOverEvent(0, e, f, g, h, false, false, false, false, 0, null);
        NativeEvent k = i.createMouseDownEvent(1, e, f, g, h, false, false, false, false, Event.BUTTON_LEFT);
        DomEvent.fireNativeEvent(j, this, evt.getRelativeElement());
        DomEvent.fireNativeEvent(k, this, evt.getRelativeElement());
        Element l = a == null ? null : a.getElement();
        if (l != null) {
            l.dispatchEvent(j);
            l.dispatchEvent(k);
        }
    }

    /**
	 * {@inheritDoc}
	 */
    @Override
    protected void onTouchMove(TouchMoveEvent evt) {
        if (evt.getTouches().length() != 1) return;
        Touch d = evt.getTouches().get(0);
        int e = d.getScreenX();
        int f = d.getScreenY();
        b.set(e, f);
        int g = d.getClientX();
        int h = d.getClientY();
        c.set(g, h);
        Document i = Document.get();
        NativeEvent j = i.createMouseMoveEvent(1, e, f, g, h, false, false, false, false, Event.BUTTON_LEFT);
        DomEvent.fireNativeEvent(j, this, evt.getRelativeElement());
        Element k = a == null ? null : a.getElement();
        if (k != null) k.dispatchEvent(j);
    }

    /**
	 * {@inheritDoc}
	 */
    @Override
    protected void onTouchEnd(TouchEndEvent evt) {
        if (evt.getTouches().length() != 0) return;
        int d = b.getX();
        int e = b.getY();
        int f = c.getX();
        int g = c.getY();
        Document h = Document.get();
        Element i = a == null ? null : a.getElement();
        NativeEvent j = h.createMouseUpEvent(1, d, e, f, g, false, false, false, false, Event.BUTTON_LEFT);
        NativeEvent k = h.createMouseOutEvent(0, d, e, f, g, false, false, false, false, 0, null);
        DomEvent.fireNativeEvent(j, this, evt.getRelativeElement());
        DomEvent.fireNativeEvent(k, this, evt.getRelativeElement());
        if (i != null) {
            i.dispatchEvent(j);
            i.dispatchEvent(k);
        }
    }

    /**
	 * Invoked when a touch is cancelled.
	 */
    @Override
    protected void onTouchCancel(TouchCancelEvent evt) {
        if (evt.getTouches().length() != 0) return;
        int d = b.getX();
        int e = b.getY();
        int f = c.getX();
        int g = c.getY();
        Document h = Document.get();
        NativeEvent i = h.createMouseUpEvent(1, d, e, f, g, false, false, false, false, Event.BUTTON_LEFT);
        NativeEvent j = h.createMouseOutEvent(0, d, e, f, g, false, false, false, false, 0, null);
        DomEvent.fireNativeEvent(i, this, evt.getRelativeElement());
        DomEvent.fireNativeEvent(j, this, evt.getRelativeElement());
        Element k = a == null ? null : a.getElement();
        if (k != null) {
            k.dispatchEvent(i);
            k.dispatchEvent(j);
        }
    }
}
