package com.maryanovsky.map.client;

import java.util.*;
import com.maryanovsky.gwtutils.client.MathUtils;
import com.maryanovsky.gwtutils.client.geom.Size;
import com.maryanovsky.gwtutils.client.geom.SizeView;
import com.maryanovsky.map.client.tiles.ImageTile;

/**
 * An implementation of tile layers in which each tile consists of a single
 * image retrieved from a URL. The URL for each tile is constructed from a
 * pattern; see {@link #ImageTileLayer(SizeView, String)} for more information.
 * 
 * @author Alexander Maryanovsky
 */
public class ImageTileLayer implements TileLayer {

    /***/
    private final SizeView a;

    /***/
    private final TileUrlPattern b;

    /**
	 * <p>Creates a new {@link ImageTileLayer} with the specified tile image
	 * size and tile URL pattern. The pattern is used to obtain a URL from the x
	 * index, y index and the zoom of a tile in the following manner:
	 * <ul>
	 *   <li>Each occurrence of the string <code>{x}</code> in the pattern is
	 *   replaced with the x index.</li>
	 *   <li>Each occurrence of the string <code>{y}</code> in the pattern is
	 *   replaced with the y index.</li>
	 *   <li>Each occurrence of the string <code>{zoom}</code> in the pattern is
	 *   replaced with the x index.</li>
	 *   <li>Each occurrence of a substring that starts with <code>[</code>
	 *   (square bracket open), ends with <code>]</code> (square bracket close),
	 *   and contains a list of comma separated strings is replaced with a
	 *   pseudorandomly chosen string from that list.</li>
	 * </ul></p>
	 * <p>The first three replacement rules allow specifying the positions of
	 * the x and y indices and the zoom value within the URL. The last rule
	 * allows specifying the different domains at which the tiles are available.
	 * This is commonly used to work around the browser limitation on the number
	 * of simultaneous connections to a single domain.</p>
	 * <p><strong>Example</strong>: the pattern
	 * <code>http://tile[1,2,3].mymaps.com/plain/{zoom}/{x}_{y}.png</code>
	 * could produce the URL
	 * <code>http://tile2.mymaps.com/plain/9/734_983.png</code>. 
	 */
    public ImageTileLayer(SizeView tileImageSize, String tileUrlPattern) {
        if (tileImageSize == null) throw new IllegalArgumentException("tileImageSize may not be null");
        this.a = new Size(tileImageSize);
        this.b = new TileUrlPattern(tileUrlPattern);
    }

    /**
	 * Returns the size of tiles.
	 */
    public SizeView getTileSize() {
        return a;
    }

    /**
	 * {@inheritDoc}
	 */
    public Tile getTile(int x, int y, int zoom) {
        int c = MathUtils.roundDownToMultiple(x, a.getWidth()) / a.getWidth();
        int d = MathUtils.roundDownToMultiple(y, a.getHeight()) / a.getHeight();
        String e = getTileUrl(c, d, zoom);
        if (e == null) return null;
        return ImageTile.makeImageTile(e);
    }

    /**
	 * {@inheritDoc}
	 */
    public void releaseTile(Tile tile) {
    }

    /**
	 * Returns the URL of the tile image at the specified zoom and indices, or
	 * <code>null</code> if none. Note that <code>xIndex</code> and
	 * <code>yIndex</code> are not pixel coordinates like in
	 * {@link TileLayer#getTile(int, int, int)}, but the indices of the tiles
	 * in their respective row and column at the tile grid of the specified
	 * zoom.
	 * The default implementation converts the three arguments to string
	 * representations via {@link String#valueOf(int)} and invokes
	 * {@link #getTileUrl(int, int, int)}. You should override this method if
	 * you need to perform some checks or possibly change the values of the
	 * parameters, or if you need a different representation of them to be
	 * inserted into the pattern.
	 */
    protected String getTileUrl(int xIndex, int yIndex, int zoom) {
        return getTileUrl(String.valueOf(xIndex), String.valueOf(yIndex), String.valueOf(zoom));
    }

    /**
	 * Returns the URL of the tile image at the specified zoom and indices.
	 * The arguments have already been checked and converted to their final
	 * string form. The default implementation applies them to the tile URL
	 * pattern and returns the result.
	 */
    protected String getTileUrl(String xArg, String yArg, String zoomArg) {
        return b.apply(xArg, yArg, zoomArg);
    }

    /***/
    private static class TileUrlPattern {

        /***/
        private final List<String> c;

        /***/
        private final List<List<String>> d;

        /***/
        private final List<String> e = new ArrayList<String>(1);

        /***/
        private final List<String> f = new ArrayList<String>(1);

        /***/
        private final List<String> g = new ArrayList<String>(1);

        /**
		 * Creates a new <code>TileUrlPattern</code> from the specified pattern.
		 */
        public TileUrlPattern(String pattern) {
            this.c = new LinkedList<String>();
            this.d = new LinkedList<List<String>>();
            int i = 0;
            int j = 0;
            int k = 0;
            while (true) {
                j = k;
                while ((j < pattern.length()) && ("{[".indexOf(pattern.charAt(j)) == -1)) j++;
                if (j == pattern.length()) {
                    c.add(pattern.substring(i));
                    break;
                }
                List<String> l = null;
                if (pattern.charAt(j) == '[') {
                    int m = pattern.indexOf(']', j);
                    if (m == -1) {
                        k++;
                        continue;
                    }
                    l = Arrays.asList(pattern.substring(j + 1, m).split(",", -1));
                    k = m + 1;
                } else if (pattern.startsWith("{x}", k)) {
                    l = e;
                    k = j + "{x}".length();
                } else if (pattern.startsWith("{y}", k)) {
                    l = f;
                    k = j + "{y}".length();
                } else if (pattern.startsWith("{zoom}", k)) {
                    l = g;
                    k = j + "{zoom}".length();
                } else {
                    k++;
                    continue;
                }
                c.add(pattern.substring(i, j));
                d.add(l);
                i = k;
                if (i >= pattern.length()) break;
            }
        }

        /**
		 * Applies the specified arguments to the pattern and returns the
		 * resulting URL.
		 */
        public synchronized String apply(String x, String y, String zoom) {
            e.add(x);
            f.add(y);
            g.add(zoom);
            StringBuilder i = new StringBuilder();
            Iterator<String> j = c.iterator();
            Iterator<List<String>> k = d.iterator();
            while (j.hasNext()) {
                i.append(j.next());
                if (k.hasNext()) {
                    List<String> l = k.next();
                    if (l.isEmpty()) continue;
                    String m = (l.size() == 1) ? l.get(0) : l.get(h(i.toString() + x + y + zoom, l.size()));
                    i.append(m);
                }
            }
            e.clear();
            f.clear();
            g.clear();
            return i.toString();
        }

        /***/
        private int h(String i, int j) {
            return Math.abs(i.hashCode() % j);
        }
    }
}
