package com.maryanovsky.map.client;

import com.maryanovsky.gwtutils.client.MathUtils;
import com.maryanovsky.gwtutils.client.Utils;
import com.maryanovsky.gwtutils.client.geom.DoublePointView;

/**
 * Represents a location on the surface of the Earth.
 * 
 * @author Maryanovsky Alexander
 */
public final class LatLng implements DoublePointView {

    /***/
    private final double a;

    /***/
    private final double b;

    /**
	 * Creates a new {@link LatLng} at the specified latitude and longitude.
	 * 
	 * @param latitude The latitude, in degrees, a value in the range [-90, 90].
	 * @param longitude The longitude, in degrees, a value in the range
	 * [-180, 180].
	 */
    public LatLng(double latitude, double longitude) {
        if ((latitude < -90) || (latitude > 90)) throw new IllegalArgumentException("Latitude out of range: " + latitude);
        if ((longitude < -180) || (longitude > 180)) throw new IllegalArgumentException("Longitude out of range " + longitude);
        this.a = latitude;
        this.b = longitude;
    }

    /**
	 * Creates a new {@link LatLng} object from the specified string, in the
	 * format of {@link #toUrlValue()}.
	 */
    public static LatLng fromUrlValue(String urlValue) {
        String[] c = urlValue.split(",");
        double d = Double.parseDouble(c[0]);
        double e = Double.parseDouble(c[1]);
        return new LatLng(d, e);
    }

    /**
	 * Returns the latitude, in degrees.
	 */
    public double getLatitude() {
        return a;
    }

    /**
	 * Returns the longitude, in degrees.
	 */
    public double getLongitude() {
        return b;
    }

    /**
	 * Returns the longitude, in degrees.
	 */
    public double getX() {
        return b;
    }

    /**
	 * Returns the latitude, in degrees.
	 */
    public double getY() {
        return a;
    }

    /**
	 * Returns a string representing this <code>LatLng</code> suitable for
	 * use as the value of a URL parameter. The returned value contains the
	 * latitude and longitude, with the specified number of digits after the
	 * decimal point, separated by a comma.  
	 */
    public String toUrlValue(int precision) {
        String c = MathUtils.doubleToFixedString(getLatitude(), precision);
        String d = MathUtils.doubleToFixedString(getLongitude(), precision);
        return c + "," + d;
    }

    /**
	 * Same as {@linkplain #toUrlValue(int)}, but with a fixed precision of 6
	 * digits after the decimal point.
	 */
    public String toUrlValue() {
        return toUrlValue(6);
    }

    /**
	 * Returns whether this <code>LatLng</code> represents the same location on
	 * the globe as the specified one. Note that the latitude and longitude
	 * values may be different while the location is still the same (for
	 * example, any two {@link LatLng} objects with latitude equal to 90 are
	 * equal).
	 */
    public boolean equals(Object object) {
        if (!(object instanceof LatLng)) return false;
        LatLng c = (LatLng) object;
        double d = c.getLatitude();
        double e = c.getLongitude();
        if (d == 90) return getLatitude() == 90; else if (d == -90) return getLatitude() == -90; else if (e == 180) return (getLatitude() == d) && ((getLongitude() == 180) || (getLongitude() == -180)); else return (getLatitude() == d) && (getLongitude() == e);
    }

    /**
	 * Returns the hashcode of this object.
	 */
    public int hashCode() {
        double c = getLatitude();
        double d = getLongitude();
        if (getLatitude() == 90) return 1; else if (getLatitude() == -90) return -1;
        if (d == -180) d = 180;
        return Utils.hashCode(Utils.hashCode(c), Utils.hashCode(d));
    }

    /**
	 * Returns a textual representation of this coordinate.
	 */
    public String toString() {
        return "[lat=" + getLatitude() + ", lng=" + getLongitude() + "]";
    }
}
