package com.maryanovsky.map.client;

/**
 * Represents an area bounded by minimum and maximum latitude and longitude
 * values (a rectangle, in geographical coordinates). The area can cross the 180
 * degree meridian.
 * 
 * @author Maryanovsky Alexander
 */
public class LatLngBounds {

    /***/
    private final LatLng a;

    /***/
    private final LatLng b;

    /**
	 * Creates a new {@link LatLngBounds} that represents an empty rectangle
	 * centered at the specified point.
	 */
    public LatLngBounds(LatLng point) {
        this(point, point);
    }

    /**
	 * Creates a new {@link LatLngBounds} that represents the rectangle bounded
	 * by the specified south-west and north-east points. Note that the latitude
	 * of <code>southWest</code> may not be larger than that of
	 * <code>northEast</code>, but its longitue may be, in which case the
	 * represented area will cross the 180 degree meridian.
	 */
    public LatLngBounds(LatLng southWest, LatLng northEast) {
        if (southWest == null) throw new IllegalArgumentException("southWest may not be null");
        if (northEast == null) throw new IllegalArgumentException("northEast may not be null");
        if (southWest.getLatitude() > northEast.getLatitude()) throw new IllegalArgumentException("The latitude of southWest may not be larger than that of northEast");
        this.a = southWest;
        this.b = northEast;
    }

    /**
	 * Creates a new {@link LatLngBounds} that represents the rectangle bounded
	 * by <code>point1</code> and <code>point2</code> and that contains the
	 * <code>include</code> point (there two are possible rectangles bounded by
	 * two points).
	 */
    public static LatLngBounds create(LatLng point1, LatLng point2, LatLng include) {
        LatLngBounds c = new LatLngBounds(point1, point2);
        if (!c.contains(include)) {
            double d = point1.getLongitude();
            double e = point2.getLongitude();
            point1 = new LatLng(point1.getLatitude(), d);
            point2 = new LatLng(point2.getLatitude(), e);
            c = new LatLngBounds(point1, point2);
        }
        return c;
    }

    /**
	 * Returns the south-western point.
	 */
    public LatLng getSouthWest() {
        return a;
    }

    /**
	 * Returns the north-eastern point.
	 */
    public LatLng getNorthEast() {
        return b;
    }

    /**
	 * Returns the center point.
	 */
    public LatLng getCenter() {
        double c = (a.getLatitude() + b.getLatitude()) / 2;
        double d = (a.getLongitude() + b.getLongitude()) / 2;
        if (crosses180Meridian()) {
            d = d + 180;
            if (d > 180) d -= 360;
        }
        return new LatLng(c, d);
    }

    /**
	 * Returns whether this {@link LatLngBounds} crosses the 180 meridian. This
	 * occurs when the longitude of the south-west point is larger than that of
	 * the north-east point.
	 */
    public boolean crosses180Meridian() {
        return a.getLongitude() > b.getLongitude();
    }

    /**
	 * Returns whether this {@link LatLngBounds} spans the entire planet
	 * latitudally.
	 */
    public boolean spansFullLatitude() {
        return getLatitudalSpan() == 180;
    }

    /**
	 * Returns whether this {@link LatLngBounds} spans the entire planet
	 * longitudally.
	 */
    public boolean spansFullLongitude() {
        return getLongitudalSpan() == 360;
    }

    /**
	 * Returns whether this {@link LatLngBounds} contains the specified point.
	 */
    public boolean contains(LatLng point) {
        double c = point.getLatitude();
        boolean d = (a.getLatitude() <= c) && (c <= b.getLatitude());
        if (!d) return false;
        double e = point.getLongitude();
        if (crosses180Meridian()) return (e >= a.getLongitude()) || (e <= b.getLongitude()); else return (e >= a.getLongitude()) && (e <= b.getLongitude());
    }

    /**
	 * Returns whether this {@link LatLngBounds} contains the specified one.
	 */
    public boolean contains(LatLngBounds bounds) {
        return contains(bounds.getSouthWest()) && contains(bounds.getNorthEast()) && (spansFullLongitude() || (crosses180Meridian() == bounds.crosses180Meridian()));
    }

    /**
	 * Returns the latitudal span of this {@link LatLngBounds}.
	 */
    public double getLatitudalSpan() {
        return b.getLatitude() - a.getLatitude();
    }

    /**
	 * Returns the longitudal span of this {@link LatLngBounds}.
	 */
    public double getLongitudalSpan() {
        double c = b.getLongitude() - a.getLongitude();
        if (crosses180Meridian()) c += 360;
        return c;
    }

    /**
	 * Returns a new {@link LatLngBounds} which extends the current one in the
	 * minimal possible way to include the specified point too.
	 */
    public LatLngBounds getExtended(LatLng point) {
        if (contains(point)) return this;
        double c = Math.min(a.getLatitude(), point.getLatitude());
        double d = Math.max(b.getLatitude(), point.getLatitude());
        LatLng e = new LatLng(c, point.getLongitude());
        LatLng f = new LatLng(c, a.getLongitude());
        LatLng g = new LatLng(d, point.getLongitude());
        LatLng h = new LatLng(d, b.getLongitude());
        LatLngBounds i = new LatLngBounds(e, h);
        LatLngBounds j = new LatLngBounds(f, g);
        return (i.getLongitudalSpan() < j.getLongitudalSpan()) ? i : j;
    }

    /**
	 * Returns a textual representation of this object.
	 */
    public String toString() {
        return "[sw=" + a + ", ne=" + b + "]";
    }
}
