package com.maryanovsky.map.client;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import com.maryanovsky.gwtutils.client.BrowserUtils;
import com.maryanovsky.gwtutils.client.geom.Point;
import com.maryanovsky.gwtutils.client.geom.PointView;
import com.maryanovsky.gwtutils.client.geom.Size;
import com.maryanovsky.gwtutils.client.geom.SizeView;

/**
 * Represents a map, which is a stack of {@linkplain TileLayer}s.
 * 
 * @author Maryanovsky Alexander
 */
public class Map {

    /***/
    private final Projection projection;

    /***/
    private final List<TileLayer> a;

    /***/
    private final int b;

    /***/
    private final int c;

    /**
	 * Creates a new {@link Map} with the specified projection, stack
	 * of tile layers and zoom range.
	 */
    public Map(Projection projection, List<TileLayer> tileLayers, int minZoom, int maxZoom) {
        if (projection == null) throw new IllegalArgumentException("projection may not be null");
        if (tileLayers == null) throw new IllegalArgumentException("tileLayers may not be null");
        if (tileLayers.isEmpty()) throw new IllegalArgumentException("tileLayers may not be empty");
        if (minZoom > maxZoom) throw new IllegalArgumentException("minZoom may not be greater than maxZoom");
        this.projection = projection;
        this.a = Collections.unmodifiableList(new ArrayList<TileLayer>(tileLayers));
        this.b = minZoom;
        this.c = maxZoom;
    }

    /**
	 * Creates a new {@link Map} with the specified projection, tile layer and
	 * zoom range.
	 */
    public Map(Projection projection, TileLayer tileLayer, int minZoom, int maxZoom) {
        this(projection, Collections.singletonList(tileLayer), minZoom, maxZoom);
    }

    /**
	 * Returns a variant of this <code>Map</code> properly scaled to be
	 * displayed on displays where the number of logical pixels per device pixel
	 * is not 1, such as iPhone retina displays. See also
	 * {@link BrowserUtils#getDevicePixelRatio()}. Note that if you invoke this
	 * method on an already-scaled <code>Map</code>, it will scale it again.
	 * <strong>Important:</strong> This method will throw an exception if not
	 * all the tile layers have the same tile size. 
	 */
    public Map getVersionScaledToDisplay() {
        double d = BrowserUtils.getDevicePixelRatio();
        if (d == 1.0) return this; else {
            SizeView e = getLayers().get(0).getTileSize();
            for (TileLayer f : getLayers()) {
                SizeView g = f.getTileSize();
                if ((e.getWidth() != g.getWidth()) || (e.getHeight() != g.getHeight())) throw new IllegalStateException("Scaled versions can't work properly when more than one tile size is used");
            }
            int f = (int) (e.getWidth() / d);
            int g = (int) (e.getHeight() / d);
            double h = ((double) e.getWidth()) / f;
            double i = ((double) e.getHeight()) / g;
            Projection j = new ScaledProjection(getProjection(), h, i);
            List<TileLayer> k = new ArrayList<TileLayer>(getLayers().size());
            for (TileLayer l : getLayers()) k.add(new ScaledTileLayer(l, h, i));
            return new Map(j, k, getMinZoom(), getMaxZoom());
        }
    }

    /**
	 * Returns the projection of this map.
	 */
    public Projection getProjection() {
        return projection;
    }

    /**
	 * Returns the stack of {linkplain TileLayer}s in this map. Layers at lower
	 * indices are behind layers at higher indices. The returned list is
	 * unmodifiable.
	 */
    public List<TileLayer> getLayers() {
        return a;
    }

    /**
	 * Returns the smallest available zoom for this <code>Map</code>.
	 */
    public int getMinZoom() {
        return b;
    }

    /**
	 * Returns the biggest available zoom for this <code>Map</code>.
	 */
    public int getMaxZoom() {
        return c;
    }

    /***/
    private static class ScaledProjection implements Projection {

        /***/
        private final Projection projection;

        /***/
        private final double xScale;

        /***/
        private final double yScale;

        /**
		 * Creates a new {@link ScaledProjection} with the specified delegate
		 * projection and scale.
		 */
        public ScaledProjection(Projection projection, double xScale, double yScale) {
            this.projection = projection;
            this.xScale = xScale;
            this.yScale = yScale;
        }

        /**
		 * {@inheritDoc}
		 */
        public Point fromLatLngToPixel(LatLng latLng, int zoom) {
            Point d = new Point(projection.fromLatLngToPixel(latLng, zoom));
            d.set((int) (d.getX() / xScale), (int) (d.getY() / yScale));
            return d;
        }

        /**
		 * {@inheritDoc}
		 */
        public LatLng fromPixelToLatLng(PointView pixel, int zoom) {
            Point d = new Point((int) (pixel.getX() * xScale), (int) (pixel.getY() * yScale));
            return projection.fromPixelToLatLng(d, zoom);
        }

        /**
		 * {@inheritDoc}
		 */
        public SizeView getWrapSize(int zoom) {
            Size d = new Size(projection.getWrapSize(zoom));
            d.set((int) (d.getWidth() / xScale), (int) (d.getHeight() / yScale));
            return d;
        }

        /**
		 * {@inheritDoc}
		 */
        public double getZoomMagnification(int startZoom, int endZoom) {
            return projection.getZoomMagnification(startZoom, endZoom);
        }
    }

    /***/
    private static class ScaledTileLayer implements TileLayer {

        /***/
        private final TileLayer d;

        /***/
        private final double xScale;

        /***/
        private final double yScale;

        /***/
        private final SizeView e;

        /**
		 * Creates a new <code>ScaledTileLayer</code> with the specified 
		 * requested scale.
		 */
        public ScaledTileLayer(TileLayer tileLayer, double xScale, double yScale) {
            this.d = tileLayer;
            this.xScale = xScale;
            this.yScale = yScale;
            Size f = new Size(tileLayer.getTileSize());
            f.set((int) (f.getWidth() / xScale), (int) (f.getHeight() / yScale));
            this.e = f;
        }

        /**
		 * {@inheritDoc} 
		 */
        public SizeView getTileSize() {
            return e;
        }

        /**
		 * {@inheritDoc}
		 */
        public Tile getTile(int x, int y, int zoom) {
            Tile f = d.getTile((int) (x * xScale), (int) (y * yScale), zoom);
            if (f == null) return null;
            f.setTileSizePixels(e);
            return f;
        }

        /**
		 * Releases any resources associated with the specified tile that was
		 * returned by {@link #getTile(int, int, int)}.
		 */
        public void releaseTile(Tile tile) {
            d.releaseTile(tile);
        }
    }
}
