package com.maryanovsky.map.client;

import com.maryanovsky.gwtutils.client.event.ChangeListener;
import com.maryanovsky.gwtutils.client.event.ChangeListenerList;
import com.maryanovsky.gwtutils.client.event.HasChange;

/**
 * Represents the location on the map displayed by a {@link MapWidget}: latitude
 * and longitude coordinates and a zoom.
 * 
 * @author Maryanovsky Alexander
 */
public class MapLocationModel implements HasChange {

    /***/
    private final int a;

    /***/
    private final int b;

    /***/
    private LatLng c;

    /***/
    private int d;

    /***/
    private final ChangeListenerList e = new ChangeListenerList();

    /**
	 * Creates a new <code>MapLocationModel</code> with the specified arguments.
	 * 
	 * @param minZoom The minimum possible zoom.
	 * @param maxZoom The maximum possible zoom.
	 * @param center The location of the center of the view.
	 * @param zoom The zoom of the view.
	 */
    public MapLocationModel(int minZoom, int maxZoom, LatLng center, int zoom) {
        if (minZoom > maxZoom) throw new IllegalArgumentException("minZoom must be less than or equal to maxZoom");
        this.a = minZoom;
        this.b = maxZoom;
        set(center, zoom);
    }

    /**
	 * Adds the specified <code>ChangeListener</code> to receive notifications
	 * when this <code>MapLocationModel</code> changes.
	 */
    public void addChangeListener(ChangeListener listener) {
        e.add(listener);
    }

    /**
	 * Removes the specified <code>ChangeListener</code> from receiving
	 * notifications when this <code>MapLocationModel</code> changes.
	 */
    public void removeChangeListener(ChangeListener listener) {
        e.remove(listener);
    }

    /***/
    private void fireStateChanged() {
        e.fireStateChanged(this);
    }

    /**
	 * Returns the smallest possible zoom in this model.
	 */
    public int getMinZoom() {
        return a;
    }

    /**
	 * Returns the biggest possible zoom in this model.
	 */
    public int getMaxZoom() {
        return b;
    }

    /**
	 * Returns whether the specified zoom is within the zoom bounds of this
	 * <code>MapLocationModel</code>.
	 */
    public boolean isLegalZoom(int zoom) {
        return (zoom >= a) && (zoom <= b);
    }

    /**
	 * Checks whether the specified zoom is within our zoom bounds of our
	 * and throws an <code>IllegalArgumentException</code> exception if not.
	 */
    public void zoomCheck(int zoom) {
        if (!isLegalZoom(zoom)) throw new IllegalArgumentException("Zoom " + zoom + " is outside the supported zoom bounds [" + a + "-" + b + "]");
    }

    /**
	 * Returns the nearest legal zoom to the specified one.
	 */
    public int clampToLegal(int zoom) {
        return Math.min(getMaxZoom(), Math.max(getMinZoom(), zoom));
    }

    /**
	 * Returns the location of the center.
	 */
    public LatLng getCenter() {
        return c;
    }

    /**
	 * Sets the center to the specified location.
	 */
    public void setCenter(LatLng center) {
        if (center.equals(this.c)) return;
        f(center);
        fireStateChanged();
    }

    /***/
    private void f(LatLng h) {
        if (h == null) throw new IllegalArgumentException("center may not be null");
        this.c = h;
    }

    /**
	 * Returns the zoom.
	 */
    public int getZoom() {
        return d;
    }

    /**
	 * Sets the zoom to the specified value.
	 */
    public void setZoom(int zoom) {
        if (this.d == zoom) return;
        g(zoom);
        fireStateChanged();
    }

    /***/
    private void g(int h) {
        zoomCheck(h);
        this.d = h;
    }

    /**
	 * Sets both the center and zoom in one call (to avoid a temporary state
	 * where either the zoom or the center are wrong).
	 */
    public void set(LatLng center, int zoom) {
        if (center.equals(this.c) && (this.d == zoom)) return;
        f(center);
        g(zoom);
        fireStateChanged();
    }

    /**
	 * Copies all relevant information from the specified
	 * <code>MapLocationModel</code> to this one.
	 */
    public void set(MapLocationModel mapLocationModel) {
        set(mapLocationModel.getCenter(), mapLocationModel.getZoom());
    }
}
