package com.maryanovsky.map.client;

import com.google.gwt.event.dom.client.HasAllMouseHandlers;
import com.google.gwt.event.dom.client.HasAllTouchHandlers;
import com.google.gwt.user.client.DOM;
import com.google.gwt.user.client.ui.HasHorizontalAlignment;
import com.google.gwt.user.client.ui.HasVerticalAlignment;
import com.google.gwt.user.client.ui.Widget;
import com.maryanovsky.gwtutils.client.event.PropertyChangeListener;
import com.maryanovsky.gwtutils.client.event.PropertyChangeListenerList;
import com.maryanovsky.gwtutils.client.geom.*;
import com.maryanovsky.gwtutils.client.ui.Alignment;
import com.maryanovsky.gwtutils.client.ui.UiUtils;
import com.maryanovsky.gwtutils.client.ui.mouse.gestures.drag.*;
import com.maryanovsky.gwtutils.client.ui.touch.gestures.mouse.TouchMouseEventsConverter;

/**
 * An overlay which positions a single widget at a specified location on the
 * map.
 * 
 * @author Maryanovsky Alexander
 */
public class WidgetOverlay implements Overlay {

    /**
	 * The name of the anchor property.
	 */
    public static final String ANCHOR_PROPERTY_NAME = "anchor";

    /**
	 * The name of the offset property.
	 */
    public static final String OFFSET_PROPERTY_NAME = "offset";

    /***/
    private final Widget a;

    /***/
    private LatLng b;

    /***/
    private final Point c = new Point();

    /***/
    private Alignment d = Alignment.BOTTOM_RIGHT;

    /***/
    private MapWidget e = null;

    /***/
    private PropertyChangeListenerList f = null;

    /***/
    private Object g = null;

    /***/
    private Size h = null;

    /***/
    private boolean i = true;

    /**
	 * Creates a new <code>WidgetOverlay</code> which places the specified
	 * widget at the specified offset from the specified anchor point. You may
	 * also pass the size of the widget, which can significantly reduce up the
	 * time it takes to add the overlay. If you pass <code>null</code> for the
	 * <code>widgetSize</code> parameter, it will be determined automatically,
	 * which, if there are many {@link WidgetOverlay}s, may take a long time.
	 * Additionally, if 
	 */
    public WidgetOverlay(Widget widget, LatLng anchor, PointView offset, SizeView widgetSize) {
        if (widget == null) throw new IllegalArgumentException("widget may not be null");
        if (anchor == null) throw new IllegalArgumentException("anchor may not be null");
        if (offset == null) throw new IllegalArgumentException("offset may not be null");
        this.a = widget;
        this.b = anchor;
        setOffset(offset);
        this.h = (widgetSize == null) ? null : new Size(widgetSize);
        widget.setVisible(i);
    }

    /**
	 * Creates a new <code>WidgetOverlay</code> which places the specified
	 * widget at the specified offset from the specified anchor point.
	 */
    public WidgetOverlay(Widget widget, LatLng anchor, PointView offset) {
        this(widget, anchor, offset, null);
    }

    /**
	 * Registers the specified property change listener to be notified when the
	 * specified property of this overlay changes.
	 */
    public void addPropertyChangeListener(String propertyName, PropertyChangeListener listener) {
        if (f == null) f = new PropertyChangeListenerList(new String[] { ANCHOR_PROPERTY_NAME, OFFSET_PROPERTY_NAME });
        f.add(propertyName, listener);
    }

    /**
	 * Unregisters the specified property change listener from being notified
	 * when the specified property of this overlay changes.
	 */
    public void removePropertyChangeListener(String propertyName, PropertyChangeListener listener) {
        if (f != null) f.remove(propertyName, listener);
    }

    /**
	 * Returns the widget this overlay adds to the map.
	 */
    public Widget getWidget() {
        return a;
    }

    /**
	 * Returns the anchor point at which the widget is added.
	 */
    public LatLng getAnchor() {
        return b;
    }

    /**
	 * Sets the anchor point at which the widget is added.
	 */
    public void setAnchor(LatLng anchor) {
        LatLng k = anchor;
        this.b = anchor;
        if (e != null) updatePosition();
        if (f != null) f.firePropertyChanged(this, ANCHOR_PROPERTY_NAME, k, anchor);
    }

    /**
	 * Sets the offset of the widget from the anchor point. 
	 */
    public void setOffset(PointView offset) {
        Point k = new Point(this.c);
        this.c.set(offset);
        if (e != null) updatePosition();
        if (f != null) f.firePropertyChanged(this, OFFSET_PROPERTY_NAME, k, offset);
    }

    /**
	 * Returns the current offset of the widget from the anchor point.
	 */
    public PointView getOffset() {
        return c;
    }

    /**
	 * Returns the alignment of the widget relative to the location specified by
	 * the anchor and offset.
	 */
    public Alignment getAlignment() {
        return d;
    }

    /**
	 * Sets the alignment of the widget relative to the location specified by
	 * the anchor and offset.
	 */
    public void setAlignment(Alignment alignment) {
        if (alignment == null) throw new IllegalArgumentException("alignment may not be null");
        this.d = alignment;
        if (e != null) updatePosition();
    }

    /**
	 * This method should be called when the widget's size changes. If you want
	 * the widget's size to be (re)calculated automatically, pass a
	 * <code>null</code> value.
	 */
    public void updateWidgetSize(SizeView widgetSize) {
        this.h = (widgetSize == null) ? null : new Size(widgetSize);
        updatePosition();
    }

    /**
	 * Adds the widget to the map's overlay panel.
	 */
    public void added(MapWidget mapWidget) {
        this.e = mapWidget;
        DOM.setStyleAttribute(a.getElement(), "position", "absolute");
        mapWidget.getOverlayPanel().add(a);
        updatePosition();
    }

    /**
	 * {@inheritDoc}
	 */
    public void updated(MapWidget mapWidget, boolean isTemporary) {
        updatePosition();
    }

    /**
	 * Updates the widget's position. Does nothing if it hasn't been added to
	 * a map widget yet.
	 */
    public void updatePosition() {
        if ((e == null) || !UiUtils.getVisibility(a) || !UiUtils.getVisibility(e.getOverlayPanel())) return;
        if (h == null) {
            int k = a.getOffsetWidth();
            int l = a.getOffsetHeight();
            if ((k != 0) && (l != 0)) h = new Size(k, l);
        }
        PointView k = j();
        if (h != null) {
            SizeView l = e.getMapSize();
            Rectangle m = new Rectangle(0, 0, l.getWidth(), l.getHeight());
            Rectangle n = new Rectangle(k, h);
            boolean o = n.intersects(m);
            if (o != i) {
                i = o;
                a.setVisible(i);
            }
        }
        if (i) UiUtils.setTopLeft(a, k);
    }

    /***/
    private PointView j() {
        Point k = e.mapToLocal(b);
        k.moveBy(c.getX(), c.getY());
        if (h != null) {
            int l = h.getWidth();
            int m = h.getHeight();
            Alignment n = getAlignment();
            HasHorizontalAlignment.HorizontalAlignmentConstant o = n.getHorizontalAlignment();
            HasVerticalAlignment.VerticalAlignmentConstant p = n.getVerticalAlignment();
            if (o == HasHorizontalAlignment.ALIGN_CENTER) k.moveBy(-l / 2, 0); else if (o == HasHorizontalAlignment.ALIGN_LEFT) k.moveBy(-l, 0);
            if (p == HasVerticalAlignment.ALIGN_MIDDLE) k.moveBy(0, -m / 2); else if (p == HasVerticalAlignment.ALIGN_TOP) k.moveBy(0, -m);
        }
        return k;
    }

    /**
	 * Removes the widget from the map's overlay panel.
	 */
    public void removed(MapWidget mapWidget) {
        mapWidget.getOverlayPanel().remove(a);
        this.e = null;
        this.h = null;
    }

    /**
	 * Makes the overlay draggable (via the specified drag handle) using the
	 * mouse in such a way that dragging it adjusts the offset, but not the
	 * anchor.
	 * <code>updateContinuously</code> specifies whether the offset will be
	 * updated as the overlay is dragged or only at the end of the drag gesture.
	 */
    public void makeOffsetDraggableByMouse(HasAllMouseHandlers dragHandle, final boolean updateContinuously) {
        if (g != null) UiUtils.disableDragging(g);
        g = UiUtils.makeDraggable(a, dragHandle, new HandlesAllDragEvents() {

            private Point initialOffset = null;

            public void onDragStart(DragStartEvent evt) {
                initialOffset = new Point(getOffset());
            }

            public void onDragMove(DragMoveEvent evt) {
                if (updateContinuously) updateOffset(evt.getOffset());
            }

            public void onDragEnd(DragEndEvent evt) {
                updateOffset(evt.getOffset());
            }

            public void onDragCancel(DragCancelEvent evt) {
            }

            private void updateOffset(PointView k) {
                initialOffset.moveBy(k.getX(), k.getY());
                setOffset(initialOffset);
                initialOffset.moveBy(-k.getX(), -k.getY());
            }
        });
    }

    /**
	 * Makes the overlay draggable (via the specified drag handle) using the
	 * mouse in such a way that dragging it adjusts its anchor, but not the
	 * offset.
	 * <code>updateContinuously</code> specifies whether the anchor will be
	 * updated as the overlay is dragged or only at the end of the drag gesture.
	 */
    public void makeAnchorDraggableByMouse(HasAllMouseHandlers dragHandle, final boolean updateContinuously) {
        if (g != null) UiUtils.disableDragging(g);
        g = UiUtils.makeDraggable(a, dragHandle, new HandlesAllDragEvents() {

            private LatLng initialAnchor = null;

            public void onDragStart(DragStartEvent evt) {
                initialAnchor = getAnchor();
            }

            public void onDragMove(DragMoveEvent evt) {
                if (updateContinuously) updateAnchor(evt.getOffset());
            }

            public void onDragEnd(DragEndEvent evt) {
                updateAnchor(evt.getOffset());
            }

            public void onDragCancel(DragCancelEvent evt) {
            }

            private void updateAnchor(PointView k) {
                int l = e.getMapLocationModel().getZoom();
                Projection m = e.getMap().getProjection();
                Point n = new Point(m.fromLatLngToPixel(initialAnchor, l));
                n.moveBy(k);
                setAnchor(m.fromPixelToLatLng(n, l));
            }
        });
    }

    /**
	 * Makes the overlay draggable (via the specified drag handle) by touch in
	 * such a way that dragging it adjusts the offset, but not the anchor.
	 * <code>updateContinuously</code> specifies whether the offset will be
	 * updated as the overlay is dragged or only at the end of the drag gesture.
	 */
    public void makeOffsetDraggableByTouch(HasAllTouchHandlers dragHandle, final boolean updateContinuously) {
        makeOffsetDraggableByMouse(new TouchMouseEventsConverter(dragHandle, null, dragHandle), updateContinuously);
    }

    /**
	 * Makes the overlay draggable (via the specified drag handle) by touch
	 * in such a way that dragging it adjusts its anchor, but not the offset.
	 * <code>updateContinuously</code> specifies whether the anchor will be
	 * updated as the overlay is dragged or only at the end of the drag gesture.
	 */
    public void makeAnchorDraggableByTouch(HasAllTouchHandlers dragHandle, final boolean updateContinuously) {
        makeAnchorDraggableByMouse(new TouchMouseEventsConverter(dragHandle, null, dragHandle), updateContinuously);
    }

    /**
	 * Makes the overlay non-draggable.
	 */
    public void makeNonDraggable() {
        if (g != null) UiUtils.disableDragging(g);
    }
}
