package com.maryanovsky.map.client.actions.mouse;

import com.google.gwt.dom.client.NativeEvent;
import com.google.gwt.event.dom.client.MouseDownEvent;
import com.google.gwt.event.dom.client.MouseEvent;
import com.google.gwt.event.dom.client.MouseUpEvent;
import com.google.gwt.event.dom.client.MouseWheelEvent;
import com.google.gwt.event.shared.EventHandler;
import com.maryanovsky.gwtutils.client.Condition;
import com.maryanovsky.gwtutils.client.ui.mouse.gestures.SyntheticMouseGestureEvent;
import com.maryanovsky.gwtutils.client.ui.mouse.gestures.click.ClickEvent;
import com.maryanovsky.gwtutils.client.ui.mouse.gestures.click.ClickHandler;
import com.maryanovsky.gwtutils.client.ui.mouse.gestures.click.ClickRecognizer;
import com.maryanovsky.gwtutils.client.ui.mouse.gestures.drag.DragRecognizer;
import com.maryanovsky.gwtutils.client.ui.mouse.gestures.drag.DragStartEvent;
import com.maryanovsky.gwtutils.client.ui.mouse.gestures.drag.DragStartHandler;
import com.maryanovsky.gwtutils.client.ui.mouse.gestures.wheelscroll.WheelScrollDownEvent;
import com.maryanovsky.gwtutils.client.ui.mouse.gestures.wheelscroll.WheelScrollDownHandler;
import com.maryanovsky.gwtutils.client.ui.mouse.gestures.wheelscroll.WheelScrollRecognizer;
import com.maryanovsky.gwtutils.client.ui.mouse.gestures.wheelscroll.WheelScrollUpEvent;
import com.maryanovsky.gwtutils.client.ui.mouse.gestures.wheelscroll.WheelScrollUpHandler;
import com.maryanovsky.map.client.MapWidget;
import com.maryanovsky.map.client.actions.StandardActions;

/**
 * Implementats the standard actions which allow the user to interact with the
 * map on a regular desktop computer, utilizing a mouse/keyboard.
 * 
 * @author Maryanovsky Alexander
 */
public class StandardMouseActions extends StandardActions {

    /***/
    private static final Condition<ClickEvent> a = new Condition<ClickEvent>() {

        public boolean eval(ClickEvent evt) {
            return evt.getClickCount() % 2 == 0;
        }
    };

    /***/
    private static <T1 extends EventHandler, T2 extends MouseEvent<? extends EventHandler>> Condition<SyntheticMouseGestureEvent<T1, T2>> b() {
        return new Condition<SyntheticMouseGestureEvent<T1, T2>>() {

            public boolean eval(SyntheticMouseGestureEvent<T1, T2> argument) {
                MouseEvent<?> d = argument.getSourceEvent();
                return !(d.isAltKeyDown() || d.isControlKeyDown() || d.isShiftKeyDown() || d.isMetaKeyDown());
            }
        };
    }

    /***/
    private static <T1 extends EventHandler, T2 extends MouseEvent<? extends EventHandler>> Condition<SyntheticMouseGestureEvent<T1, T2>> c(final int d) {
        return new Condition<SyntheticMouseGestureEvent<T1, T2>>() {

            public boolean eval(SyntheticMouseGestureEvent<T1, T2> argument) {
                MouseEvent<?> e = argument.getSourceEvent();
                return e.getNativeButton() == d;
            }
        };
    }

    /**
	 * Adds an action to the specified map widget allowing the user to pan the
	 * map by dragging it with the left mouse button.
	 * Returns the {@link DragRecognizer} responsible for recognizing the drag
	 * gestures.
	 */
    public static DragRecognizer addDragPan(MapWidget mapWidget) {
        Condition<DragStartEvent> d = Condition.and(StandardMouseActions.<DragStartHandler, MouseDownEvent>b(), StandardMouseActions.<DragStartHandler, MouseDownEvent>c(NativeEvent.BUTTON_LEFT));
        DragRecognizer e = new DragRecognizer(mapWidget, mapWidget.getMouseEventsSource(), d);
        new DragPanAction(mapWidget).registerWith(e);
        return e;
    }

    /**
	 * Adds an action to the specified map widget allowing the user to pan the
	 * map to a specified location by right-clicking it. The transition is
	 * optionally animated.
	 * Returns the {@link ClickRecognizer} responsible for recognizing the click
	 * gestures.
	 */
    public static ClickRecognizer addRightClickPan(MapWidget mapWidget, boolean isAnimated) {
        Condition<ClickEvent> d = Condition.and(StandardMouseActions.<ClickHandler, MouseUpEvent>b(), StandardMouseActions.<ClickHandler, MouseUpEvent>c(NativeEvent.BUTTON_RIGHT));
        ClickRecognizer e = new ClickRecognizer(mapWidget, mapWidget.getMouseEventsSource(), d);
        new ClickPanAction(mapWidget, isAnimated).registerWith(e);
        return e;
    }

    /**
	 * Adds an action to the specified map widget allowing the user to zoom-in
	 * and recenter on a location by double-clicking it. The transition is
	 * optionally animated.
	 * Returns the {@link ClickRecognizer} responsible for recognizing the click
	 * gestures.
	 */
    public static ClickRecognizer addDoubleClickZoomIn(MapWidget mapWidget, boolean isAnimated) {
        Condition<ClickEvent> d = Condition.and(StandardMouseActions.<ClickHandler, MouseUpEvent>b(), StandardMouseActions.<ClickHandler, MouseUpEvent>c(NativeEvent.BUTTON_LEFT), a);
        ClickRecognizer e = new ClickRecognizer(mapWidget, mapWidget.getMouseEventsSource(), d);
        new ClickZoomInAction(mapWidget, isAnimated).registerWith(e);
        return e;
    }

    /**
	 * Adds an action to the specified map widget allowing the user to zoom in
	 * or out using his mouse wheel. The transition is optionally animated.
	 * Returns the {@link WheelScrollRecognizer} responsible for recognizing the
	 * wheel-scroll gestures.
	 */
    public static WheelScrollRecognizer addWheelZoom(MapWidget mapWidget, boolean isAnimated) {
        Condition<WheelScrollUpEvent> d = Condition.cast(StandardMouseActions.<WheelScrollUpHandler, MouseWheelEvent>b());
        Condition<WheelScrollDownEvent> e = Condition.cast(StandardMouseActions.<WheelScrollDownHandler, MouseWheelEvent>b());
        WheelScrollRecognizer f = new WheelScrollRecognizer(mapWidget, mapWidget.getMouseEventsSource(), d, e);
        new WheelScrollZoomAction(mapWidget, isAnimated).registerWith(f);
        return f;
    }

    /**
	 * Adds an action to the specified map widget allowing the user to select
	 * the area he wants to zoom-in on by dragging a rectangle on the map while
	 * holding down the SHIFT key. Optionally, the rectangle is forced to the
	 * same width-to-height ratio as map widget itself. The transition is
	 * optionally animated.
	 * Returns the {@link DragRecognizer} responsible for recognizing the drag
	 * gestures.
	 */
    public static DragRecognizer addShiftDragAreaSelection(MapWidget mapWidget, boolean isKeepRatio, boolean isAnimated) {
        Condition<DragStartEvent> d = new Condition<DragStartEvent>() {

            public boolean eval(DragStartEvent evt) {
                return evt.getSourceEvent().isShiftKeyDown();
            }
        };
        Condition<DragStartEvent> e = Condition.and(d, StandardMouseActions.<DragStartHandler, MouseDownEvent>c(NativeEvent.BUTTON_LEFT));
        DragRecognizer f = new DragRecognizer(mapWidget, mapWidget.getMouseEventsSource(), e);
        new AreaSelectionAction(mapWidget, isKeepRatio, isAnimated).registerWith(f);
        return f;
    }

    /**
	 * Adds all the standard actions to the specified map widget, with
	 * optionally animated transitions.
	 */
    public static void addAll(MapWidget mapWidget, boolean isAnimated) {
        addDragPan(mapWidget);
        addRightClickPan(mapWidget, isAnimated);
        addDoubleClickZoomIn(mapWidget, isAnimated);
        addWheelZoom(mapWidget, isAnimated);
        addShiftDragAreaSelection(mapWidget, true, isAnimated);
    }

    /**
	 * Adds all the standard actions to the specified map widget.
	 */
    @Override
    public void addAll(MapWidget mapWidget) {
        addAll(mapWidget, true);
    }
}
