package com.maryanovsky.map.client.actions.touch;

import com.google.gwt.event.shared.HandlerRegistration;
import com.maryanovsky.gwtutils.client.MathUtils;
import com.maryanovsky.gwtutils.client.event.MultipleHandlerRegistration;
import com.maryanovsky.gwtutils.client.geom.Point;
import com.maryanovsky.gwtutils.client.ui.Animation;
import com.maryanovsky.gwtutils.client.ui.UiUtils;
import com.maryanovsky.gwtutils.client.ui.touch.gestures.drag.*;
import com.maryanovsky.map.client.MapWidget;
import com.maryanovsky.map.client.actions.MapAction;

/**
 * Pans the map as the user drags his finger across the map widget, continuing
 * the panning after the user lifted his finger, as if the map had inertia over
 * a surface with friction.
 * 
 * @author Maryanovsky Alexander
 */
public class InertialDragPanAction extends MapAction<HasAllDragHandlers> implements DragStartHandler, DragMoveHandler, DragEndHandler, DragCancelHandler {

    /***/
    private static final int a = 50;

    /***/
    private final int b;

    /***/
    private DecelerationAnimation c = null;

    /***/
    private Point d = new Point(0, 0);

    /***/
    private long e = 0;

    /***/
    private Point f = new Point(0, 0);

    /***/
    private long g = 0;

    /**
	 * Creates a new {@link InertialDragPanAction} for the specified map widget
	 * with the specified deceleration duration, in milliseconds.
	 * 
	 * @param mapWidget The target map widget.
	 * @param decelerationDuration The time it takes the map to decelerate to
	 * full stop, in milliseconds.
	 */
    public InertialDragPanAction(MapWidget mapWidget, int decelerationDuration) {
        super(mapWidget);
        if (decelerationDuration < 0) throw new IllegalArgumentException("decelerationDuration must be non-negative");
        this.b = decelerationDuration;
    }

    /**
	 * Registers for handling drag events from the specified source.
	 */
    public HandlerRegistration registerWith(HasAllDragHandlers dragSource) {
        return new MultipleHandlerRegistration(dragSource.addDragStartHandler(this), dragSource.addDragMoveHandler(this), dragSource.addDragEndHandler(this), dragSource.addDragCancelHandler(this));
    }

    /**
	 * {@inheritDoc}
	 */
    public void onDragStart(DragStartEvent evt) {
        if ((c != null) && c.isRunning()) c.end();
        d.set(0, 0);
        f.set(0, 0);
        e = System.currentTimeMillis();
        g = e;
    }

    /**
	 * {@inheritDoc}
	 */
    public void onDragMove(DragMoveEvent evt) {
        if (UiUtils.getVisibility(mapWidget.getOverlayPanel())) UiUtils.setVisibility(mapWidget.getOverlayPanel(), false);
        f.set(d);
        d.set(evt.getOffset());
        g = e;
        e = System.currentTimeMillis();
        mapWidget.setTemporaryOffset(evt.getOffset().getOpposite());
    }

    /**
	 * {@inheritDoc}
	 */
    public void onDragEnd(DragEndEvent evt) {
        mapWidget.setTemporaryOffset(evt.getOffset().getOpposite());
        long h = System.currentTimeMillis();
        int i = (int) (h - g);
        int j = d.getX() - f.getX();
        int k = d.getY() - f.getY();
        boolean l = false;
        int m;
        if (i == 0) m = 1000; else m = (int) (Math.sqrt(j * j + k * k) * 1000 / i);
        if (m > 1000) m = 1000;
        double n;
        if ((j == 0) && (k == 0)) {
            n = 0;
            l = true;
        } else n = MathUtils.atan2(-k, -j);
        if (!l && (m >= a)) {
            c = new DecelerationAnimation(b, m, n);
            c.start();
        } else {
            UiUtils.setVisibility(mapWidget.getOverlayPanel(), true);
            mapWidget.applyTemporaryOffset();
        }
    }

    /**
	 * {@inheritDoc}
	 */
    public void onDragCancel(DragCancelEvent evt) {
        UiUtils.setVisibility(mapWidget.getOverlayPanel(), true);
        mapWidget.setTemporaryOffset(new Point(0, 0));
    }

    /***/
    private class DecelerationAnimation extends Animation {

        /***/
        private final int h;

        /***/
        private double i;

        /**
		 * Creates a new {@link DecelerationAnimation} with the specified
		 * duration (in milliseconds), initial speed (in pixels per second) and 
		 * angle.
		 */
        public DecelerationAnimation(int duration, int speed, double angle) {
            super(duration);
            this.h = speed;
            this.i = angle;
        }

        /**
		 * {@inheritDoc}
		 */
        @Override
        protected void animationStart() {
        }

        /**
		 * {@inheritDoc}
		 */
        @Override
        protected void nextFrame() {
            int j = getDuration();
            double k = h * 1000 / j;
            double l = getLinearProgress() * j / 1000;
            int m = (int) ((h * l - k * l * l / 2) * Math.cos(i));
            int n = (int) ((h * l - k * l * l / 2) * Math.sin(i));
            mapWidget.setTemporaryOffset(new Point(m, n).moveBy(d.getOpposite()));
        }

        /**
		 * {@inheritDoc}
		 */
        @Override
        protected void animationEnd() {
            UiUtils.setVisibility(mapWidget.getOverlayPanel(), true);
            mapWidget.applyTemporaryOffset();
        }
    }
}
