package com.maryanovsky.map.client.actions.touch;

import com.maryanovsky.gwtutils.client.Condition;
import com.maryanovsky.gwtutils.client.ui.touch.gestures.drag.DragRecognizer;
import com.maryanovsky.gwtutils.client.ui.touch.gestures.mftap.MultiFingerTapEvent;
import com.maryanovsky.gwtutils.client.ui.touch.gestures.mftap.MultiFingerTapRecognizer;
import com.maryanovsky.gwtutils.client.ui.touch.gestures.pinch.PinchRecognizer;
import com.maryanovsky.gwtutils.client.ui.touch.gestures.tap.TapEvent;
import com.maryanovsky.gwtutils.client.ui.touch.gestures.tap.TapRecognizer;
import com.maryanovsky.map.client.MapWidget;
import com.maryanovsky.map.client.actions.StandardActions;

/**
 * Implementats the standard actions which allow the user to interact with the
 * map on an iPhone.
 * 
 * @author Maryanovsky Alexander
 */
public class StandardMultiTouchActions extends StandardActions {

    /**
	 * Adds an action to the specified map widget which will zoom in on the
	 * point the user double-tapped on. The zoom-in transition is optionally
	 * animated.
	 * Returns the {@link TapRecognizer} responsible for recognizing the
	 * double-tap gestures.
	 */
    public static TapRecognizer addDoubleTapZoomInAction(MapWidget mapWidget, boolean isAnimated) {
        Condition<TapEvent> a = new Condition<TapEvent>() {

            public boolean eval(TapEvent evt) {
                return evt.getTapCount() == 2;
            }
        };
        TapRecognizer b = new TapRecognizer(mapWidget, mapWidget.getTouchEventsSource(), a);
        new TapZoomInAction(mapWidget, isAnimated).registerWith(b);
        return b;
    }

    /**
	 * Adds an action to the specified map widget which will zoom out when the
	 * user taps with two fingers. The zoom-out transition is optionally
	 * animated.
	 * Returns the {@link MultiFingerTapRecognizer} responsible for recognizing
	 * the two-finger tap.
	 */
    public static MultiFingerTapRecognizer addTwoFingerTapZoomOutAction(MapWidget mapWidget, boolean isAnimated) {
        Condition<MultiFingerTapEvent> a = new Condition<MultiFingerTapEvent>() {

            public boolean eval(MultiFingerTapEvent evt) {
                return evt.getFingerCount() == 2;
            }
        };
        MultiFingerTapRecognizer b = new MultiFingerTapRecognizer(mapWidget, mapWidget.getTouchEventsSource(), a);
        new MultiFingerTapZoomOutAction(mapWidget, isAnimated).registerWith(b);
        return b;
    }

    /**
	 * Adds an action to the specified map widget which will pan the map when
	 * the user drags his finger. Returns the {@link DragRecognizer}
	 * responsible for recognizing the drag gesture.
	 */
    public static DragRecognizer addDragPanAction(MapWidget mapWidget) {
        DragRecognizer a = new DragRecognizer(mapWidget, mapWidget.getTouchEventsSource());
        new InertialDragPanAction(mapWidget, 250).registerWith(a);
        return a;
    }

    /**
	 * Adds an action to the specified map widget which will magnify and zoom
	 * the map when the user pinches the map widget. The final zoom transition
	 * is optionally animated.
	 * Returns the {@link PinchRecognizer} responsible for recognizing the pinch
	 * gesture.
	 */
    public static PinchRecognizer addPinchMagnifyAction(MapWidget mapWidget, boolean isAnimated) {
        PinchRecognizer a = new PinchRecognizer(mapWidget, mapWidget.getTouchEventsSource());
        new PinchMagnifyAction(mapWidget, isAnimated).registerWith(a);
        return a;
    }

    /**
	 * Adds all the standard actions, with optionally animated transitions, to
	 * the specified map widget.
	 */
    public static void addAll(MapWidget mapWidget, boolean isAnimated) {
        addDoubleTapZoomInAction(mapWidget, isAnimated);
        addTwoFingerTapZoomOutAction(mapWidget, isAnimated);
        addDragPanAction(mapWidget);
        addPinchMagnifyAction(mapWidget, isAnimated);
    }

    /**
	 * Adds all the standard actions to the specified map widget.
	 */
    public void addAll(MapWidget mapWidget) {
        addAll(mapWidget, true);
    }
}
