package com.maryanovsky.map.client.opengis.wms;

import com.google.gwt.xml.client.Element;
import com.maryanovsky.gwtutils.client.text.ParseException;
import com.maryanovsky.gwtutils.client.text.ParseUtils;

/**
 * Encapsulates a bounding box as defined by the
 * <a href="http://portal.opengeospatial.org/files/?artifact_id=1081&version=1&format=pdf">
 * Web Map Service Implementation Specification</a>. Note that this class is
 * used for LatLonBoundingBox as well, with the SRS always set to "EPSG:4326"
 * and unspecified resolution values.
 * 
 * @author Maryanovsky Alexander
 */
public final class BoundingBox {

    /***/
    private final String a;

    /***/
    private final double b;

    /***/
    private final double c;

    /***/
    private final double d;

    /***/
    private final double e;

    /***/
    private final double f;

    /***/
    private final double g;

    /**
	 * Creates a new {@link BoundingBox} from the specified
	 * <code>BoundingBox</code> or <code>LatLonBoundingBox</code> element that
	 * defines it.
	 */
    public BoundingBox(Element element) throws ParseException {
        String h = element.getTagName();
        if (!("LatLonBoundingBox".equals(h) || "BoundingBox".equals(h))) throw new ParseException("Element tag name is \"" + h + "\";" + "must be either \"LatLonBoundingBox\" or \"BoundingBox\"");
        boolean i = "LatLonBoundingBox".equals(h);
        String j = i ? "EPSG:4326" : ParseUtils.getMandatoryAttribute(element, "SRS");
        this.a = j;
        this.b = ParseUtils.getMandatoryDoubleAttribute(element, "minx");
        this.c = ParseUtils.getMandatoryDoubleAttribute(element, "miny");
        this.d = ParseUtils.getMandatoryDoubleAttribute(element, "maxx");
        this.e = ParseUtils.getMandatoryDoubleAttribute(element, "maxy");
        this.f = ParseUtils.getOptionalDoubleAttribute(element, "resx", 0);
        this.g = ParseUtils.getOptionalDoubleAttribute(element, "resy", 0);
    }

    /**
	 * Retuns the spatial reference system in which this box is defined.
	 */
    public String getSrs() {
        return a;
    }

    /**
	 * Returns the minimum X coordinate.
	 */
    public double getMinX() {
        return b;
    }

    /**
	 * Returns the minimum Y coordinate.
	 */
    public double getMinY() {
        return c;
    }

    /**
	 * Returns the maximum X coordinate.
	 */
    public double getMaxX() {
        return d;
    }

    /**
	 * Returns the maximum Y coordinate.
	 */
    public double getMaxY() {
        return e;
    }

    /**
	 * Returns the resolution on the X axis; 0 if unspecified.
	 */
    public double getResX() {
        return f;
    }

    /**
	 * Returns the resolution on the Y axis; 0 if unspecified.
	 */
    public double getResY() {
        return g;
    }
}
