package com.maryanovsky.map.client.opengis.wms;

import java.util.*;
import com.google.gwt.http.client.RequestBuilder;
import com.google.gwt.http.client.RequestException;
import com.google.gwt.xml.client.Document;
import com.google.gwt.xml.client.Element;
import com.maryanovsky.gwtutils.client.geom.Size;
import com.maryanovsky.gwtutils.client.geom.SizeView;
import com.maryanovsky.gwtutils.client.net.ResultCallback;
import com.maryanovsky.gwtutils.client.net.UrlBuilder;
import com.maryanovsky.gwtutils.client.net.XmlResponseHandler;
import com.maryanovsky.gwtutils.client.text.ParseException;
import com.maryanovsky.gwtutils.client.text.ParseUtils;

/**
 * <p>Represents the capabilities of a Web Map Service as defined by the
 * <a href="http://portal.opengeospatial.org/files/?artifact_id=1081&version=1&format=pdf">
 * Web Map Service Implementation Specification</a>. Additionally, tile sets are
 * defined by the <a href="http://wiki.osgeo.org/wiki/WMS_Tiling_Client_Recommendation">
 * WMS Tiling Client Recommendation</a>.
 * 
 * <p>Note that the javadoc comments in this class are short, and unsufficient
 * for adequate documentation. To understand the various classes and their
 * properties you will need to read the specs mentioned above.
 * 
 * @author Maryanovsky Alexander
 */
public class Capabilities {

    /***/
    private final Document a;

    /***/
    private final String name;

    /***/
    private final String title;

    /***/
    private final String abztract;

    /***/
    private final List<String> keywordList;

    /***/
    private final String url;

    /***/
    private final ContactInformation b;

    /***/
    private final String c;

    /***/
    private final String d;

    /***/
    private final GetMapRequestInfo e;

    /***/
    private final Layer f;

    /***/
    private final Map<String, Layer> g;

    /***/
    private final List<TileSet> h;

    /**
	 * Parses the specified Capabilities response XML document and creates a
	 * {@link Capabilities} object from it.
	 */
    public Capabilities(Document document) throws ParseException {
        if (document == null) throw new IllegalArgumentException("document may not be null");
        Wms.checkVersionCompatible(document);
        Element l = document.getDocumentElement();
        ParseUtils.checkTagName(l, "WMT_MS_Capabilities");
        Element m = ParseUtils.getMandatoryNamedElement(l, "Service");
        this.a = document;
        this.name = ParseUtils.getMandatoryElementTextContent(m, "Name");
        this.title = ParseUtils.getMandatoryElementTextContent(m, "Title");
        this.abztract = ParseUtils.getOptionalElementTextContent(m, "Abstract", null);
        Element n = ParseUtils.getOptionalNamedElement(m, "KeywordList");
        this.keywordList = n == null ? Collections.<String>emptyList() : Collections.unmodifiableList(Arrays.asList(ParseUtils.getElementListTextContent(n, "Keyword")));
        this.url = j(m, true);
        Element o = ParseUtils.getOptionalNamedElement(m, "ContactInformation");
        this.b = o == null ? null : new ContactInformation(o);
        this.c = ParseUtils.getOptionalElementTextContent(m, "Fees", null);
        this.d = ParseUtils.getOptionalElementTextContent(m, "AccessConstraints", null);
        Element p = ParseUtils.getMandatoryNamedElement(l, "Capability");
        Element q = ParseUtils.getMandatoryNamedElement(p, "Request");
        Element r = ParseUtils.getMandatoryNamedElement(q, "GetMap");
        this.e = new GetMapRequestInfo(r);
        Element s = ParseUtils.getOptionalNamedElement(p, "Layer");
        this.f = s == null ? null : new Layer(s, null);
        this.g = Collections.unmodifiableMap(k(f, new HashMap<String, Layer>()));
        Element t = ParseUtils.getOptionalNamedElement(p, "VendorSpecificCapabilities");
        if (t == null) this.h = Collections.<TileSet>emptyList(); else {
            List<Element> u = ParseUtils.getNamedChildElements(t, "TileSet");
            List<TileSet> v = new ArrayList<TileSet>(u.size());
            for (Element w : u) v.add(new TileSet(w));
            this.h = Collections.unmodifiableList(v);
        }
    }

    /***/
    private static String i(Element l, String m) throws ParseException {
        Element n = ParseUtils.getMandatoryNamedElement(l, "HTTP");
        Element o = ParseUtils.getMandatoryNamedElement(n, m);
        return j(o, true);
    }

    /***/
    private static String j(Element l, boolean m) throws ParseException {
        Element n = ParseUtils.getNamedChildElement(l, "OnlineResource", m);
        return n == null ? null : ParseUtils.getMandatoryAttribute(n, "xlink:href");
    }

    /***/
    private static Map<String, Layer> k(Layer l, Map<String, Layer> m) {
        if (l.getName() != null) m.put(l.getName(), l);
        List<Layer> n = l.getChildren();
        for (Layer o : n) k(o, m);
        return m;
    }

    /**
	 * Sends an HTTP GET GetCapabilities request using the specified base URL.
	 * The response, a {@link Capabilities} object, is delivered to the
	 * specified {@link ResultCallback}.
	 */
    public static void sendGetRequest(String baseUrl, ResultCallback<Capabilities> callback) throws RequestException {
        String l = makeGetRequestUrl(baseUrl);
        RequestBuilder m = new RequestBuilder(RequestBuilder.GET, l);
        m.sendRequest(null, new ResponseHandler(l, callback));
    }

    /**
	 * Returns the URL of a full HTTP GET GetCapabilities request using the
	 * specified base URL. 
	 */
    public static String makeGetRequestUrl(String baseUrl) {
        UrlBuilder l = new UrlBuilder(baseUrl);
        l.addParam("version", Wms.VERSION);
        l.addParam("service", "WMS");
        l.addParam("request", "GetCapabilities");
        return l.getUrl();
    }

    /**
	 * Returns the full Capabilities response XML document from which this
	 * {@link Capabilities} object was created. Through this object you can
	 * retrieve any additional information sent that this class does not parse. 
	 */
    public Document getDocument() {
        return a;
    }

    /**
	 * Returns the name of the service.
	 */
    public String getName() {
        return name;
    }

    /**
	 * Returns the title of the service.
	 */
    public String getTitle() {
        return title;
    }

    /**
	 * Returns a short description of the service. May be <code>null</code>.
	 */
    public String getAbstract() {
        return abztract;
    }

    /**
	 * Returns a list of keywords pertaining to the service. The returned list
	 * is unmodifiable.
	 */
    public List<String> getKeywordList() {
        return keywordList;
    }

    /**
	 * Returns the URL of the service. May be <code>null</code>.
	 */
    public String getUrl() {
        return url;
    }

    /**
	 * Returns the information of the contact person for the service. May be
	 * <code>null</code>. 
	 */
    public ContactInformation getContactInformation() {
        return b;
    }

    /**
	 * Returns a description of the fees accociated with the service. May be
	 * <code>null</code>.
	 */
    public String getFees() {
        return c;
    }

    /**
	 * Returns any access constraints regarding the use of the service. May be
	 * <code>null</code>.
	 */
    public String getAccessConstraints() {
        return d;
    }

    /**
	 * Returns information specifying how to make a <code>GetMap</code> request
	 * to the service.
	 */
    public GetMapRequestInfo getGetMapRequestInfo() {
        return e;
    }

    /**
	 * Returns the toplevel layer provided by this service. May be
	 * <code>null</code>.
	 */
    public Layer getTopLevelLayer() {
        return f;
    }

    /**
	 * Returns a Map of layer names to layers. The returned map is unmodifiable.
	 * Note that only named layers appear here.
	 */
    public Map<String, Layer> getLayerByName() {
        return g;
    }

    /**
	 * Returns the list of {@link TileSet}s supported by the service. The
	 * returned list is unmodifiable. 
	 */
    public List<TileSet> getTileSets() {
        return h;
    }

    /**
	 * Returns a list of {@link TileSet}s which display the specified layer. The
	 * list will be sorted in increasing order by the number of layers displayed
	 * by the tile set. The returned list is unmodifiable.
	 */
    public List<TileSet> getTileSetsByLayerName(String layerName) {
        if (layerName == null) throw new IllegalArgumentException("layerName may not be null");
        List<TileSet> l = new ArrayList<TileSet>();
        for (TileSet m : h) {
            String[] n = m.getLayers().split(",");
            for (String o : n) {
                if (layerName.equals(o)) {
                    l.add(m);
                    break;
                }
            }
        }
        Collections.sort(l, new Comparator<TileSet>() {

            public int compare(TileSet tileSet1, TileSet tileSet2) {
                return tileSet1.getLayers().split(",").length - tileSet2.getLayers().split(",").length;
            }
        });
        return Collections.unmodifiableList(l);
    }

    /**
	 * The base class for various classes which specify a URL and the format
	 * (mime-type) of the document at that URL.
	 */
    public static class UrlWithFormat {

        /***/
        private final String url;

        /***/
        private final String l;

        /**
		 * Creates a new {@link UrlWithFormat} object from the element that
		 * defines it. The element must have <code>Format</code> and
		 * <code>OnlineResource</code> child elements.
		 */
        protected UrlWithFormat(Element element) throws ParseException {
            this.url = j(element, true);
            this.l = ParseUtils.getMandatoryElementTextContent(element, "Format");
        }

        /**
		 * Returns the URL of the document.
		 */
        public final String getUrl() {
            return url;
        }

        /**
		 * Returns the format (mime-type) of the document at the URL.
		 */
        public final String getFormat() {
            return l;
        }
    }

    /**
	 * The class for URLs of images, which also specify the format (mime-type)
	 * and size of the image. 
	 */
    public static final class ImageUrl extends UrlWithFormat {

        /***/
        private final SizeView l;

        /**
		 * Creates an {@link ImageUrl} from the specified element that defines
		 * it. The element, in addition to the requirements of
		 * {@link UrlWithFormat}'s constructor, must have <code>width</code> and
		 * <code>height</code> attributes.
		 */
        private ImageUrl(Element m) throws ParseException {
            super(m);
            int n = ParseUtils.getMandatoryIntAttribute(m, "width");
            int o = ParseUtils.getMandatoryIntAttribute(m, "height");
            this.l = new Size(n, o);
        }

        /**
		 * Returns the size of the image.
		 */
        public SizeView getImageSize() {
            return l;
        }
    }

    /**
	 * Encapsulates the contact information sent as part of a
	 * <code>Capabilities</code> response.
	 */
    public static final class ContactInformation {

        /***/
        private final String l;

        /***/
        private final String m;

        /***/
        private final String n;

        /***/
        private final Address address;

        /***/
        private final String o;

        /***/
        private final String p;

        /***/
        private final String q;

        /**
		 * Creates a new {@link ContactInformation} object from the specified
		 * <code>ContactInformation</code> response element that defines it.
		 */
        private ContactInformation(Element r) throws ParseException {
            ParseUtils.checkTagName(r, "ContactInformation");
            Element s = ParseUtils.getOptionalNamedElement(r, "ContactPersonPrimary");
            this.l = s == null ? null : ParseUtils.getMandatoryElementTextContent(s, "ContactPerson");
            this.m = s == null ? null : ParseUtils.getMandatoryElementTextContent(s, "ContactOrganization");
            this.n = ParseUtils.getOptionalElementTextContent(r, "ContactPosition", null);
            Element t = ParseUtils.getOptionalNamedElement(r, "ContactAddress");
            this.address = t == null ? null : new Address(t);
            this.o = ParseUtils.getOptionalElementTextContent(r, "ContactVoiceTelephone", null);
            this.p = ParseUtils.getOptionalElementTextContent(r, "ContactFacsimileTelephone", null);
            this.q = ParseUtils.getOptionalElementTextContent(r, "ElectronicMailAddress", null);
        }

        /**
		 * Returns the contact person. May be <code>null</code>.
		 */
        public String getPerson() {
            return l;
        }

        /**
		 * Returns the contact organization. May be <code>null</code>.
		 */
        public String getOrganization() {
            return m;
        }

        /**
		 * Returns the position title of the contact person. May be
		 * <code>null</code>.
		 */
        public String getPosition() {
            return n;
        }

        /**
		 * Returns the address of the contact. May be <code>null</code>.
		 */
        public Address getAddress() {
            return address;
        }

        /**
		 * Returns the voice telephone number of the contact. May be
		 * <code>null</code>.
		 */
        public String getPhone() {
            return o;
        }

        /**
		 * Returns the facsimile telephone number of the contact. May be
		 * <code>null</code>.
		 */
        public String getFax() {
            return p;
        }

        /**
		 * Returns the electronic mail address of the contact. May be
		 * <code>null</code>.
		 */
        public String getEmail() {
            return q;
        }

        /**
		 * Encapsulates the contact address.
		 */
        public static final class Address {

            /***/
            private final String r;

            /***/
            private final String address;

            /***/
            private final String s;

            /***/
            private final String t;

            /***/
            private final String u;

            /***/
            private final String v;

            /**
			 * Creates a new {@link Address} object from the specified
			 * <code>ContactAddress</code> element that defines it.
			 */
            private Address(Element w) throws ParseException {
                ParseUtils.checkTagName(w, "ContactAddress");
                this.r = ParseUtils.getMandatoryElementTextContent(w, "AddressType");
                this.address = ParseUtils.getMandatoryElementTextContent(w, "Address");
                this.s = ParseUtils.getMandatoryElementTextContent(w, "City");
                this.t = ParseUtils.getMandatoryElementTextContent(w, "StateOrProvince");
                this.u = ParseUtils.getMandatoryElementTextContent(w, "PostCode");
                this.v = ParseUtils.getMandatoryElementTextContent(w, "Country");
            }

            /**
			 * Returns the address type.
			 */
            public String getAddressType() {
                return r;
            }

            /**
			 * Returns the street address.
			 */
            public String getAddress() {
                return address;
            }

            /**
			 * Returns the city.
			 */
            public String getCity() {
                return s;
            }

            /**
			 * Returns the state or province.
			 */
            public String getStateOrProvince() {
                return t;
            }

            /**
			 * Returns the zip or postal code.
			 */
            public String getPostCode() {
                return u;
            }

            /**
			 * Returns the country.
			 */
            public String getCountry() {
                return v;
            }
        }
    }

    /**
	 * Encapsulates information about the GetMap capabilities of the service.
	 */
    public static final class GetMapRequestInfo {

        /***/
        private final List<String> l;

        /***/
        private final String m;

        /**
		 * Creates a new {@link GetMapRequestInfo} from the specified <code>GetMap</code>
		 * element that defines it.
		 */
        private GetMapRequestInfo(Element n) throws ParseException {
            this.l = Collections.unmodifiableList(Arrays.asList(ParseUtils.getElementListTextContent(n, "Format")));
            this.m = i(ParseUtils.getMandatoryNamedElement(n, "DCPType"), "Get");
        }

        /**
		 * Returns the list of supported image formats (mime types). The
		 * returned list is unmodifiable.
		 */
        public List<String> getImageFormats() {
            return l;
        }

        /**
		 * Returns the prefix of the URLs for retrieving map images via a GetMap
		 * request.
		 */
        public String getUrlPrefix() {
            return m;
        }
    }

    /**
	 * Encapsulates information about a map layer.
	 */
    public static final class Layer {

        /***/
        private final Layer l;

        /***/
        private final boolean m;

        /***/
        private final int n;

        /***/
        private final boolean o;

        /***/
        private final boolean p;

        /***/
        private final SizeView q;

        /***/
        private final String name;

        /***/
        private final String title;

        /***/
        private final String abztract;

        /***/
        private final List<String> keywordList;

        /***/
        private final List<String> r;

        /***/
        private final BoundingBox s;

        /***/
        private final Map<String, BoundingBox> t;

        /***/
        private final List<Dimension> u;

        /***/
        private final Map<String, Extent> v;

        /***/
        private final Attribution w;

        /***/
        private final Map<String, String> x;

        /***/
        private final List<Identifier> y;

        /***/
        private final List<MetadataUrl> z;

        /***/
        private final List<UrlWithFormat> aa;

        /***/
        private final List<UrlWithFormat> ab;

        /***/
        private final List<Style> styles;

        /***/
        private final ScaleHint ac;

        /***/
        private final List<Layer> ad;

        /**
		 * Creates a new {@link Layer} from the specified <code>Layer</code>
		 * element that defines it and with the specified parent layer
		 * (<code>null</code> if the created layer is a toplevel one).
		 */
        private Layer(Element ae, Layer af) throws ParseException {
            this.l = af;
            this.m = ParseUtils.getOptionalIntAttribute(ae, "queryable", 0) == 1;
            this.n = ParseUtils.getOptionalIntAttribute(ae, "cascaded", 0);
            this.o = ParseUtils.getOptionalIntAttribute(ae, "opaque", 0) == 1;
            this.p = ParseUtils.getOptionalIntAttribute(ae, "noSubsets", 0) == 0;
            int ag = ParseUtils.getOptionalIntAttribute(ae, "fixedWidth", 0);
            int ah = ParseUtils.getOptionalIntAttribute(ae, "fixedHeight", 0);
            this.q = ag == 0 ? null : new Size(ag, ah);
            this.name = ParseUtils.getOptionalElementTextContent(ae, "Name", null);
            this.title = ParseUtils.getMandatoryElementTextContent(ae, "Title");
            this.abztract = ParseUtils.getOptionalElementTextContent(ae, "Abstract", null);
            Element ai = ParseUtils.getOptionalNamedElement(ae, "KeywordList");
            this.keywordList = ai == null ? Collections.<String>emptyList() : Collections.unmodifiableList(Arrays.asList(ParseUtils.getElementListTextContent(ai, "Keyword")));
            List<Element> aj = ParseUtils.getNamedChildElements(ae, "SRS");
            String[] ak = (aj.size() == 1) ? ParseUtils.getMandatoryElementTextContent(ae, "SRS").split(" ") : ParseUtils.getElementListTextContent(ae, "SRS");
            this.r = Collections.unmodifiableList(Arrays.asList(ak));
            Element al = ParseUtils.getOptionalNamedElement(ae, "LatLonBoundingBox");
            this.s = al == null ? null : new BoundingBox(al);
            List<Element> am = ParseUtils.getNamedChildElements(ae, "BoundingBox");
            Map<String, BoundingBox> an = new HashMap<String, BoundingBox>();
            for (Element ao : am) {
                BoundingBox ap = new BoundingBox(ao);
                an.put(ap.getSrs(), ap);
            }
            this.t = Collections.unmodifiableMap(an);
            List<Element> ao = ParseUtils.getNamedChildElements(ae, "Dimension");
            List<Dimension> ap = new ArrayList<Dimension>(ao.size());
            for (Element aq : ao) ap.add(new Dimension(aq));
            this.u = Collections.unmodifiableList(ap);
            List<Element> aq = ParseUtils.getNamedChildElements(ae, "Extent");
            Map<String, Extent> ar = new HashMap<String, Extent>();
            for (Element as : aq) {
                Extent at = new Extent(as);
                ar.put(at.getName().toLowerCase(), at);
            }
            this.v = Collections.unmodifiableMap(ar);
            Element as = ParseUtils.getOptionalNamedElement(ae, "Attribution");
            this.w = as == null ? null : new Attribution(as);
            List<Element> at = ParseUtils.getNamedChildElements(ae, "AuthorityURL");
            Map<String, String> au = new HashMap<String, String>();
            for (Element av : at) {
                String aw = ParseUtils.getMandatoryAttribute(av, "name");
                String ax = j(av, true);
                au.put(aw, ax);
            }
            this.x = Collections.unmodifiableMap(au);
            List<Element> av = ParseUtils.getNamedChildElements(ae, "Identifier");
            List<Identifier> aw = new ArrayList<Identifier>(av.size());
            for (Element ax : av) aw.add(new Identifier(ax));
            this.y = Collections.unmodifiableList(aw);
            List<Element> ax = ParseUtils.getNamedChildElements(ae, "MetadataURL");
            List<MetadataUrl> ay = new ArrayList<MetadataUrl>(ax.size());
            for (Element az : ax) ay.add(new MetadataUrl(az));
            this.z = Collections.unmodifiableList(ay);
            List<Element> az = ParseUtils.getNamedChildElements(ae, "DataURL");
            List<UrlWithFormat> ba = new ArrayList<UrlWithFormat>(az.size());
            for (Element bb : az) ba.add(new UrlWithFormat(bb));
            this.aa = Collections.unmodifiableList(ba);
            List<Element> bb = ParseUtils.getNamedChildElements(ae, "FeatureListURL");
            List<UrlWithFormat> bc = new ArrayList<UrlWithFormat>(bb.size());
            for (Element bd : bb) bc.add(new UrlWithFormat(bd));
            this.ab = Collections.unmodifiableList(bc);
            List<Element> bd = ParseUtils.getNamedChildElements(ae, "Style");
            List<Style> be = new ArrayList<Style>(bd.size());
            for (Element bf : bd) be.add(new Style(bf));
            this.styles = Collections.unmodifiableList(be);
            Element bf = ParseUtils.getOptionalNamedElement(ae, "ScaleHint");
            this.ac = bf == null ? null : new ScaleHint(bf);
            List<Element> bg = ParseUtils.getNamedChildElements(ae, "Layer");
            List<Layer> bh = new ArrayList<Layer>(bg.size());
            for (Element bi : bg) bh.add(new Layer(bi, this));
            this.ad = Collections.unmodifiableList(bh);
        }

        /**
		 * Returns the parent layer; <code>null</code> if this is a toplevel
		 * layer.
		 */
        public Layer getParent() {
            return l;
        }

        /**
		 * Returns whether this layer is queryable.
		 */
        public boolean isQueryable() {
            return m;
        }

        /**
		 * Returns the cascaded count.
		 */
        public int getCascadedCount() {
            return n;
        }

        /**
		 * Returns whether this layer is opaque.
		 */
        public boolean isOpaque() {
            return o;
        }

        /**
		 * Returns whether this layer is subsettable.
		 */
        public boolean isSubsettable() {
            return p;
        }

        /**
		 * If this layer has a fixed size, in pixels, returns it; otherwise,
		 * returns <code>null</code>.
		 */
        public SizeView getFixedSize() {
            return q;
        }

        /**
		 * Returns the name of this layer; <code>null</code> if the layer is
		 * unnamed.
		 */
        public String getName() {
            return name;
        }

        /**
		 * Returns the title of this layer.
		 */
        public String getTitle() {
            return title;
        }

        /**
		 * Returns the abstract (a short description) of this layer.
		 */
        public String getAbstract() {
            return abztract;
        }

        /**
		 * Returns a list of keywords pertaining to this layer. The returned
		 * list is unmodifiable.
		 */
        public List<String> getKeywordList() {
            return keywordList;
        }

        /**
		 * Returns the list of SRSes for this layer. The returned list is
		 * unmodifiable.
		 */
        public List<String> getSrses() {
            return r;
        }

        /**
		 * Returns the LatLonBoundingBox of this layer. 
		 */
        public BoundingBox getLatLonBoundingBox() {
            return s;
        }

        /**
		 * Returns a map of SRSes to bounding boxes of this layer in that SRS.
		 * The returned map is unmodifiable.
		 */
        public Map<String, BoundingBox> getSrsToBoundingBoxMap() {
            return t;
        }

        /**
		 * Returns the list of Dimensions for this layer. The returned list is
		 * unmodifiable.
		 */
        public List<Dimension> getDimensions() {
            return u;
        }

        /**
		 * Returns a map from dimension names in lowercase to the corresponding
		 * Extent instances. The returned map is unmodifiable.
		 */
        public Map<String, Extent> getDimensionNamesToExtentsMap() {
            return v;
        }

        /**
		 * Returns the Attribution of this layer; may be <code>null</code>.
		 */
        public Attribution getAttribution() {
            return w;
        }

        /**
		 * Returns a map from Authority names to Authority URLs. The returned
		 * map is unmodifiable.
		 */
        public Map<String, String> getAuthorityNamesToUrlsMap() {
            return x;
        }

        /**
		 * Returns the list of Identifiers for this layer. The returned list is
		 * unmodifiable.
		 */
        public List<Identifier> getIdentifiers() {
            return y;
        }

        /**
		 * Returns the list of MetadataURLs for this layer. The returned list is
		 * unmodifiable.
		 */
        public List<MetadataUrl> getMetadataUrls() {
            return z;
        }

        /**
		 * Returns the list of DataURLs for this layer. The returned list is
		 * unmodifiable.
		 */
        public List<UrlWithFormat> getDataUrls() {
            return aa;
        }

        /**
		 * Returns the list of FeatureListURLs for this layer. The returned list
		 * is unmodifiable.
		 */
        public List<UrlWithFormat> getFeatureListUrls() {
            return ab;
        }

        /**
		 * Returns the list of styles for this layer. The returned list is
		 * unmodifiable.
		 */
        public List<Style> getStyles() {
            return styles;
        }

        /**
		 * Returns the ScaleHint for this layer; <code>null</code> if none.
		 */
        public ScaleHint getScaleHint() {
            return ac;
        }

        /**
		 * Returns the list of child layers of this layer. The returned list is
		 * unmodifiable. 
		 */
        public List<Layer> getChildren() {
            return ad;
        }

        /**
		 * Encapsulates a Dimension.
		 */
        public static final class Dimension {

            /***/
            private final String name;

            /***/
            private final String ae;

            /***/
            private final String af;

            /**
			 * Creates a new {@link Dimension} from the specified
			 * <code>Dimension</code> element that defines it.
			 */
            private Dimension(Element ag) throws ParseException {
                this.name = ParseUtils.getMandatoryAttribute(ag, "name");
                this.ae = ParseUtils.getMandatoryAttribute(ag, "units");
                this.af = ParseUtils.getOptionalAttribute(ag, "unitSymbol", null);
            }

            /**
			 * Returns the name of the Dimension.
			 */
            public String getName() {
                return name;
            }

            /**
			 * Returns the units of the Dimension.
			 */
            public String getUnits() {
                return ae;
            }

            /**
			 * Returns the unit symbol of the Dimension units. <code>null</code>
			 * if unspecified.
			 */
            public String getUnitSymbol() {
                return af;
            }
        }

        /**
		 * Encapsulates the Extent of a Dimension.
		 */
        public static final class Extent {

            /***/
            private final String name;

            /***/
            private final String ae;

            /***/
            private final boolean af;

            /***/
            private final boolean ag;

            /***/
            private boolean ah;

            /***/
            private final String value;

            /**
			 * Creates a new {@link Extent} from the specified
			 * <code>Extent</code> element that defines it.
			 */
            private Extent(Element ai) throws ParseException {
                this.name = ParseUtils.getMandatoryAttribute(ai, "name");
                this.ae = ParseUtils.getOptionalAttribute(ai, "default", null);
                this.af = ParseUtils.getOptionalAttribute(ai, "multipleValues", "0").equals("1");
                this.ag = ParseUtils.getOptionalAttribute(ai, "nearestValues", "0").equals("1");
                this.ah = ParseUtils.getOptionalAttribute(ai, "current", "0").equals("1");
                this.value = ParseUtils.getTextContent(ai);
            }

            /**
			 * Returns the name of the Extent.
			 */
            public String getName() {
                return name;
            }

            /**
			 * Returns the default value of the Extent; may be
			 * <code>null</code>.
			 */
            public String getDefaultValue() {
                return ae;
            }

            /**
			 * Returns whether the client may specify multiple values in the
			 * request.
			 */
            public boolean isMulipleValuesAllowed() {
                return af;
            }

            /**
			 * Returns whether the server will round off imprecise values.
			 */
            public boolean isNearestValues() {
                return ag;
            }

            /**
			 * Returns whether the temporal data are normally kept current and
			 * the request parameter TIME may include the keyword "current"
			 * instead of an ending value. Only valid for Time extents.
			 */
            public boolean isCurrent() {
                return ah;
            }

            /**
			 * Returns the value of the Extent. This is the unparsed content of
			 * the Extent tag. See table C.1 in the spec. for a description of
			 * the format.
			 */
            public String getValue() {
                return value;
            }
        }

        /**
		 * Encapsulates the Attribution of a layer.
		 */
        public static final class Attribution {

            /***/
            private final String title;

            /***/
            private final String url;

            /***/
            private final ImageUrl ae;

            /**
			 * Creates a new {@link Attribution} from the specified
			 * <code>Attribution</code> element that defines it.
			 */
            private Attribution(Element af) throws ParseException {
                this.title = ParseUtils.getOptionalElementTextContent(af, "Title", null);
                this.url = j(af, false);
                Element ag = ParseUtils.getOptionalNamedElement(af, "LogoURL");
                this.ae = ag == null ? null : new ImageUrl(ag);
            }

            /**
			 * Returns a short description of the provider; may be
			 * <code>null</code>.
			 */
            public String getTitle() {
                return title;
            }

            /**
			 * Returns the URL of the provider; may be <code>null</code>.
			 */
            public String getUrl() {
                return url;
            }

            /**
			 * Returns the URL of the provider's logo; may be <code>null</code>.
			 */
            public ImageUrl getLogoUrl() {
                return ae;
            }
        }

        /**
		 * Encapsulates an Identifier of a layer.
		 */
        public static final class Identifier {

            /***/
            private final String ae;

            /***/
            private final String value;

            /**
			 * Creates a new {@link Identifier} from the specified
			 * <code>Identifier</code> element that defines it.
			 */
            private Identifier(Element af) throws ParseException {
                this.ae = ParseUtils.getMandatoryAttribute(af, "authority");
                this.value = ParseUtils.getTextContent(af);
            }

            /**
			 * Returns the name of the authority that defines this identifier.
			 */
            public String getAuthorityName() {
                return ae;
            }

            /**
			 * Returns the value of the identifier.
			 */
            public String getValue() {
                return value;
            }
        }

        /**
		 * Encapsulates a Metadata URL regarding a layer.
		 */
        public static final class MetadataUrl extends UrlWithFormat {

            /***/
            private final String ae;

            /**
			 * Creates a new {@link MetadataUrl} from the specified
			 * <code>MetadataURL</code> element that defines it.
			 */
            private MetadataUrl(Element af) throws ParseException {
                super(af);
                this.ae = ParseUtils.getMandatoryAttribute(af, "type");
            }

            /**
			 * Returns the standard to which the metadata complies.
			 */
            public String getType() {
                return ae;
            }
        }

        /**
		 * Encapsulates a Style for a layer.
		 */
        public static final class Style {

            /***/
            private final String name;

            /***/
            private final String title;

            /***/
            private final String abztract;

            /***/
            private final List<ImageUrl> ae;

            /***/
            private final UrlWithFormat af;

            /***/
            private final UrlWithFormat ag;

            /**
			 * Creates a new {@link Style} from the specified
			 * <code>Style</code> element that defines it.
			 */
            private Style(Element ah) throws ParseException {
                this.name = ParseUtils.getMandatoryElementTextContent(ah, "Name");
                this.title = ParseUtils.getMandatoryElementTextContent(ah, "Title");
                this.abztract = ParseUtils.getOptionalElementTextContent(ah, "Abstract", null);
                List<Element> ai = ParseUtils.getNamedChildElements(ah, "LegendURL");
                List<ImageUrl> aj = new ArrayList<ImageUrl>(ai.size());
                for (Element ak : ai) aj.add(new ImageUrl(ak));
                this.ae = Collections.unmodifiableList(aj);
                Element ak = ParseUtils.getOptionalNamedElement(ah, "StyleSheetURL");
                this.af = ak == null ? null : new UrlWithFormat(ak);
                Element al = ParseUtils.getOptionalNamedElement(ah, "StyleURL");
                this.ag = al == null ? null : new UrlWithFormat(al);
            }

            /**
			 * Returns the name of the style. . This is meant for inter-computer
			 * communication, not for humans.
			 */
            public String getName() {
                return name;
            }

            /**
			 * Returns the title of the style. This is meant to be displayed to
			 * humans.
			 */
            public String getTitle() {
                return title;
            }

            /**
			 * Returns a short description of the style. May be
			 * <code>null</code>.
			 */
            public String getAbstract() {
                return abztract;
            }

            /**
			 * Returns a list of URLs of legend images. The returned list is
			 * unmodifiable.
			 */
            public List<ImageUrl> getLegendUrls() {
                return ae;
            }

            /**
			 * Returns the style sheet URL; may be <code>null</code>.
			 */
            public UrlWithFormat getStyleSheetUrl() {
                return af;
            }

            /**
			 * Returns the style URL; may be <code>null</code>.
			 */
            public UrlWithFormat getStyleUrl() {
                return ag;
            }
        }

        /**
		 * Encapsulates a ScaleHint; a hint of minimum and maximum scales at
		 * which it is appropriate to display a layer.
		 */
        public static final class ScaleHint {

            /***/
            private final double ae;

            /***/
            private final double af;

            /**
			 * Creates a new {@link ScaleHint} from the specified
			 * <code>ScaleHint</code> element that defines it.
			 */
            private ScaleHint(Element ag) throws ParseException {
                this.ae = ParseUtils.getMandatoryDoubleAttribute(ag, "min");
                this.af = ParseUtils.getMandatoryDoubleAttribute(ag, "max");
            }

            /**
			 * Returns the minimum scale at which it is appropriate to display a
			 * layer.
			 */
            public double getMinimum() {
                return ae;
            }

            /**
			 * Returns the maximum scale at which it is appropriate to display a
			 * layer.
			 */
            public double getMaximum() {
                return af;
            }
        }
    }

    /**
	 * Encapsulates information about a tile set as defined by the
	 * <a href="http://wiki.osgeo.org/wiki/WMS_Tiling_Client_Recommendation">
	 * WMS Tiling Client Recommendation</a>. 
	 */
    public static final class TileSet {

        /***/
        private final String l;

        /***/
        private final BoundingBox m;

        /***/
        private final List<Double> n;

        /***/
        private final Size o;

        /***/
        private final String p;

        /***/
        private final String q;

        /***/
        private final String styles;

        /**
		 * Creates a new {@link TileSet} from the specified <code>TileSet</code>
		 * element that defines it.
		 */
        private TileSet(Element r) throws ParseException {
            if (!"TileSet".equals(r.getTagName())) throw new ParseException("Element tag name is " + r.getTagName() + ";" + " must be \"TileSet\"");
            this.l = ParseUtils.getMandatoryElementTextContent(r, "SRS");
            this.m = new BoundingBox(ParseUtils.getMandatoryNamedElement(r, "BoundingBox"));
            String[] s = ParseUtils.getMandatoryElementTextContent(r, "Resolutions").split(" ");
            List<Double> t = new ArrayList<Double>(s.length);
            try {
                for (String u : s) t.add(new Double(u));
            } catch (NumberFormatException u) {
                throw new ParseException(u.getMessage());
            }
            this.n = Collections.unmodifiableList(t);
            try {
                int v = Integer.parseInt(ParseUtils.getMandatoryElementTextContent(r, "Width"));
                int w = Integer.parseInt(ParseUtils.getMandatoryElementTextContent(r, "Height"));
                this.o = new Size(v, w);
            } catch (NumberFormatException v) {
                throw new ParseException(v.getMessage());
            }
            this.p = ParseUtils.getMandatoryElementTextContent(r, "Format");
            this.q = ParseUtils.getMandatoryElementTextContent(r, "Layers");
            this.styles = ParseUtils.getMandatoryElementTextContent(r, "Styles");
        }

        /**
		 * Returns the spatial reference system (SRS) of this tile set.
		 */
        public String getSrs() {
            return l;
        }

        /**
		 * Returns the bounding box of this tile set.
		 */
        public BoundingBox getBoundingBox() {
            return m;
        }

        /**
		 * Returns the list of resolutions (map units per pixel) at which this
		 * tile set has tiles. The returned list is unmodifable. 
		 */
        public List<Double> getResolutions() {
            return n;
        }

        /**
		 * Returns the size of the tiles.
		 */
        public SizeView getTileSize() {
            return o;
        }

        /**
		 * Returns the format (mime-type) of the tile images.
		 */
        public String getImageFormat() {
            return p;
        }

        /**
		 * Returns the unparsed value of the <code>Layers</code> element.
		 */
        public String getLayers() {
            return q;
        }

        /**
		 * Returns the unparsed value of the <code>Styles</code> element.
		 */
        public String getStyles() {
            return styles;
        }
    }

    /**
	 * The handler of the response to a GetCapabilities request. 
	 */
    public static class ResponseHandler extends XmlResponseHandler<Capabilities> {

        /**
		 * Creates a new {@link ResponseHandler} with the specified request URL
		 * and callback to which the resulting {@link Capabilities} object is
		 * delivered.
		 */
        public ResponseHandler(String requestUrl, ResultCallback<Capabilities> callback) {
            super(requestUrl, "WMT_MS_Capabilities", ServiceExceptionReport.ROOT_ELEMENT_NAME, null, callback);
        }

        /**
		 * Parses the document of a successfuly response.
		 */
        protected Capabilities parseResponse(Document document) throws ParseException {
            return new Capabilities(document);
        }
    }
}
