package com.maryanovsky.map.client.opengis.wms;

import java.util.List;
import com.maryanovsky.gwtutils.client.geom.DoublePointView;
import com.maryanovsky.gwtutils.client.geom.Point;
import com.maryanovsky.gwtutils.client.geom.Size;
import com.maryanovsky.gwtutils.client.geom.SizeView;
import com.maryanovsky.map.client.projections.ResolutionListZoomStrategy;
import com.maryanovsky.map.client.projections.ScaleProjection;
import com.maryanovsky.map.client.projections.ZoomStrategy;

/**
 * The projection for the EPSG:4326 SRS.
 * 
 * @author Maryanovsky Alexander
 */
public class Epsg4326Projection extends ScaleProjection implements WmsProjection {

    /**
	 * Creates a new {@link Epsg4326Projection}.
	 * 
	 * @param zoomStrategy The strategy that defines zoom magnification for us.
	 * @param nativeZoom The "native" zoom. There is no special meaning to this
	 * zoom other than that worldSize is given at that zoom. 
	 * @param worldSize The size of the world, in pixels, at the native zoom.
 	 * @param longitudeWraps Whether the map wraps around the X (longitude)
	 * axis.
	 */
    private Epsg4326Projection(ZoomStrategy a, int b, SizeView c, boolean d) {
        super(a, b, c, d, 0);
    }

    /**
	 * Returns an {@link Epsg4326Projection} with the specified list of
	 * resolutions and longitude wrapping strategy.
	 * 
	 * @param resolutions The list of resolutions (map units per pixel) for each
	 * zoom. The first value corresponds to zoom 0 and so on.
	 * @param longitudeWraps Whether the projection will cause the map to wrap
	 * around the X (longitude) axis. 
	 */
    public static Epsg4326Projection getInstance(List<Double> resolutions, boolean longitudeWraps) {
        if (resolutions.size() == 0) throw new IllegalArgumentException("resolutions may not be an empty list");
        ZoomStrategy a = new ResolutionListZoomStrategy(resolutions);
        double b = resolutions.get(0);
        SizeView c = new Size((int) Math.round(360 / b), (int) Math.round(180 / b));
        return new Epsg4326Projection(a, 0, c, longitudeWraps);
    }

    /**
	 * {@inheritDoc}
	 */
    public DoublePointView fromPixelToNative(int x, int y, int zoom) {
        if (Wms.isOutsideWorld(x, y, zoom, worldWidth, worldHeight, nativeZoom, zoomStrategy, longitudeWraps)) return null; else return fromPixelToLatLng(new Point(x, y), zoom);
    }
}
