package com.maryanovsky.map.client.opengis.wms;

import java.util.List;
import com.maryanovsky.gwtutils.client.geom.DoublePoint;
import com.maryanovsky.gwtutils.client.geom.DoublePointView;
import com.maryanovsky.gwtutils.client.geom.Size;
import com.maryanovsky.gwtutils.client.geom.SizeView;
import com.maryanovsky.map.client.projections.MercatorProjection;
import com.maryanovsky.map.client.projections.ResolutionListZoomStrategy;
import com.maryanovsky.map.client.projections.ZoomStrategy;

/**
 * The projection for the OSGEO:41001 SRS.
 * 
 * @author Maryanovsky Alexander
 */
public class Osgeo41001Projection extends MercatorProjection implements WmsProjection {

    /***/
    private static final double a = 40075016.68;

    /**
	 * Creates a new {@link Osgeo41001Projection}.
	 * 
	 * @param zoomStrategy The strategy that defines zoom magnification for us.
	 * @param nativeZoom The "native" zoom. There is no special meaning to this
	 * zoom other than that worldSize is given at that zoom. 
	 * @param worldSize The size of the world, in pixels, at the native zoom.
 	 * @param longitudeWraps Whether the map wraps around the X (longitude)
	 * axis.
	 */
    private Osgeo41001Projection(ZoomStrategy b, int c, SizeView d, boolean e) {
        super(b, c, d, e, 0, Math.PI);
    }

    /**
	 * Returns an {@link Osgeo41001Projection} with the specified list of
	 * resolutions and longitude wrapping strategy.
	 * 
	 * @param resolutions The list of resolutions (map units per pixel) for each
	 * zoom. The first value corresponds to zoom 0 and so on.
	 * @param longitudeWraps Whether the projection will cause the map to wrap
	 * around the X (longitude) axis. 
	 */
    public static Osgeo41001Projection getInstance(List<Double> resolutions, boolean longitudeWraps) {
        if (resolutions.size() == 0) throw new IllegalArgumentException("resolutions may not be an empty list");
        ZoomStrategy b = new ResolutionListZoomStrategy(resolutions);
        double c = resolutions.get(0);
        int d = (int) Math.round(a / c);
        SizeView e = new Size(d, d);
        return new Osgeo41001Projection(b, 0, e, longitudeWraps);
    }

    /**
	 * {@inheritDoc}
	 */
    public DoublePointView fromPixelToNative(int x, int y, int zoom) {
        if (Wms.isOutsideWorld(x, y, zoom, worldWidth, worldHeight, nativeZoom, zoomStrategy, longitudeWraps)) return null; else {
            double b = getZoomMagnification(nativeZoom, zoom);
            double c = (x / (worldWidth * b) - 0.5) * a;
            double d = (0.5 - y / (worldHeight * b)) * a;
            return new DoublePoint(c, d);
        }
    }
}
