package com.maryanovsky.map.client.opengis.wms;

import java.util.ArrayList;
import java.util.List;
import com.google.gwt.xml.client.*;
import com.maryanovsky.gwtutils.client.text.ParseException;
import com.maryanovsky.gwtutils.client.text.ParseUtils;

/**
 * Encapsulates a WMS service exception report. Note that "exception" here means
 * an exceptional response to a WMS request, not exception in the Java sense.
 * 
 * @author Maryanovsky Alexander
 */
public final class ServiceExceptionReport {

    /**
	 * The name of the root element in a service exception response.
	 */
    static final String ROOT_ELEMENT_NAME = "ServiceExceptionReport";

    /***/
    private final Document a;

    /***/
    private final List<ServiceException> b;

    /**
	 * Creates a {@link ServiceExceptionReport} from the specified
	 * Service Exception response XML document.
	 */
    public ServiceExceptionReport(Document document) throws ParseException {
        if (document == null) throw new IllegalArgumentException("document may not be null");
        Wms.checkVersionCompatible(document);
        this.a = document;
        Element c = document.getDocumentElement();
        List<Element> d = ParseUtils.getNamedChildElements(c, ServiceException.ELEMENT_NAME);
        b = new ArrayList<ServiceException>(d.size());
        for (Element e : d) b.add(new ServiceException(e));
    }

    /**
	 * Returns the Service Exception response XML document.
	 */
    public Document getDocument() {
        return a;
    }

    /**
	 * Encapsulates a Service Exception.
	 */
    public static class ServiceException {

        /**
		 * The InvalidFormat exception code.
		 */
        public static final String INVALID_FORMAT_EXCEPTION_CODE = "InvalidFormat";

        /**
		 * The InvalidSRS exception code.
		 */
        public static final String INVALID_SRS_EXCEPTION_CODE = "InvalidSRS";

        /**
		 * The LayerNotDefined exception code.
		 */
        public static final String LAYER_NOT_DEFINED_EXCEPTION_CODE = "LayerNotDefined";

        /**
		 * The StyleNotDefined exception code.
		 */
        public static final String STYLE_NOT_DEFINED_EXCEPTION_CODE = "StyleNotDefined";

        /**
		 * The LayerNotQueryable exception code.
		 */
        public static final String LAYER_NOT_QUERIABLE_EXCEPTION_CODE = "LayerNotQueriable";

        /**
		 * The CurrentUpdateSequence exception code.
		 */
        public static final String CURRENT_UPDATE_SEQUENCE_EXCEPTION_CODE = "CurrentUpdateSequence";

        /**
		 * The InvalidUpdateSequence exception code.
		 */
        public static final String INVALID_UPDATE_SEQUENCE_EXCEPTION_CODE = "InvalidUpdateSequence";

        /**
		 * The MissingDimensionValue exception code.
		 */
        public static final String MISSING_DIMENSION_VALUE_EXCEPTION_CODE = "MissingDimensionValue";

        /**
		 * The InvalidDimensionValue exception code.
		 */
        public static final String INVALID_DIMENSION_VALUE_EXCEPTION_CODE = "InvalidDimensionValue";

        /**
		 * The name of the element describing a ServiceException.
		 */
        static final String ELEMENT_NAME = "ServiceException";

        /***/
        private final Element c;

        /***/
        private final String d;

        /***/
        private final String e;

        /**
		 * Creates a new {@link ServiceException} object from the specified
		 * ServiceException element.
		 */
        private ServiceException(Element f) throws ParseException {
            if (f == null) throw new IllegalArgumentException("element may not be null");
            this.c = f;
            this.d = f.getAttribute("code");
            Node g = f.hasChildNodes() ? f.getFirstChild() : null;
            Text h = (g instanceof Text) ? (Text) g : null;
            this.e = h == null ? null : h.getData();
        }

        /**
		 * Returns the XML element from which this ServiceException object was
		 * created.
		 */
        public Element getElement() {
            return c;
        }

        /**
		 * Returns the exception code. Note that this may be <code>null</code>
		 * (if no exception code was given), one of the exception code
		 * constants defined by this class, or any other string. Comparison
		 * between the returned value and the exception codes defined by this
		 * class is to be done via the {@link String#equals(Object)} method.
		 */
        public String getCode() {
            return d;
        }

        /**
		 * Returns the exception message. May be <code>null</code>.
		 */
        public String getMessage() {
            return e;
        }
    }
}
