package com.maryanovsky.map.client.opengis.wms;

import com.maryanovsky.gwtutils.client.geom.DoublePointView;
import com.maryanovsky.gwtutils.client.geom.SizeView;
import com.maryanovsky.gwtutils.client.net.UrlBuilder;
import com.maryanovsky.map.client.Tile;
import com.maryanovsky.map.client.TileLayer;
import com.maryanovsky.map.client.opengis.wms.Capabilities.TileSet;
import com.maryanovsky.map.client.tiles.ImageTile;

/**
 * A {@link TileLayer} implementation for a WMS server.
 * 
 * @author Maryanovsky Alexander
 */
public class TileSetLayer implements TileLayer {

    /***/
    private final Capabilities a;

    /***/
    private final TileSet b;

    /***/
    private final WmsProjection c;

    /***/
    private final int d;

    /***/
    private final int e;

    /***/
    private final UrlBuilder f;

    /**
	 * Creates a new {@link TileSetLayer} for the specified {@link TileSet}.
	 * 
	 * @param capabilities The {@link Capabilities} of the service.
	 * @param tileSet The tileset to display.
	 * @param projection The projection.
	 * @param minZoom The zoom corresponding to the first (outermost)
	 * resolution of the tileset.
	 * @param bgColor The background color of the tiles, in 0xffRRGGBB format.
	 * For transparent tiles, the high 8 bits should be 0 instead of 0xff.
	 */
    public TileSetLayer(Capabilities capabilities, TileSet tileSet, WmsProjection projection, int minZoom, int bgColor) {
        if (capabilities == null) throw new IllegalArgumentException("capabilities may not be null");
        if (tileSet == null) throw new IllegalArgumentException("tileSet may not be null");
        if (projection == null) throw new IllegalArgumentException("projection may not be null");
        int h = (bgColor >>> 24);
        if ((h != 0xff) && (h != 0)) throw new IllegalArgumentException("The 8 high bits of bgColor may be either 0 or 0xff (got " + Integer.toHexString(h) + ")");
        this.a = capabilities;
        this.b = tileSet;
        this.c = projection;
        this.d = minZoom;
        this.e = bgColor;
        this.f = g();
    }

    /***/
    private UrlBuilder g() {
        UrlBuilder h = new UrlBuilder(a.getGetMapRequestInfo().getUrlPrefix());
        h.addParam("version", Wms.VERSION);
        h.addParam("request", "GetMap");
        h.addParam("tiled", "true");
        h.addParam("layers", b.getLayers());
        h.addParam("styles", b.getStyles());
        h.addParam("srs", b.getSrs());
        h.addParam("format", b.getImageFormat());
        h.addParam("width", b.getTileSize().getWidth());
        h.addParam("height", b.getTileSize().getHeight());
        boolean i = (e >>> 24) == 0;
        int j = e & 0x00ffffff;
        h.addParam("transparent", i ? "TRUE" : "FALSE");
        if (!i) h.addParam("bgcolor", "0x" + Integer.toHexString(j));
        return h;
    }

    /**
	 * {@inheritDoc}
	 */
    public SizeView getTileSize() {
        return b.getTileSize();
    }

    /**
	 * {@inheritDoc}
	 */
    public Tile getTile(int x, int y, int zoom) {
        UrlBuilder h = new UrlBuilder(f);
        if ((zoom < d) || (zoom > d + b.getResolutions().size() - 1)) return null;
        double i = b.getResolutions().get(zoom - d);
        BoundingBox j = b.getBoundingBox();
        double k = b.getTileSize().getWidth() * i;
        double l = b.getTileSize().getHeight() * i;
        DoublePointView m = c.fromPixelToNative(x, y, zoom);
        if (m == null) return null;
        int n = (int) Math.floor((m.getX() - j.getMinX()) / k);
        int o = (int) Math.floor((m.getY() - j.getMinY()) / l);
        double p = j.getMinX() + n * k;
        double q = j.getMinY() + o * l;
        double r = p + k;
        double s = q + l;
        if ((p < j.getMinX()) || (r > j.getMaxX()) || (q < j.getMinY()) || (s > j.getMaxY())) return null;
        h.addParam("bbox", p + "," + q + "," + r + "," + s);
        return ImageTile.makeImageTile(h.getUrl());
    }

    /**
	 * {@inheritDoc}
	 */
    public void releaseTile(Tile tile) {
    }
}
