package com.maryanovsky.map.client.opengis.wms;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import com.google.gwt.xml.client.Document;
import com.google.gwt.xml.client.Element;
import com.maryanovsky.gwtutils.client.text.ParseException;
import com.maryanovsky.map.client.Map;
import com.maryanovsky.map.client.TileLayer;
import com.maryanovsky.map.client.opengis.wms.Capabilities.TileSet;
import com.maryanovsky.map.client.projections.ZoomStrategy;

/**
 * A class for various static WMS values and utilities. Please refer to
 * <a href="http://portal.opengeospatial.org/files/?artifact_id=1081&version=1&format=pdf">
 * the WMS 1.1.1 Implementation Specification</a> for the description of the WMS
 * client/server protocol currently implemented by this package.
 * 
 * @author Maryanovsky Alexander
 */
public final class Wms {

    /**
	 * The WMS version implemented by this package.
	 */
    public static final String VERSION = "1.1.1";

    /**
	 * Checks that the specified document uses a WMS version compatible with the
	 * one implemented by this package. If not, an appropriate
	 * {@link ParseException} is thrown.
	 */
    public static void checkVersionCompatible(Document document) throws ParseException {
        Element b = document.getDocumentElement();
        String c = b.getAttribute("version");
        if (!c.equals(Wms.VERSION)) throw new ParseException("Wrong WMS version (supported=" + Wms.VERSION + ", actual=" + c);
    }

    /**
	 * Creates a {@link Map} from the specified list of {@link TileSet}s. Note
	 * that all the tile sets must have the same SRS.
	 * 
	 * @param capabilities The {@link Capabilities} of the service.
	 * @param tileSets The list of tile sets to show on the map.
	 * @param longitudeWraps Whether the projection will cause the map to wrap
	 * around the X (longitude) axis. 
	 */
    public static Map createMap(Capabilities capabilities, List<TileSet> tileSets, boolean longitudeWraps) {
        if (tileSets == null) throw new IllegalArgumentException("tileSets may not be null");
        if (tileSets.size() == 0) throw new IllegalArgumentException("tileSets may not be empty");
        String b = null;
        for (TileSet c : tileSets) {
            String d = c.getSrs();
            if (b == null) b = d; else if (!b.equals(d)) throw new IllegalArgumentException("All the tile sets must have the same SRS (got " + b + " and " + d + ")");
        }
        List<Double> c = a(tileSets);
        WmsProjection d = getProjection(b, c, longitudeWraps);
        List<TileLayer> e = new ArrayList<TileLayer>(tileSets.size());
        for (TileSet f : tileSets) {
            List<Double> g = f.getResolutions();
            int h = c.indexOf(g.get(0));
            e.add(new TileSetLayer(capabilities, f, d, h, 0x00));
        }
        return new Map(d, e, 0, c.size() - 1);
    }

    /**
	 * Creates a {@link Map} from the specified {@link TileSet}.
	 * 
	 * @param capabilities The {@link Capabilities} of the service.
	 * @param tileSet The tile set to show on the map.
	 * @param longitudeWraps Whether the projection will cause the map to wrap
	 * around the X (longitude) axis. 
	 */
    public static Map createMap(Capabilities capabilities, TileSet tileSet, boolean longitudeWraps) {
        return createMap(capabilities, Collections.singletonList(tileSet), longitudeWraps);
    }

    /***/
    private static List<Double> a(List<TileSet> b) {
        List<Double> c = new ArrayList<Double>();
        int[] d = new int[b.size()];
        int e = 0;
        while (e < b.size()) {
            double f = 0;
            for (int g = 0; g < b.size(); g++) {
                TileSet h = b.get(g);
                List<Double> i = h.getResolutions();
                if (d[g] >= i.size()) continue;
                double j = i.get(d[g]);
                if (j > f) f = j;
            }
            c.add(new Double(f));
            for (int g = 0; g < b.size(); g++) {
                TileSet h = b.get(g);
                List<Double> i = h.getResolutions();
                if (d[g] >= i.size()) continue;
                double j = i.get(d[g]);
                if (j == f) {
                    d[g]++;
                    if (d[g] == i.size()) e++;
                } else if (g > 0) {
                    double k = i.get(d[g - 1]);
                    if ((j < f) && (k > f)) throw new IllegalArgumentException("Received overlapping resolution lists with non-identical resolutions");
                }
            }
        }
        return c;
    }

    /**
	 * Returns a {@link WmsProjection} for the specified SRS and list of
	 * resolutions for each zoom. Currently supported SRSes are
	 * <code>EPSG:4326</code> and <code>OSGEO:41001</code>. Throws an
	 * {@link IllegalArgumentException} if the specified SRS is unsupported.
	 * 
	 * @param srs The SRS.
	 * @param resolutions The list of resolutions (map units per pixel) for each
	 * zoom. The first value corresponds to zoom 0 and so on.
	 * @param longitudeWraps Whether the projection will cause the map to wrap
	 * around the X (longitude) axis. 
	 */
    public static WmsProjection getProjection(String srs, List<Double> resolutions, boolean longitudeWraps) {
        if ("EPSG:4326".equals(srs)) return Epsg4326Projection.getInstance(resolutions, longitudeWraps); else if ("OSGEO:41001".equals(srs)) return Osgeo41001Projection.getInstance(resolutions, longitudeWraps); else throw new IllegalArgumentException("Unsupported SRS: " + srs);
    }

    /**
	 * Package friendly utility method for checking whether the specified pixel
	 * coordinates are inside the world map.
	 */
    static boolean isOutsideWorld(int x, int y, int zoom, double worldWidthNative, double worldHeightNative, int nativeZoom, ZoomStrategy zoomStrategy, boolean longitudeWraps) {
        double b = zoomStrategy.getZoomMagnification(nativeZoom, zoom);
        int c = (int) (worldWidthNative * b);
        int d = (int) (worldHeightNative * b);
        return ((y < 0) || (y >= d)) || (!longitudeWraps && ((x < 0) || (x >= c)));
    }
}
