package com.maryanovsky.map.client.projections;

import java.util.HashMap;
import java.util.Map;
import com.maryanovsky.gwtutils.client.geom.Point;
import com.maryanovsky.gwtutils.client.geom.PointView;
import com.maryanovsky.gwtutils.client.geom.Size;
import com.maryanovsky.gwtutils.client.geom.SizeView;
import com.maryanovsky.map.client.LatLng;
import com.maryanovsky.map.client.Projection;

/**
 * A useful base class for {@link Projection} implementations.
 * <p>It divides the
 * responsibility of handling the projection itself and the zoom magnification
 * to two different objects. The projection at the native zoom is delegated to
 * the subclass using the abstract
 * {@link #fromPixelToLatLngImpl(double, double)} and
 * {@link #fromLatLngToPixelImpl(LatLng)} methods. The zoom magnification, on
 * the other hand, is delegated to a {@link ZoomStrategy}. An implementation
 * of that interface defines the magnification between each two zooms.
 * <p>Additionally, this class takes care of implementing the
 * {@link Projection#getWrapSize(int)} method.
 * 
 * @author Maryanovsky Alexander
 */
public abstract class AbstractProjection implements Projection {

    /**
	 * The "native" zoom, at which world size is specified.
	 */
    protected final int nativeZoom;

    /**
	 * The width of the world, in pixels, at the native zoom.
	 */
    protected final double worldWidth;

    /**
	 * The height of the world, in pixels, at the native zoom.
	 */
    protected final double worldHeight;

    /**
	 * Whether the map wraps around the X (longitude) axis.
	 */
    protected final boolean longitudeWraps;

    /**
	 * The zoom magnification strategy.
	 */
    protected final ZoomStrategy zoomStrategy;

    /***/
    private final Map<Integer, SizeView> a = new HashMap<Integer, SizeView>();

    /**
	 * Creates a new {@link AbstractProjection} with the specified arguments.
	 * 
	 * @param zoomStrategy The strategy that defines zoom magnification for us.
	 * @param nativeZoom The "native" zoom. There is no special meaning to this
	 * zoom other than that <code>worldSize</code> is given at that zoom.
	 * @param worldSize The size of the world, in pixels, at the native zoom.
 	 * @param longitudeWraps Whether the map wraps around the X (longitude)
	 * axis. 
	 */
    public AbstractProjection(ZoomStrategy zoomStrategy, int nativeZoom, SizeView worldSize, boolean longitudeWraps) {
        if (worldSize == null) throw new IllegalArgumentException("worldSize may not be null");
        if (worldSize.getWidth() <= 0) throw new IllegalArgumentException("world width must be positive");
        if (worldSize.getHeight() <= 0) throw new IllegalArgumentException("world height must be positive");
        if (zoomStrategy == null) throw new IllegalArgumentException("zoomMagnificationStrategy may not be null");
        this.nativeZoom = nativeZoom;
        this.worldWidth = worldSize.getWidth();
        this.worldHeight = worldSize.getHeight();
        this.longitudeWraps = longitudeWraps;
        this.zoomStrategy = zoomStrategy;
    }

    /**
	 * Delegates to {@link #fromLatLngToPixelImpl(LatLng)} and then converts the
	 * results to the actual zoom before returning.
	 */
    public Point fromLatLngToPixel(LatLng latLng, int zoom) {
        double[] b = fromLatLngToPixelImpl(latLng);
        double c = getZoomMagnification(nativeZoom, zoom);
        double d = b[0] * c;
        double e = b[1] * c;
        return new Point((int) d, (int) e);
    }

    /**
	 * Converts the pixel coordinates to the native zoom and then delegates to
	 * {@link #fromPixelToLatLngImpl(double, double)}.
	 */
    public LatLng fromPixelToLatLng(PointView pixel, int zoom) {
        double b = getZoomMagnification(zoom, nativeZoom);
        double c = pixel.getX() * b;
        double d = pixel.getY() * b;
        d = Math.max(0, Math.min(worldHeight, d));
        if (longitudeWraps) {
            while (c >= worldWidth) c -= worldWidth;
            while (c < 0) c += worldWidth;
        } else c = Math.max(0, Math.min(worldWidth, c));
        return fromPixelToLatLngImpl(c, d);
    }

    /**
	 * Converts from latitude-longitude to pixel coordinates at the native zoom.
	 * Returns the X and Y coordinates in a 2-element array. The returned values
	 * must be in the range <code>[0, worldWidth or worldHeight]</code>.
	 */
    protected abstract double[] fromLatLngToPixelImpl(LatLng latLng);

    /**
	 * Converts from pixel coordinates to latitude-longitude at the native zoom.
	 * The passed values are guaranteed to be in the range 
	 * <code>[0, worldWidth or worldHeight]</code>.
	 */
    protected abstract LatLng fromPixelToLatLngImpl(double x, double y);

    /**
	 * Delegates to {@link #getWrapSizeImpl(int)} and caches the results.
	 */
    public SizeView getWrapSize(int zoom) {
        SizeView b = a.get(zoom);
        if (b == null) {
            b = getWrapSizeImpl(zoom);
            a.put(zoom, b);
        }
        return b;
    }

    /**
	 * Returns the wrap size at the specified zoom. This method will be called
	 * at most once for each zoom.
	 */
    protected SizeView getWrapSizeImpl(int zoom) {
        int b = longitudeWraps ? (int) (worldWidth * getZoomMagnification(nativeZoom, zoom)) : 0;
        return new Size(b, 0);
    }

    /**
	 * Delegates to the {@link ZoomStrategy} given to us in the
	 * constructor.
	 */
    public double getZoomMagnification(int startZoom, int endZoom) {
        return zoomStrategy.getZoomMagnification(startZoom, endZoom);
    }
}
