package com.maryanovsky.map.client.projections;

import com.maryanovsky.gwtutils.client.MathUtils;
import com.maryanovsky.gwtutils.client.geom.SizeView;
import com.maryanovsky.map.client.LatLng;

/**
 * An implementation of the mercator projection.
 */
public class MercatorProjection extends AbstractProjection {

    /***/
    private static final double DEG_TO_RAD = MathUtils.DEG_TO_RAD;

    /***/
    private static final double RAD_TO_DEG = MathUtils.RAD_TO_DEG;

    /***/
    private final double a;

    /***/
    private final double b;

    /***/
    private final double c;

    /**
	 * Creates a new MercatorProjection with the specified arguments.
	 *
	 * @param zoomStrategy The strategy that defines zoom magnification for us.
	 * @param nativeZoom The "native" zoom. There is no special meaning to this
	 * zoom other than that worldSize is given at that zoom. 
	 * @param worldSize The size of the world, in pixels, at the native zoom.
	 * @param longitudeWraps Whether the map wraps around the X (longitude)
	 * axis. 
	 * @param centralLongitude The longitude at the center of the map.
	 * @param maxMercatorY The Y mercator value at the top (and the opposite of
	 * the value at the bottom) of the map. 
	 */
    public MercatorProjection(ZoomStrategy zoomStrategy, int nativeZoom, SizeView worldSize, boolean longitudeWraps, double centralLongitude, double maxMercatorY) {
        super(zoomStrategy, nativeZoom, worldSize, longitudeWraps);
        if ((centralLongitude <= -180) || (centralLongitude > 180)) throw new IllegalArgumentException("centralLongitude must be in the range (-180,180]");
        this.a = centralLongitude;
        this.b = maxMercatorY;
        this.c = (Math.atan(Math.exp(maxMercatorY)) - Math.PI / 4) * 2 * RAD_TO_DEG;
    }

    /**
	 * {@inheritDoc}
	 */
    protected double[] fromLatLngToPixelImpl(LatLng latLng) {
        double d = worldWidth * (180 + latLng.getLongitude() - a) / 360;
        if (d < 0) d += worldWidth; else if (d > worldWidth) d -= worldWidth;
        double e = Math.max(-c, Math.min(c, latLng.getLatitude()));
        double f = (-Math.log(Math.tan(Math.PI / 4 + e * DEG_TO_RAD / 2)) / b + 1) * worldHeight / 2;
        return new double[] { d, f };
    }

    /**
	 * {@inheritDoc}
	 */
    protected LatLng fromPixelToLatLngImpl(double x, double y) {
        double d = (Math.atan(Math.exp(-(y * 2 / worldHeight - 1) * b)) - Math.PI / 4) * 2 * RAD_TO_DEG;
        double e = x * 360 / worldWidth + a - 180;
        if (e < -180) e += 360; else if (e > 180) e -= 360;
        return new LatLng(d, e);
    }
}
