package com.maryanovsky.map.client.projections;

import com.maryanovsky.gwtutils.client.geom.SizeView;
import com.maryanovsky.map.client.LatLng;

/**
 * A projection which simply scales the latitude and longitude values into the
 * world size rectangle specified in the constructor.   
 * 
 * @author Maryanovsky Alexander
 */
public class ScaleProjection extends AbstractProjection {

    /***/
    private final double a;

    /**
	 * Creates a new {@link ScaleProjection} with the specified arguments.
	 * 
	 * @param zoomStrategy The strategy that defines zoom magnification for us.
	 * @param nativeZoom The "native" zoom. There is no special meaning to this
	 * zoom other than that worldSize is given at that zoom. 
	 * @param worldSize The size of the world, in pixels, at the native zoom.
 	 * @param longitudeWraps Whether the map wraps around the X (longitude)
	 * axis.
	 * @param centralLongitude The longitude at the center of the map.
	 */
    public ScaleProjection(ZoomStrategy zoomStrategy, int nativeZoom, SizeView worldSize, boolean longitudeWraps, double centralLongitude) {
        super(zoomStrategy, nativeZoom, worldSize, longitudeWraps);
        if ((centralLongitude <= -180) || (centralLongitude > 180)) throw new IllegalArgumentException("centralLongitude must be in the range (-180,180]");
        this.a = centralLongitude;
    }

    /**
	 * {@inheritDoc}
	 */
    protected double[] fromLatLngToPixelImpl(LatLng latLng) {
        double b = worldWidth * (180 + latLng.getLongitude() - a) / 360;
        if (b < 0) b += worldWidth; else if (b > worldWidth) b -= worldWidth;
        double c = worldHeight * (90 - latLng.getLatitude()) / 180;
        return new double[] { b, c };
    }

    /**
	 * {@inheritDoc}
	 */
    protected LatLng fromPixelToLatLngImpl(double x, double y) {
        double b = 90 - y * 180 / worldHeight;
        double c = x * 360 / worldWidth + a - 180;
        if (c < -180) c += 360; else if (c > 180) c -= 360;
        return new LatLng(b, c);
    }
}
