package com.maryanovsky.map.client.widgets.overview;

import com.google.gwt.user.client.DOM;
import com.google.gwt.user.client.Event;
import com.google.gwt.user.client.Timer;
import com.google.gwt.user.client.ui.FocusPanel;
import com.google.gwt.user.client.ui.SimplePanel;
import com.google.gwt.user.client.ui.Widget;
import com.maryanovsky.gwtutils.client.Utils;
import com.maryanovsky.gwtutils.client.event.ChangeEvent;
import com.maryanovsky.gwtutils.client.event.ChangeListener;
import com.maryanovsky.gwtutils.client.event.PropertyChangeEvent;
import com.maryanovsky.gwtutils.client.event.PropertyChangeListener;
import com.maryanovsky.gwtutils.client.geom.*;
import com.maryanovsky.gwtutils.client.ui.UiUtils;
import com.maryanovsky.gwtutils.client.ui.mouse.gestures.drag.*;
import com.maryanovsky.map.client.*;

/**
 * A <code>MapWidget</code> which displays an overview map for a certain other
 * <code>MapWidget</code>.
 * The style name of this widget is <code>map-OverviewMapWidget</code>. The
 * rectangle that displays the bounds of the main map consists of 3 elements
 * placed into each other - outer, middle and inner one. Their style names are
 * <code>map-OverviewMapWidgetRectOuter</code>,
 * <code>map-OverviewMapWidgetRectMiddle</code> and
 * <code>map-OverviewMapWidgetRectInner</code> respectively.
 * In order to properly use this widget's default look, you need to import the
 * "widgets/overview/overview.css" file into the page's CSS.
 * 
 * @author Maryanovsky Alexander
 */
public class OverviewMapWidget extends MapWidget {

    /***/
    private final MapWidget a;

    /***/
    private final Synchronization b;

    /***/
    private Widget c;

    /***/
    private LatLng d;

    /***/
    private boolean e = false;

    /**
	 * Creates a new <code>OverviewMapWidget</code> for the specified main map
	 * widget and {@linkplain Synchronization} between them.
	 * 
	 * @param mainMapWidget The main map widget with which we are linked.
	 * @param synchronization The synchronization between the main and overview
	 * map widgets. 
	 */
    public OverviewMapWidget(MapWidget mainMapWidget, Synchronization synchronization) {
        super(synchronization.createOverviewLocationModel(mainMapWidget));
        if (mainMapWidget == null) throw new IllegalArgumentException("mainMapWidget may not be null");
        this.a = mainMapWidget;
        this.b = synchronization;
        mainMapWidget.getMapLocationModel().addChangeListener(new ChangeListener() {

            public void stateChanged(ChangeEvent evt) {
                b.syncOverview(a, OverviewMapWidget.this);
                updateBoundsWidget();
            }
        });
        this.getMapLocationModel().addChangeListener(new ChangeListener() {

            public void stateChanged(ChangeEvent evt) {
                b.syncMain(a, OverviewMapWidget.this);
                updateBoundsWidget();
            }
        });
        PropertyChangeListener f = new PropertyChangeListener() {

            public void propertyChanged(PropertyChangeEvent evt) {
                Map g = (Map) evt.getOldValue();
                Map h = (Map) evt.getNewValue();
                Projection i = g == null ? null : g.getProjection();
                Projection j = h == null ? null : h.getProjection();
                if (!Utils.areEqual(i, j)) replaceBoundsWidget();
            }
        };
        mainMapWidget.addPropertyChangeListener(MAP_PROPERTY_NAME, f);
        this.addPropertyChangeListener(MAP_PROPERTY_NAME, f);
        PropertyChangeListener g = new PropertyChangeListener() {

            public void propertyChanged(PropertyChangeEvent evt) {
                b.syncOverview(a, OverviewMapWidget.this);
                updateBoundsWidget();
            }
        };
        mainMapWidget.addPropertyChangeListener(SIZE_PROPERTY_NAME, g);
        this.addPropertyChangeListener(SIZE_PROPERTY_NAME, g);
        PropertyChangeListener h = new PropertyChangeListener() {

            public void propertyChanged(PropertyChangeEvent evt) {
                updateBoundsWidget();
            }
        };
        mainMapWidget.addPropertyChangeListener(TEMPORARY_OFFSET_PROPERTY_NAME, h);
        this.addPropertyChangeListener(TEMPORARY_OFFSET_PROPERTY_NAME, h);
        setStyleName("map-OverviewMapWidget");
    }

    /**
	 * Returns the main map widget whose overview this widget displays.
	 */
    public final MapWidget getMainMapWidget() {
        return a;
    }

    /**
	 * Pins the bounds widget to the specified location, or unpins it if the
	 * argument is <code>null</code>. This allows forcing it to a location other
	 * than the one displayed by the main widget, when such a thing is necessary
	 * (usually temporarily, during certain animations). 
	 */
    protected final void setPinnedBoundsWidgetLocation(LatLng location) {
        d = location;
    }

    /**
	 * Returns the location of the bounds widget when it is pinned;
	 * <code>null</code> otherwise.  It is up to
	 * {@linkplain #updateBoundsWidget()} to check whether the bounds
	 * widget is pinned and set its location appropriately.
	 */
    protected final LatLng getPinnedBoundsWidgetLocation() {
        return d;
    }

    /**
	 * Tells the overview map widget whether its bounds widget it being dragged.
	 * During a drag, its location should not be touched by
	 * {@linkplain #updateBoundsWidget()}.
	 */
    protected final void setBoundsWidgetDragged(boolean isBoundsWidgetDragged) {
        this.e = isBoundsWidgetDragged;
    }

    /**
	 * Returns whether the bounds widget is currently being dragged. During a
	 * drag, its location should not be touched by
	 * {@linkplain #updateBoundsWidget()}. 
	 */
    protected boolean isBoundsWidgetDragged() {
        return e;
    }

    /**
	 * Returns the widget displaying the bounds of the main map on this overview
	 * map; <code>null</code> if none.
	 */
    protected final Widget getBoundsWidget() {
        return c;
    }

    /**
	 * Replaces the current <code>boundsWidget</code>, if any, with a new one,
	 * as created by {@linkplain #createBoundsWidget()}.
	 */
    protected void replaceBoundsWidget() {
        if (c != null) c.removeFromParent();
        c = createBoundsWidget();
        if (c != null) getWidgetPanel().add(c);
        updateBoundsWidget();
    }

    /**
	 * Creates the widget to display the display bounds of the main map widget
	 * on the overview widget. Returns <code>null</code> for none. This method
	 * may be invoked (by {@linkplain #replaceBoundsWidget()} more than once,
	 * such as when the projection of one of the map(s) changes.
	 */
    protected Widget createBoundsWidget() {
        if ((a.getMap() == null) || (getMap() == null)) return null;
        Projection f = a.getMap().getProjection();
        Projection g = getMap().getProjection();
        if (f.equals(g)) {
            SimplePanel h = new FocusPanel();
            h.setStyleName("map-OverviewMapWidgetRectOuter");
            SimplePanel i = new SimplePanel();
            i.setStyleName("map-OverviewMapWidgetRectMiddle");
            Widget j = new SimplePanel();
            j.setStyleName("map-OverviewMapWidgetRectInner");
            DOM.setStyleAttribute(j.getElement(), "fontSize", "0px");
            UiUtils.addFullSize(h, i);
            UiUtils.addFullSize(i, j);
            FocusPanel k = new FocusPanel() {

                public void onBrowserEvent(Event event) {
                    super.onBrowserEvent(event);
                    if ((event.getTypeInt() & Event.MOUSEEVENTS) != 0) UiUtils.cancelEvent(event);
                }
            };
            DOM.setStyleAttribute(k.getElement(), "position", "absolute");
            UiUtils.setGrabCursor(k);
            UiUtils.makeDraggable(k, k, new HandlesAllDragEvents() {

                public void onDragStart(DragStartEvent evt) {
                    UiUtils.setGrabbingCursor(OverviewMapWidget.this.c);
                }

                public void onDragEnd(DragEndEvent evt) {
                    Widget l = OverviewMapWidget.this.c;
                    setBoundsWidgetDragged(false);
                    UiUtils.setGrabCursor(l);
                    int m = l.getOffsetWidth();
                    int n = l.getOffsetHeight();
                    Point o = UiUtils.getTopLeft(l).moveBy(m / 2, n / 2);
                    LatLng p = localToMap(o);
                    b.moveBoth(a, OverviewMapWidget.this, p);
                }

                public void onDragMove(DragMoveEvent evt) {
                    endCurrentAnimation();
                    setBoundsWidgetDragged(true);
                    if (!isScrolling) maybeScroll();
                }

                public void onDragCancel(DragCancelEvent evt) {
                    UiUtils.setGrabCursor(OverviewMapWidget.this.c);
                }

                private Timer scrollTimer = new Timer() {

                    public void run() {
                        maybeScroll();
                    }
                };

                private boolean isScrolling = false;

                private void maybeScroll() {
                    Widget l = OverviewMapWidget.this.c;
                    Point m = UiUtils.getTopLeft(l);
                    SizeView n = getMapSize();
                    int o;
                    if (m.getX() < 0) o = m.getX(); else if (m.getX() + l.getOffsetWidth() > n.getWidth()) o = m.getX() + l.getOffsetWidth() - n.getWidth(); else o = 0;
                    int p;
                    if (m.getY() < 0) p = m.getY(); else if (m.getY() + l.getOffsetHeight() > n.getHeight()) p = m.getY() + l.getOffsetHeight() - n.getHeight(); else p = 0;
                    o /= 2;
                    p /= 2;
                    if ((o == 0) && (p == 0)) {
                        isScrolling = false;
                        return;
                    } else {
                        isScrolling = true;
                        PointView q = getTemporaryOffset();
                        setTemporaryOffset(new Point(q).moveBy(o, p));
                        scrollTimer.schedule(50);
                    }
                }
            });
            UiUtils.addFullSize(k, h);
            return k;
        } else return null;
    }

    /**
	 * Updates the current <code>boundsWidget</code> to accurately reflect the
	 * bounds of the main map widget display. See also
	 * {@linkplain #getPinnedBoundsWidgetLocation()} and
	 * {@linkplain #isBoundsWidgetDragged()} for special circumstances when the
	 * bounds widget may not necessarily reflect the bounds of the main map 
	 * widget.
	 */
    protected void updateBoundsWidget() {
        if ((c == null) || isBoundsWidgetDragged()) return;
        Projection f = getMap().getProjection();
        SizeView g = a.getMapSize();
        SizeView h = this.getMapSize();
        int i = a.getMapLocationModel().getZoom();
        int j = getMapLocationModel().getZoom();
        Size k = new Size(g).multiplyBy(f.getZoomMagnification(i, j));
        if ((k.getWidth() > h.getWidth() * 0.9) || (k.getHeight() > h.getHeight() * 0.9)) {
            c.setVisible(false);
        } else {
            c.setPixelSize(k.getWidth(), k.getHeight());
            Point l = mapToLocal(d != null ? d : a.getCenter());
            UiUtils.setTopLeft(c, new Point(l.getX() - k.getWidth() / 2, l.getY() - k.getHeight() / 2));
            c.setVisible(true);
        }
    }

    /**
	 * The interface to implement when you want to decide on a custom
	 * relationship between the location of a main map widget and its overview.
	 */
    public static interface Synchronization {

        /**
		 * Creates the {@link MapLocationModel} to be used for the overview
		 * map widget.
		 */
        MapLocationModel createOverviewLocationModel(MapWidget mainWidget);

        /**
		 * Invoked when the location of the main map changes to apply these
		 * changes to the specified overview map widget.
		 */
        void syncOverview(MapWidget mainWidget, OverviewMapWidget overviewWidget);

        /**
		 * Invoked when the location of the overview map changes to apply these
		 * changes to the specified main map widget.
		 */
        void syncMain(MapWidget mainWidget, OverviewMapWidget overviewWidget);

        /**
		 * Invoked to center both the main and the overview map widgets on the
		 * specified coordinate.
		 */
        void moveBoth(MapWidget mainWidget, OverviewMapWidget overviewWidget, LatLng location);
    }

    /**
	 * The most commonly used implementation of {@linkplain Synchronization}.
	 * It positions the overview map at the same location as the main map,
	 * zoomed out by the specified number of zooms.
	 */
    public static class StandardSynchronization implements Synchronization {

        /***/
        private final double f;

        /***/
        private final boolean g;

        /***/
        private int h = 0;

        /***/
        private int i = 0;

        /**
		 * Creates a new <code>StandardSynchronization</code> with the specified
		 * main map portion threshold and optionally animated transitions on
		 * both maps.
		 * 
		 * @param mainMapPortionThreshold The maximum portion of the width and
		 * height of the overview map that the area of the main map is to be
		 * allowed to occupy (whenever possible). This determines the zoom at
		 * which the overview map is displayed.
		 * @param animate Whether to animate transitions on both maps. Note that
		 * a <code>true</code> value is merely a suggestion - certain
		 * transitions will not be animated regardless of the value of this
		 * parameter.
		 */
        public StandardSynchronization(double mainMapPortionThreshold, boolean animate) {
            if (mainMapPortionThreshold <= 0) throw new IllegalArgumentException("mainMapPortionThreshold must be positive");
            this.f = mainMapPortionThreshold;
            this.g = animate;
        }

        /**
		 * Creates a new {@link StandardSynchronization} with optionally
		 * animated transitions and the default main map portion threshold.
		 */
        public StandardSynchronization(boolean animate) {
            this(0.4, animate);
        }

        /**
		 * {@inheritDoc}
		 */
        public MapLocationModel createOverviewLocationModel(MapWidget mapWidget) {
            MapLocationModel k = mapWidget.getMapLocationModel();
            return new MapLocationModel(k.getMinZoom(), k.getMaxZoom(), k.getCenter(), k.getZoom());
        }

        /**
		 * {@inheritDoc}
		 */
        public void syncMain(MapWidget mainWidget, OverviewMapWidget overviewWidget) {
            if (i > 0) {
                i--;
                return;
            }
            LatLng k = overviewWidget.getMapLocationModel().getCenter();
            h++;
            mainWidget.moveTo(k, g && mainWidget.shouldAnimateTo(k));
        }

        /***/
        private int j(MapWidget k, OverviewMapWidget l) {
            MapLocationModel m = k.getMapLocationModel();
            MapLocationModel n = l.getMapLocationModel();
            Projection o = k.getMap().getProjection();
            Projection p = k.getMap().getProjection();
            SizeView q = k.getMapSize();
            SizeView r = l.getMapSize();
            if (o.equals(p)) {
                Projection s = o;
                double t = f * Math.min(((double) r.getWidth()) / q.getWidth(), ((double) r.getHeight()) / q.getHeight());
                int u = m.getZoom();
                int v = u;
                while (v > n.getMinZoom()) {
                    double w = s.getZoomMagnification(u, v);
                    if (w <= t) break;
                    v--;
                }
                return v;
            } else {
                LatLng s = k.localToMap(new Point(0, 0));
                LatLng t = k.localToMap(new Point(q));
                int u = n.getMaxZoom();
                while (u > n.getMinZoom()) {
                    PointView v = p.fromLatLngToPixel(s, u);
                    PointView w = p.fromLatLngToPixel(t, u);
                    Rectangle x = Rectangle.makeCover(v, w);
                    if ((((double) x.getWidth()) / r.getWidth() <= f) && (((double) x.getHeight()) / r.getHeight() <= f)) break;
                    u--;
                }
                return u;
            }
        }

        /**
		 * {@inheritDoc}
		 */
        public void syncOverview(MapWidget mainWidget, OverviewMapWidget overviewWidget) {
            if (h > 0) {
                h--;
                return;
            }
            if (mainWidget.getMapSize().isEmpty() || overviewWidget.getMapSize().isEmpty()) return;
            MapLocationModel k = mainWidget.getMapLocationModel();
            MapLocationModel l = overviewWidget.getMapLocationModel();
            LatLng m = k.getCenter();
            int n = j(mainWidget, overviewWidget);
            i++;
            if (n == l.getZoom()) overviewWidget.moveTo(m, g && overviewWidget.shouldAnimateTo(m)); else l.set(m, n);
        }

        /**
		 * {@inheritDoc}
		 */
        public void moveBoth(MapWidget mainWidget, final OverviewMapWidget overviewWidget, LatLng location) {
            h++;
            i++;
            overviewWidget.setPinnedBoundsWidgetLocation(location);
            overviewWidget.getMapLocationModel().addChangeListener(new ChangeListener() {

                public void stateChanged(ChangeEvent evt) {
                    overviewWidget.setPinnedBoundsWidgetLocation(null);
                    overviewWidget.getMapLocationModel().removeChangeListener(this);
                }
            });
            mainWidget.moveTo(location, g && mainWidget.shouldAnimateTo(location));
            overviewWidget.moveTo(location, g && overviewWidget.shouldAnimateTo(location));
        }
    }
}
