package com.maryanovsky.map.client.widgets.scale;

import com.maryanovsky.gwtutils.client.MathUtils;
import com.maryanovsky.gwtutils.client.geom.Point;
import com.maryanovsky.gwtutils.client.geom.SizeView;
import com.maryanovsky.map.client.LatLng;
import com.maryanovsky.map.client.MapWidget;

/**
 * Implements {@link ScaleWidget.Scale} for maps of the planet earth. The earth
 * is modelled as a perfect sphere of its actual equatorial radius. This scale
 * is therefore imprecise, especially at locations far from the equator (but
 * it's good enough for most uses).
 * 
 * @author Maryanovsky Alexander
 */
public class EarthScale implements ScaleWidget.Scale {

    /**
	 * The equatorial radius of the earth in meters.
	 */
    public static final double EARTH_RADIUS_METERS = 6378137;

    /***/
    private static final double[] a = new double[] { 0.01, 0.02, 0.05, 0.1, 0.2, 0.5, 1, 2, 5, 10, 20, 50, 100, 200, 500, 1000, 2000, 5000, 10000, 20000, 50000, 100000, 200000, 500000, 1000000, 2000000, 5000000, 10000000 };

    /***/
    private static final String[] b = new String[] { "1cm", "2cm", "5cm", "10cm", "20cm", "50cm", "1m", "2m", "5m", "10m", "20m", "50m", "100m", "200m", "500m", "1km", "2km", "5km", "10km", "20km", "50km", "100km", "200km", "500km", "1000km", "2000km", "5000km", "10000km" };

    /***/
    private static final double[] c = new double[] { 0.0127, 0.0254, 0.0508, 0.1016, 0.1524, 0.3048, 0.6096, 1.524, 3.048, 6.096, 15.24, 30.48, 60.96, 152.4, 304.8, 609.6, 1609.344, 3218.688, 8046.72, 16093.44, 32186.88, 80467.2, 160934.4, 321868.8, 804672, 1609344, 3218688, 8046720, 16093440 };

    /***/
    private static final String[] d = new String[] { "0.5in", "1in", "2in", "4in", "6in", "1ft", "2ft", "5ft", "10ft", "20ft", "50ft", "100ft", "200ft", "500ft", "1000ft", "2000ft", "1mi", "2mi", "5mi", "10mi", "20mi", "50mi", "100mi", "200mi", "500mi", "1000mi", "2000mi", "5000mi", "10000mi" };

    /***/
    private final int e;

    /***/
    private final double[] f;

    /***/
    private final String[] g;

    /**
	 * Creates a new {@link EarthScale} with the specified preferred pixel size,
	 * array of possible sizes and their names.
	 * 
	 * @param preferredSize The preferred size of the scale, in pixels.
	 * @param sizes The possible displayed sizes, in meters, in increasing
	 * order.
	 * @param sizeNames The names of the possible displayed sizes, in the same
	 * order as sizes.
	 */
    public EarthScale(int preferredSize, double[] sizes, String[] sizeNames) {
        if (preferredSize <= 0) throw new IllegalArgumentException("preferredSize must be positive");
        if ((sizes == null) || (sizes.length == 0)) throw new IllegalArgumentException("sizes must be non-null and non-empty");
        if ((sizeNames == null) || (sizeNames.length == 0)) throw new IllegalArgumentException("sizeNames must be non-null and non-empty");
        if (sizeNames.length != sizes.length) throw new IllegalArgumentException("sizes and sizeNames must be of the same length");
        for (int i = 0; i < sizes.length - 1; i++) if (sizes[i] >= sizes[i + 1]) throw new IllegalArgumentException("sizes must be ordered in increasing order");
        this.e = preferredSize;
        this.f = sizes;
        this.g = sizeNames;
    }

    /**
	 * Returns an instance of {@link EarthScale} which displays metric sizes in
	 * English.
	 * 
	 * @param preferredSize The preferred size of the scale, in pixels.
	 */
    public static EarthScale getEnglishMetricInstance(int preferredSize) {
        return new EarthScale(preferredSize, a, b);
    }

    /**
	 * Returns an instance of {@link EarthScale} which displays imperial sizes
	 * in English.
	 * 
	 * @param preferredSize The preferred size of the scale, in pixels.
	 */
    public static EarthScale getEnglishImperialInstance(int preferredSize) {
        return new EarthScale(preferredSize, c, d);
    }

    /**
	 * {@inheritDoc}
	 */
    public int getSizePixels(MapWidget mapWidget) {
        double i = getMetersPerPixel(mapWidget);
        if (i == 0) return 0;
        double j = i * e;
        int k = h(j);
        double l = f[k];
        int m = (int) (l / i);
        return m;
    }

    /**
	 * {@inheritDoc}
	 */
    public String getSizeText(MapWidget mapWidget) {
        double i = getMetersPerPixel(mapWidget);
        if (i == 0) return null;
        double j = i * e;
        int k = h(j);
        return g[k];
    }

    /**
	 * Returns the number of meters in a pixel displayed by the specified
	 * {@link MapWidget}. Returns 0 if unable to determine.
	 */
    public static double getMetersPerPixel(MapWidget mapWidget) {
        SizeView i = mapWidget.getMapSize();
        Point j = new Point(i.getWidth() / 2, i.getHeight() / 2);
        Point k = new Point(j.getX() + 1, j.getY());
        LatLng l = mapWidget.localToMap(j);
        LatLng m = mapWidget.localToMap(k);
        if (l.equals(m)) {
            k = new Point(j.getX() - 1, j.getY());
            m = mapWidget.localToMap(k);
        }
        if (l.equals(m)) return 0;
        return getArcLength(l, m) * EARTH_RADIUS_METERS;
    }

    /**
	 * Returns the angle of the shortest arc between the two specified
	 * locations, on a sphere of the given radius.
	 */
    public static double getArcLength(LatLng loc1, LatLng loc2) {
        double i = loc1.getLatitude() * MathUtils.DEG_TO_RAD;
        double j = loc1.getLongitude() * MathUtils.DEG_TO_RAD;
        double k = loc2.getLatitude() * MathUtils.DEG_TO_RAD;
        double l = loc2.getLongitude() * MathUtils.DEG_TO_RAD;
        double m = l - j;
        double n = Math.sin(i);
        double o = Math.sin(k);
        double p = Math.cos(i);
        double q = Math.cos(k);
        double r = Math.sin(m);
        double s = Math.cos(m);
        return MathUtils.atan2(Math.sqrt(MathUtils.sqr(q * r) + MathUtils.sqr(p * o - n * q * s)), n * o + p * q * s);
    }

    /***/
    private int h(double i) {
        int j = 0;
        double k = Math.abs(i - f[j]);
        for (int l = 0; l < f.length; l++) {
            double m = Math.abs(i - f[l]);
            if (m < k) {
                j = l;
                k = m;
            }
        }
        return j;
    }
}
